const models = require('../models');
const geoip = require('geoip-lite2');
const projectConfig = require('../config/config')['project'][
  process.env.NODE_ENV
];
let path = require('path');
let appDir = path.dirname(require.main.filename);
appDir = appDir.substring(0, appDir.lastIndexOf('/'));
const { parseAsync } = require('json2csv');
var fs = require('fs');
const moment = require('moment');
const archiver = require('archiver');
const httpErrors = require('http-errors');
const Sequelize = require('sequelize');

/**
 * GET client access log listing
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
const getLogs = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      logtype,
      clientid,
      searchKey,
      start_date,
      end_date,
      clientos
    } = req.query;

    const options = {
      where: {
        [Op.and]: []
      },
      limit: +limit,
      offset: +offset,
      order: [],
      raw: true
    };

    const publicIP = projectConfig.publicIP;

    //Managing Search Params
    let isOnlyRunning = false;
    let isRunningAlsoChecked = false;
    let isRunning = false;
    if (logtype && logtype.split(',').length !== 4) {
      let loginSuccessFullArray = [];
      let logoutSuccessFullArray = [];
      if (logtype.includes(0)) {
        loginSuccessFullArray.push(0);
        isRunning = true;
      }
      if (logtype.includes(1)) {
        isOnlyRunning = true;
        logoutSuccessFullArray.push(0);
      }
      if (logtype.includes(2)) {
        isOnlyRunning = true;
        logoutSuccessFullArray.push(1);
      }
      if (logtype.includes(3)) {
        loginSuccessFullArray.push(1);
        isRunningAlsoChecked = true;
        options.where[Op.and].push({
          DISCONNECTION_TIME: null
        });
      }

      options.where[Op.or] = [
        {
          is_login_successfull: {
            [Op.in]: loginSuccessFullArray
          }
        },
        {
          is_logout_successfull: {
            [Op.in]: logoutSuccessFullArray
          }
        }
      ];
      if (isRunningAlsoChecked && isOnlyRunning) {
        options.where[Op.and].pop();
        options.where[Op.or][0] = {};
        options.where[Op.or].push({
          [Op.and]: [
            {
              DISCONNECTION_TIME: null
            },
            {
              is_login_successfull: {
                [Op.in]: !isRunning ? [1] : [0, 1]
              }
            }
          ]
        });
      }
    }

    if (clientid) {
      options.where.client_userName = clientid;
    }    

    if (searchKey) {
      let name = `%${searchKey}%`;
      options.where[Op.and].push({
        [Op.or]: [
          {
            client_name: {
              like: name
            }
          },
          {
            ipv4_address: {
              like: name
            }
          },
          {
            location: {
              like: name
            }
          }
        ]
      });
    }

    let CONNECTION_ATTEMPT_TIME = {};
    if (start_date && end_date) {
      CONNECTION_ATTEMPT_TIME.gte = start_date;
      CONNECTION_ATTEMPT_TIME.lte = end_date + ' 23:59:59';
      options.where.CONNECTION_ATTEMPT_TIME = CONNECTION_ATTEMPT_TIME;
    }

    if (clientos) {
      options.where.client_device_os = {
        [Op.in]: clientos.split(',')
      };
    }

    const allowedSortValues = ['client_name', 'ipv4_address', 'location', 'id'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    const result = await models.Connection.paginate(options);

    const null_locations = []; //to store null location client log details

    (result.data || []).forEach(function(dat) {
      let ipToUseForLocation;
      if (!dat.ipv4_address) {
        dat.ipv4_address = publicIP;
        ipToUseForLocation = publicIP;
      } else {
        ipToUseForLocation = dat.ipv4_address;
      }
      let parts = ipToUseForLocation.split('.');
      let is_private =
        parts[0] === '10' ||
        parts[0] === '172' ||
        (parts[0] === '192' && parts[1] === '168');
      if (is_private) {
        ipToUseForLocation = publicIP;
      }
      var geo = geoip.lookup(ipToUseForLocation);
      if (dat.client_geoip_track && geo && geo.ll) {
        //Using Email as region field
        dat.region = geo.city + ', ' + geo.country;
      } else {
        dat.region = 'N/A';
      }
      if (!dat.location) {
        //choosing null location client logs to update
        null_locations.push({ id: dat.id, region: dat.region });
      }
    });

    await Promise.all(null_locations.map(location => 
      //filling null location client logs
      models.Connection.update(
        { location: location.region },
        { where: { id: location.id } }
      )
    ));

    res.json(result);
  } catch (error) {
    next(error);
  }
};

/* GET client access log listing. */
const getClients = async (req, res, next) => {
  try {
    const data = await models.Client.findAll({
      where: {
        is_active: 1
      }
    });

    const response = {
      clients: data
    };

    res.json(response);
  } catch (error) {
    next(error);
  }
};

const purge = async (req, res, next) => {
  try {
    await models.Connection.destroy({
      truncate: true
    });

    res.json({
      success: true,
      message: 'Delete Successfully'
    });
  } catch (error) {
    next(error);
  }
};

const exportCSV = async (req, res, next) => {
  try {
     const { 
      logtype,
      clientid,
      searchKey,
      start_date,
      end_date,
      clientos,
      version
    } = req.query;
    const user = req.user;
    const timestamp = moment().format("YYYY-MM-DD_HH-mm-ss");
    const fileName = `invisily-${version}-client-access-logs-${user.id}-${timestamp}`
    const zipFileName = `${fileName}.zip`;
    const csvFileName = `${fileName}.csv`;
    const Path = path.join(__dirname, '../storage/' + zipFileName);

    const options = {
      where: {
        [Op.and]: []
      },
      order: [],
      raw: true
    };

    const publicIP = projectConfig.publicIP;

    //Managing Search Params
    let isOnlyRunning = false;
    let isRunningAlsoChecked = false;
    let isRunning = false;
    if (logtype && logtype.split(',').length !== 4) {
      let loginSuccessFullArray = [];
      let logoutSuccessFullArray = [];
      if (logtype.includes(0)) {
        loginSuccessFullArray.push(0);
        isRunning = true;
      }
      if (logtype.includes(1)) {
        isOnlyRunning = true;
        logoutSuccessFullArray.push(0);
      }
      if (logtype.includes(2)) {
        isOnlyRunning = true;
        logoutSuccessFullArray.push(1);
      }
      if (logtype.includes(3)) {
        loginSuccessFullArray.push(1);
        isRunningAlsoChecked = true;
        options.where[Op.and].push({
          DISCONNECTION_TIME: null
        });
      }

      options.where[Op.or] = [
        {
          is_login_successfull: {
            [Op.in]: loginSuccessFullArray
          }
        },
        {
          is_logout_successfull: {
            [Op.in]: logoutSuccessFullArray
          }
        }
      ];
      if (isRunningAlsoChecked && isOnlyRunning) {
        options.where[Op.and].pop();
        options.where[Op.or][0] = {};
        options.where[Op.or].push({
          [Op.and]: [
            {
              DISCONNECTION_TIME: null
            },
            {
              is_login_successfull: {
                [Op.in]: !isRunning ? [1] : [0, 1]
              }
            }
          ]
        });
      }
    }

    if (clientid) {
      options.where.client_userName = clientid;
    }    

    if (searchKey) {
      let name = `%${searchKey}%`;
      options.where[Op.and].push({
        [Op.or]: [
          {
            client_name: {
              like: name
            }
          },
          {
            ipv4_address: {
              like: name
            }
          },
          {
            location: {
              like: name
            }
          }
        ]
      });
    }

    let CONNECTION_ATTEMPT_TIME = {};
    if (start_date && end_date) {
      CONNECTION_ATTEMPT_TIME.gte = start_date;
      CONNECTION_ATTEMPT_TIME.lte = end_date + ' 23:59:59';
      options.where.CONNECTION_ATTEMPT_TIME = CONNECTION_ATTEMPT_TIME;
    }

    if (clientos) {
      options.where.client_device_os = {
        [Op.in]: clientos.split(',')
      };
    }
 
      const result = await models.Connection.findAll(options);

      const null_locations = []; //to store null location client log details

      (result.data || []).forEach(function(dat) {
      let ipToUseForLocation;
      if (!dat.ipv4_address) {
        dat.ipv4_address = publicIP;
        ipToUseForLocation = publicIP;
      } else {
        ipToUseForLocation = dat.ipv4_address;
      }
      let parts = ipToUseForLocation.split('.');
      let is_private =
        parts[0] === '10' ||
        parts[0] === '172' ||
        (parts[0] === '192' && parts[1] === '168');
      if (is_private) {
        ipToUseForLocation = publicIP;
      }
      var geo = geoip.lookup(ipToUseForLocation);
      if (dat.client_geoip_track && geo && geo.ll) {
        //Using Email as region field
        dat.region = geo.city + ', ' + geo.country;
      } else {
        dat.region = 'N/A';
      }
      if (!dat.location) {
        //choosing null location client logs to update
        null_locations.push({ id: dat.id, region: dat.region });
      }
      });

      await Promise.all(null_locations.map(location => 
      //filling null location client logs
      models.Connection.update(
        { location: location.region },
        { where: { id: location.id } }
      )
      ));

      for (let log of result) {
        if (!log.is_login_successfull){
          log.eventStatus = 'Failed Login';
        } else if (!log.is_logout_successfull && log.is_logout_successfull != null){
          log.eventStatus = 'Forcefully Closed';
        } else {
          if (log.CONNECTION_ATTEMPT_TIME != null && log.DISCONNECTION_TIME != null){
            log.eventStatus = 'Gracefully Closed';
          } else if (log.DISCONNECTION_TIME == null) {
            log.eventStatus = 'Running';
          } else {
            log.eventStatus = 'Failed';
          }
        }
      }

      const fields = [
        {
          label: 'ID',
          value: 'id'
        },
        {
          label: 'Name',
          value: 'client_name'
        },
        {
          label: 'IP',
          value: 'ipv4_address'
        },
        {
          label: 'STATUS',
          value: 'eventStatus'
        },
        {
          label: 'OS',
          value: 'client_device_os'
        },
        {
          label: 'LOCATION',
          value: 'location'
        },{
          label: 'COUNTRY',
          value: 'status'
        },
        {
          label: 'TIME IN',
          value: 'CONNECTION_ATTEMPT_TIME'
        },
        {
          label: 'TIME OUT',
          value: 'DISCONNECTION_TIME'
        },
        {
          label: 'SESSION TIME',
          value: 'sessionTime'
        }
      ];
      const opts = { fields, quote: '' };
      const csv = await parseAsync(result, opts)

      const zipStream = fs.createWriteStream(Path);

      const archive = archiver('zip', {
        zlib: { level: 9 }
      });

      archive.pipe(zipStream);

      archive.append(csv, { name: csvFileName });

      archive.on('error', err => {
        res.status(500).send('Error occurred while creating the ZIP file.');
      });

      archive.finalize();
      console.log('result\n\n', opts)

      zipStream.on('close', () => {
        res.download(Path, zipFileName, err => {
          if (err) {
            throw err;
          } else {
            fs.unlinkSync(Path);
          }
        });
      });
    
  } catch (error) {
      console.log('error:', error.message)
      res.status(500).send(error.message)
  }

};

const getClientServices = async (req, res, next) => {
  try{
    const clientId = req.params.id;

    const connectionInfo = await models.Connection.findOne({
      attributes: [
        'AGENT_ID',
        'SESSION_ID'
      ],
      where: {
        id: clientId,
      },
    });

    if (connectionInfo) {
      sessionId = connectionInfo.dataValues.SESSION_ID;
      agentId = connectionInfo.dataValues.AGENT_ID;

      const serviceDetails = await models.ServiceAccessLogs.findAll({
        attributes: [
          'service_id',
          [Sequelize.fn('MAX', Sequelize.col('service_access_time')), 'latest_service_access_time'],
        ],
        where: {
          agent_id: agentId,
          session_id: sessionId,
        },
        group: ['service_id'],
        raw: true,
        nest: true,
        include: [{
          model: models.Service,
          attributes: ['Name'],
          required: true,
        }],
      });

      const formattedServiceDetails = serviceDetails.map(detail => {
        return {
          service_id: detail.service_id,
          service_access_time: new Date(detail.latest_service_access_time * 1000),
          // service_access_time: new Date(detail.latest_service_access_time * 1000)
          // .toLocaleString('en-GB', {
          //   day: '2-digit',
          //   month: '2-digit',
          //   year: 'numeric',
          //   hour: '2-digit',
          //   minute: '2-digit',
          //   hour12: false
          // }),
          Name: detail.Service.Name || 'N/A',
        };
      });

      res.status(200).json({
        success: true,
        message: 'Service details retrieved successfully',
        serviceDetails: formattedServiceDetails,
      });



    }else{
      throw new httpErrors.NotFound(`Session id or agent id not found`);
    }

  } catch (e){
    console.log('ererererer', e);
    if (e instanceof httpErrors.NotFound) {
      return res.status(404).send({
          success: false,
          message: e.message
      });
    }else{
      return res.status(500).send({
          success: false,
          message: `An error occured`
      });
    }
  }
};

module.exports = {
  getClients,
  purge,
  getLogs,
  exportCSV,
  getClientServices
};
