var httpErrors = require('http-errors');
var models = require('../models');
var Openssl = require('../openssl/openssl.js');
var OpensslObj = new Openssl();
let projectConfig = require('../config/config')['project'][
  process.env.NODE_ENV
];
let path = require('path');
let appDir = path.dirname(require.main.filename);
appDir = appDir.substring(0, appDir.lastIndexOf('/'));
let helper = require('../routes/common/helper');
let settings = require('../routes/common/settings');
let { exec } = require('child_process');
let fs = require('fs');
const {promisify} = require('util');
let fsExtra = require('fs-extra');
let archiver = require('archiver');
const moment = require('moment');

const copyFilePromise = promisify(fs.copyFile);
const scriptsPath = path.join(__dirname, '..', 'scripts');

const utils = require('../utils/common');

const { actionType, publishControllers, publishControllerSoftware } = require('../routes/common/event-publisher');

const {generateLink} = require('../routes/common/content-upload');
const {
  insertIntoAdminLogs
} = require('../routes/common/adminLogs');
const { 
  getControllerEditMessage,
  getControllerDeleteMessage,
  getControllerCreateMessage,
  getControllerEmailMessage,
  getControllerDownloadlMessage
} = require('../config/adminLogs.config')
// const { addtoQueue } = require('../queues/uploadToContent');

let installerType = {
  ubuntu18: 'linux64',
  ubuntuserver18: 'linux64',
  ubuntu: 'linux64',
  ubuntuserver20armv8: 'armv8',
  windows: 'win64',
  macintosh: 'mac'
};

let mapping = {
  ubuntu18: 'deb',
  ubuntuserver18: 'deb',
  ubuntuserver20armv8: 'deb',
  ubuntu: 'deb',
  windows: 'exe',
  macintosh: 'dmg'
};

var emailLinkExpiryInMinutes = 60;

async function setExpiryTime() {
  emailLinkExpiryInMinutes = await settings.getExpirySettings();
}

setExpiryTime();

// models.ControllerHealth.create({
//   controllerId:104,
//   OS:'test',
//   CPU:'Ubuntu'
// });

/* GET controller listing. */
const getControllers = async (req, res, next) => {
  try {
    const {
      limit = 99999,
      offset = 0,
      sort = 'id',
      dir = 'desc',
      searchKey = ''
    } = req.query;

    const options = {
      where: { },
      limit: limit,
      offset: offset,
      order: [],
      include: [
        {
          model: models.ControllerHealth,
          as: 'ControllerHealth'
        },
        {
          model: models.ComponentRole,
          as: 'Role',
          attributes: ['role']
        }
      ]
    };

    if (searchKey.trim()) {
      options.where.Name = {
        [Op.like]: `%${searchKey.trim()}%`
      };
    }

    const allowedSortValues = ['Name', 'id'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    const result = await models.Controller.paginate(options);
    res.json(result);
  } catch (error) {
    next(error);
  }
};

const status = async (req, res, next) => {
  try {
    const { controller } = req.body;

    const controllerToUpdate = await models.Controller.findOne({
      where: {
        id: controller.id
      }
    });

    if (!controllerToUpdate) {
      throw new httpErrors.NotFound('No controller found.');
    }

    controller.is_active = !controller.is_active;

    await controllerToUpdate.update(controller);

    res.json({
      success: true,
      message: 'Update Successfully'
    });
  } catch (error) {
    next(error);
  }
};

const updateController = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();

  try {

    const isExist = await models.Controller.findOne({
      where: {
        Name: req.body.controller.Name,
        id: {
          [Op.ne] : req.params.id
        }
      }
    });

    if(isExist){
      throw new httpErrors.Conflict('Controller already exists with the same name.');
    }

    const controller = await models.Controller.findOne({
      where: {
        id: req.params.id
      },
      include: [
        {
          model: models.ComponentRole,
          as: 'Role',
          attributes: ['role']
        }
      ]
    });

    if (!controller) {
      throw new httpErrors.NotFound('No controller found to update.');
    }

    generateInstallerZip(controller, '0', false);
    
    const data = await controller.update(req.body.controller, {
      transaction
    });
    
    await removeControllerDBPermission(controller, transaction);
    addControllerDBPermission(data, transaction);
    await transaction.commit();
    publishControllers(actionType.UPDATED, {
      id: data.id
    });
    insertIntoAdminLogs(
      getControllerEditMessage(controller.Name),
      req
    )
    res.json({
      success: true,
      message: 'Edited Successfully'
    });
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/* Get a single controller detail */
const getController = async (req, res, next) => {
  try {
    const data = await models.Controller.findAll({
      where: {
        id: req.params.id
      }
    });

    res.json(data);
  } catch (error) {
    next(error);
  }
};

// add new controller
const addController = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();

  try {
    let { controller } = req.body;

    const isExist = await models.Controller.findOne({
      where: {
        [Op.or]: [{
          Name: controller.Name
        }, {
          ipv4_address: controller.ipv4_address
        }]
      }
    });

    if(isExist){
      throw new httpErrors.Conflict('Controller already exists with the same Name/Private IP.');
    }

    const [encKey, hmacKey] = await Promise.all([
      OpensslObj.generateEncKey(),
      OpensslObj.generateHMacKey()

    ]);

    controller.enc_key = encKey.key;
    controller.HMacKey = hmacKey.key;

    controller.controller_id = 0;
    const data = await models.Controller.create(controller, {
      transaction
    });

    data.controller_id = data.id;
    await removeControllerDBPermission(data, transaction);

    const _data = await data.save({
      transaction
    });
    
    await addControllerDBPermission(_data, transaction);

    generateInstallerZip(_data, '0', true);
    await transaction.commit();


    publishControllers(actionType.ADDED, {
      id: data.id
    });

    insertIntoAdminLogs(
      getControllerCreateMessage(controller.Name),
      req
    )
    res.json({
      success: true,
      message: 'Add Successfully'
    });
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

// delete a single controller
const deleteController = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();

  try {
    const options = {
      where: {
        id: req.params.id
      },
      include: [
        {
          model: models.ComponentRole,
          as: 'Role',
          attributes: ['role']
        }
      ]
    };

    const controller = await models.Controller.findOne(options);

    if (!controller) {
      throw new httpErrors.NotFound('No controller to delete.');
    }

    if(controller.Role && controller.Role.role == 'Primary'){
      throw new httpErrors.BadRequest('Cannot delete primary controller.');
    }


    await Promise.all([
      controller.destroy(
        {
          transaction
        }
      ),
      removeControllerDBPermission(controller, transaction),
      models.ControllerHealth.destroy({
        where: {
          controllerId: req.params.id
        },
        transaction
      })
    ]);

    await publishControllers(actionType.DELETED, {
      id: controller.id
    });

    await transaction.commit();
    insertIntoAdminLogs(
      getControllerDeleteMessage(controller.Name),
      req
    )

    res.json({
      success: true,
      message: 'Delete Successfully'
    });
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const downloadUpgrade = async (req, res, next) => {
  try {
    const os = req.params.os;

    let folderPath =
      `${projectConfig.serviceInstallerPath}/controller/output/` + os;
    console.log('Starting Download');

    const currentTime = moment();

    const controller = await models.Controller.findOne({
      where: {
        updated_at: {
          $gt: currentTime
        }
      }
    })

    if (!controller) {
      throw new httpErrors(304, 'Controller does not need to be updated');
    }

    let filePath = `${folderPath}/invisily-controller-` + installerType[os] + '-' + controller.id + `.zip`;
    // let downloadFile = `invisily-controller-` + installerType[os] + `.zip`;

    if (!fs.existsSync(filePath)) {
      throw new httpErrors.NotFound('Controller installer not found');
    }


    let packageName = await utils.getPackageName(os,'controller','',controller.id);

    res.download(filePath, packageName, function(err) {
      // file download is successful, expire link now
      if (err) {
        console.log('Error Jenkins transfer', err);
        throw new httpErrors.BadRequest('Error Jenkins transfer: ' + error);
      }

      console.log('Successful Jenkins transfer');
    });
  } catch (error) {
    next(error);
  }
};

const downloadService = async (req, res, next) => {
  try {
    let key = req.params.key;

    const controller = await models.Controller.findOne({
      where: {
        id: key
      }
    });

    let os = controller.device_os == 'ubuntu' ?  'ubuntuserver18' : controller.device_os;
    let folderPath = `${projectConfig.serviceInstallerPath}/controller/output/` + os;

    // let downloadFile = `invisily-controller-` + installerType[os] + `.zip`;
    let message = '';

    if (controller) {
      let filePath =
    `${folderPath}/invisily-controller-` + installerType[os] + '-' + controller.id + `.zip`;
      if (!fs.existsSync(filePath)) {
        message =
          'Unable to find installation package for controller component, please request for installer';
        return helper.sendEmailLinkErrorHtmlResponse(message, 404, res);
      }

      let packageName = await utils.getPackageName(os,'controller','',controller.id);

      res.download(filePath, packageName, function(err) {
        // file download is successful, expire link now
        if (!err) {
          controller
            .update({
              link_key_expiry: moment(),
              packageDownloaded: true
            })
            .then(function() {
              console.log(
                'Updated expiry of Controller [' +
                  controller.Name +
                  '] to now'
              );
            });
        }
      });

      insertIntoAdminLogs(
        getControllerDownloadlMessage(controller.Name),
        req
      )
    } else {
      message =
        'Unable to find installation package for controller component, please request for installer';
      helper.sendEmailLinkErrorHtmlResponse(message, 404, res);
    }
  } catch (error) {
    next(error);
  }
};

const generateService = async (req, res, next) => {
  try {
    const options = {
      where: {
        id: req.params.id
      },
      include: [
        {
          model: models.ComponentRole,
          as: 'Role',
          attributes: ['role']
        }
      ]
    };

    const controller = await models.Controller.findOne(options);


    if (!controller) {
      throw new httpErrors.NotFound('No controller found.');
    }

    const os = controller.device_os === 'ubuntu' ? 'ubuntuserver18' : controller.device_os;

    const link = await generateLink({
      os: os,
      component: 'controller',
      expiry_time: emailLinkExpiryInMinutes,
      filename: `invisily-controller-` + installerType[os] + '-' + controller.id
    });

    helper.sendEmail(
      prepareEmailOptions(
        controller.email,
        'Invisily Controller package download link',
        'download_link.html',
        `${process.env.CONTENT_URL}/api/v1/links/download/${link.key}`,
        'User',
        emailLinkExpiryInMinutes,
        'Controller'
      )
    );

    insertIntoAdminLogs(
      getControllerEmailMessage(controller.Name),
      req
    )

    res.json({
      status: true,
      message: `Check your email for Controller installer download link`
    });
  } catch (error) {
    next(error);
  }
};

async function removeControllerDBPermission(controller, transaction = null) {
  let statement =
    "DROP USER IF EXISTS '" +
    process.env.CONTROLLER_DB_USER +
    "'@'" +
    controller.ipv4_address +
    "';";
  return new Promise((resolve, reject) => {
    models.sequelize
      .query(statement, {
        raw: true,
        transaction
      })
      .then(user => {
        console.log('user deleted');
        console.log(user);
      })
      .then(() => {
        console.log('flush privileges');
        statement = 'FLUSH PRIVILEGES;';
        return models.sequelize.query(statement, {
          raw: true,
          transaction
        });
      })
      .then(() => {
        console.log('returning');
        resolve('done');
      })
      .catch(error => {
        reject('some error occurred while dropping DB permission');
      });
  });
}

async function addControllerDBPermission(controller, transaction) {
  let statement =
    "CREATE USER '" +
    process.env.CONTROLLER_DB_USER +
    "'@'" +
    controller.ipv4_address +
    "' IDENTIFIED BY '" +
    process.env.CONTROLLER_DB_PASSWORD +
    "' REQUIRE X509;";

  return new Promise((resolve, reject) => {
    models.sequelize
      .query(statement, {
        raw: true,
        transaction
      })
      .then(user => {
        console.log('creating user');
        console.log(user);
      })
      .then(() => {
        console.log('granting permission');
        statement =
          "GRANT ALL PRIVILEGES ON sdp.* TO '" +
          process.env.CONTROLLER_DB_USER +
          "'@'" +
          controller.ipv4_address +
          "';";
        return models.sequelize.query(statement, {
          raw: true
        });
      })
      .then(() => {
        console.log('showing grants');
        statement =
          "SHOW GRANTS FOR '" +
          process.env.CONTROLLER_DB_USER +
          "'@'" +
          controller.ipv4_address +
          "';";
        return models.sequelize.query(statement, {
          raw: true
        });
      })
      .then(() => {
        console.log('flush privileges');
        statement = 'FLUSH PRIVILEGES;';
        return models.sequelize.query(statement, {
          raw: true
        });
      })
      .then(() => {
        console.log('returning');
        resolve('done');
      })
      .catch(error => {
        reject('some error occurred while adding DB permission');
      });
  });
}

let prepareEmailOptions = (
  recipient,
  emailSubject,
  templateFileName,
  downloadLink,
  recipientName,
  linkExpiryInMinutes,
  component
) => {
  return {
    recipient: recipient,
    emailSubject: emailSubject,
    templateFileName: templateFileName,
    downloadLink: downloadLink,
    recipientName: recipientName,
    linkExpiryInMinutes: linkExpiryInMinutes,
    component: component
  };
};

let generateControllerCfgFile = (controller, ldap) => {
  let config = {
    ControllerIP: controller.ipv4_address,
    ControllerPort: controller.port,
    ServerPrivateKey: './certs/server.pem',
    ServerCertificate: './certs/server.crt',
    CACertificate: './certs/cacert.pem',
    ServerPrivateKeyPassword: 'SZNEe62D',
    DatabaseHost: process.env.CONTROLLER_DB_REMOTE_HOST,
    Database: process.env.CONTROLLER_DB,
    agentid: controller.controller_id,
    DatabaseCertFolder: './mysqlssl',
    networkInterface: 'networkid_value_placeholder',
    DNSIP: 'DNSIP_value_placeholder',
    DNSPort: 'DNSPort_value_placeholder',
    DNSProtocol: 'DNSProtocol_value_placeholder',
    ControllerRole: controller.Role ? controller.Role.role : controller.role,
    NTPIP: process.env.CONTENT_URL,
    contentPortal: process.env.CONTENT_URL,
    contentPortalPort: process.env.CONTENT_URL.split(':').pop(),
    RedisHost: process.env.REDIS_HOST || '127.0.0.1',
    RedisPort: process.env.REDIS_PORT || '6379',
    RedisPassword: process.env.REDIS_PASSWORD,
    DatabasePort: 3306,
    NTPPort: 123,
    NumWorkers: 5,
    // ActivityTimeout: 600,
    RabbitMQIP: process.env.RABBITMQ_HOST,
    RabbitMQPort: process.env.RABBITMQ_PORT,
    MfaServerPort: controller.mfa_server_port,
    OnboardingPort: controller.onboarding_port
  };

  // config.LDAPType = 'ebryx-ladp';
  // config.webIP = config.NTPIP = utils.getHostOrIpFromUrl(
  //   projectConfig.onPremApiUrl
    
  // );
  // config.webIP = config.NTPIP = config.contentPortal = utils.getHostOrIpFromUrl(
  //   config.NTPIP
  // );

  config.NTPIP = config.contentPortal = utils.getHostOrIpFromUrl(
    config.NTPIP
  );
  
  let result =
    '[DEFAULT]\n' +
    Object.keys(config).reduce(
      (acc, key) => `${acc}${key}=${config[key]}\n`,
      ''
    ) +
    'serviceMessagesFilePath=/opt/invisily/sdp-controller/controller/agent_files\n' +
    '[LOGGING]\n' +
    '#Level can be INFO or DEBUG or ERROR in capital letters\n' +
    'Level=INFO';
  console.log(result);
  return result;
};

let generateSnifferCfgFile = controller => {
  let config = {
    SPA_pkt_dest_port: controller.spa_port,
    SPA_pkt_length: 128,
    server_port1: '"' + controller.port + '"',
    server_port1_conn_time_expiry: 5,
    del_auth_file: 0,
    server_protocol: '"' + 'tcp' + '"',
    serviceMessagesFilePath: '"' + '../controller/agent_files/' + '"',
    networkInterface: '"' + 'networkid_value_placeholder' + '"',
    component_IPAddress: '"' + controller.ipv4_address + '"',
  };

  config = {
    ...config,
    server_port2: '"' + controller.mfa_server_port + '"',
    server_port3: '"' + controller.onboarding_port + '"',
    server_port2_conn_time_expiry: 180,
    server_port3_conn_time_expiry: 60
  }
  return Object.keys(config).reduce(
    (acc, key) => `${acc}${key}=${config[key]}\n`,
    ''
  );
};

let writeCfgFile = (controller, installerDir, resolve, reject) => {
  console.log('Found LDAP record in DB');
  fs.writeFileSync(
    `${installerDir}/ctrl.ini`,
    generateControllerCfgFile(controller)
  );
  console.log('controller configuration file written');
  fs.writeFileSync(
    `${installerDir}/sniffer_config.cfg`,
    generateSnifferCfgFile(controller)
  );
  console.log('sniffer configuration file written');
  resolve();
};

let zipInstaller = (
  controller,
  installerDir,
  packageDir,
  type,
  installerType,
  resolve,
  reject
) => {
  let output = fs.createWriteStream(
    `${packageDir}/invisily-controller-` + installerType + `-` + controller.id + `.zip`
  );
  let archive = archiver('zip', {
    zlib: {
      level: 9
    } // Sets the compression level.
  });
  output.on('close', function() {
    console.log('zip generated');
    resolve();
  });
  archive.on('error', function(err) {
    reject(err);
  });
  // pipe archive data to the file
  archive.pipe(output);
  archive.append(fs.createReadStream(`${installerDir}/ctrl.ini`), {
    name: 'ctrl.ini'
  });
  archive.append(fs.createReadStream(`${installerDir}/sniffer_config.cfg`), {
    name: 'sniffer_config.cfg'
  });
  archive.append(fs.createReadStream(`${installerDir}/installer.${type}`), {
    name: `installer.${type}`
  });
  archive.append(fs.createReadStream(`${installerDir}/rup.ec`), {
    name: `rup.ec`
  });
  archive.append(fs.createReadStream(`${installerDir}/rupk.txt`), {
    name: `rupk.txt`
  });
  archive.append(
    fs.createReadStream(`${installerDir}/install_invisily_controller.sh`),
    {
      name: `install_invisily_controller.sh`
    }
  );
  archive.append(
    fs.createReadStream(`${installerDir}/obp.zip`),
    {
      name: `obp.zip`
    }
  );
  archive.append(
    fs.createReadStream(`${installerDir}/mfa.zip`),
    {
      name: `mfa.zip`
    }
  );
  archive.append(
    fs.createReadStream(`${installerDir}/Installation_Guide_Controller.txt`),
    {
      name: `Installation_Guide_Controller.txt`
    }
  );
  archive.directory(`${installerDir}/certs`, 'certs');
  archive.directory(`${installerDir}/mysqlssl`, 'mysqlssl');
  archive.finalize();
};

function generateInstallerZip(controller, packageState, isNew = true) {
  return setExpiryTime().then(function() {
    controller.packageDownloaded = false;
    controller.installer_status = 'processing';
    if (packageState === '1') {
      controller.isDownloadPackage = true;
    } else {
      controller.isDownloadPackage = false;
    }
    controller.save();
    let os =  controller.device_os === 'ubuntu' ? 'ubuntuserver18' : controller.device_os;
    let packageDir =
      `${projectConfig.serviceInstallerPath}/controller/output/` + os;
    let key = `${controller.updatedAt.getTime()}:` + os;
    let installerDir =
      `${projectConfig.serviceInstallerPath}/controller/input/` + os;
    let certDirectory = `${installerDir}/certs`;
    let certSqlDirectory = `${installerDir}/mysqlssl`;
    let mysqlCertPath = '/var/local/mysqlssl/';

    console.log('installerDir ...' + installerDir);
    console.log('packageDir ...' + packageDir);
    console.log('key ...' + key);
    console.log('certDirectory ...' + certDirectory);
    console.log('certSqlDirectory ...' + certSqlDirectory);

    helper.mkDir(certDirectory);
    helper.mkDir(packageDir);
    helper.mkDir(installerDir);
    helper.mkDir(certSqlDirectory);

    // asynchronous chaining
    return new Promise((resolve, reject) => {
      // generate certificates
      // this will also generate the certificates in the given folder
      exec(
        `sudo ${projectConfig.serviceInstallerPath}/generate-ssl-certs.sh ${key} ${certDirectory}`,
        (error, stdout, stderr) => {
          if (error) {
            reject(error);
          } else {
            // generate copy of client certificates and name them as server certificates
            fsExtra.copy(
              certDirectory + '/client.crt',
              certDirectory + '/server.crt',
              err => {
                if (err) throw err;
                console.log('client.crt was copied to server.crt');

                fsExtra.copy(
                  certDirectory + '/client.pem',
                  certDirectory + '/server.pem',
                  err => {
                    if (err) throw err;
                    console.log('client.pem was copied to server.pem');
                    fsExtra.copy(
                      mysqlCertPath + 'ca.pem',
                      certSqlDirectory + '/ca.pem',
                      err => {
                        if (err) throw err;
                        console.log('ca.pem was copied to mysqlssl ca.pem');
                        fsExtra.copy(
                          mysqlCertPath + 'client-cert.pem',
                          certSqlDirectory + '/client-cert.pem',
                          err => {
                            if (err) throw err;
                            console.log(
                              'client-cert.pem was copied to mysqlssl client-cert.pem'
                            );
                            fsExtra.copy(
                              mysqlCertPath + 'client-key.pem',
                              certSqlDirectory + '/client-key.pem',
                              err => {
                                if (err) throw err;
                                console.log(
                                  'client-key.pem was copied to mysqlssl client-key.pem'
                                );
                                resolve();
                              }
                            );
                          }
                        );
                      }
                    );
                  }
                );
              }
            );
          }
        }
      );
    })
      .then(
        () =>
          new Promise((resolve, reject) => {
            console.log('Generating encrypted files');
            exec(
              `${scriptsPath}/ctrlr_rup.sh ${process.env.CONTROLLER_DB_USER} ${process.env.CONTROLLER_DB_PASSWORD}`,
              (error, stdout, stderr) => {
                if (error) {
                  return reject(error);
                }
                Promise.all([
                  copyFilePromise('/tmp/ctrl_rup/rup.ec', installerDir + '/rup.ec'),
                  copyFilePromise('/tmp/ctrl_rup/rupk.txt', installerDir + '/rupk.txt')
                ])
                  .then(resolve)
                  .catch(reject)
              }
            )
          })
      )
      .then(
        () =>
          new Promise((resolve, reject) => {
            console.log('Generating configuration file');
            writeCfgFile(controller, installerDir, resolve, reject);
          })
      )
      .then(
        () =>
          new Promise((resolve, reject) => {
            console.log('Generating Zip file');
            zipInstaller(
              controller,
              installerDir,
              packageDir,
              mapping[os],
              installerType[os],
              resolve,
              reject
            );
          })
      )
      .then(() => {
        // save file in database
        console.log('Saving controller certificate files');
        controller.cacert = fs.readFileSync(`${certDirectory}/cacert.pem`);
        controller.clientcrt = fs.readFileSync(`${certDirectory}/client.crt`);
        controller.clientpem = fs.readFileSync(`${certDirectory}/client.pem`);
        controller.link_key = key;
        controller.link_key_expiry = moment().add(
          emailLinkExpiryInMinutes,
          'minutes'
        );
        controller.onPrenDownloadLink =
          `${projectConfig.onPremApiUrl}/controllers/download/service/` +
          new Buffer(key).toString('base64');
        controller.installer_status = 'success';
        return controller.save();
      })
      .then(async () => {
        const fileName = `invisily-controller-${installerType[os]}-${controller.id}`

        require('../queues/uploadToContent').addtoQueue({
          os: os,
          component: 'controller',
          folderPath: packageDir + '/',
          filename: fileName
        })
  
        const link = await generateLink({
          os,
          component: 'controller',
          expiry_time: 60,
          filename: fileName
        });

        const downloadLink = `${process.env.CONTENT_URL}/api/v1/links/download/${link.key}`
        
        controller.pkg_url = downloadLink
        await controller.save()

        const message = {
          updates: [
            {
              URL: downloadLink,
              Agent_ID: controller.controller_id
            }
          ],
          is_configuration_changed: !isNew
        }

        const softwareUpdate = await models.Update.findOne({
          where: { buildComponent: 'controller', buildOs: os, isDeleted: false }
        })

        if(softwareUpdate) {
          message['version'] = softwareUpdate.buildVersion
        }

        await publishControllerSoftware(actionType.UPDATED, message);
        
        return;
      })
      .catch(err => {
        console.log(err && (err.message || err));
        controller.installer_status = 'error';
        return controller.save();
      });
  });
}

const getControllerHealthStats = async (req, res, next) => {
  try {
    let { interval = 1 } = req.query;
    const { controllerId } = req.params;

    const allowedIntervals = [1, 5, 30]
    if(!allowedIntervals.includes(parseInt(interval))) {
      throw new httpErrors.BadRequest("Invalid interval value")
    }

    const controller = await models.Controller.findOne({
      id: controllerId
    })

    if(!controller) {
      throw new httpErrors.BadRequest("Invalid controller id");
    }

    function mapData(data, keys = []) {
      return data.map(row => {
        let response = {
          time: row.createdAt,
        }
        keys && keys.forEach(key => {
          if(row[key] || row[key] === 0) {
            response[key] = row[key]
          }
        })
        return response;
      })
    }

    const minDate = moment().subtract(interval, 'days').format()
    const maxDate = moment().format()

    const groupBy = (interval === 1) ? `group by createdAt`: `group by day(createdAt)`

    let query = `
        select createdAt,
        AVG(memTotal) / 1024 as memTotal, 
        AVG(memFree) / 1024 as memFree , 
        AVG(storageFree) / 1024 as storageFree , 
        AVG(storageTotal) / 1024 as storageTotal ,
        AVG(cpuUtilization) as cpuUtilization ,
        AVG(mainProcessCPUUtilization) as mainProcessCPUUtilization , 
        AVG(snifferProcessCPUUtilization) as snifferProcessCPUUtilization 
        from ControllerRuntimeHealths crh 
        where crh.controllerId = ${controllerId} AND 
        crh.createdAt >= '${minDate}' AND 
        crh.createdAt <= '${maxDate}'
        ${groupBy}, createdAt
      `


    const [result, controllerHealth] = await Promise.all([
      models.sequelize.query(query),
      models.ControllerHealth.findOne({
        where: {
          controllerId
        }
      })
    ])

    res.status(200).send({
      lastUpdated: controllerHealth && controllerHealth.updatedAt,
      os: controllerHealth && controllerHealth.OS,
      version: controllerHealth && controllerHealth.componentVersion,
      cpuDetail: controllerHealth && controllerHealth.CPU,
      lastErrorLog: controllerHealth && controllerHealth.lastErrorLog,
      lastSnifferErrorLog: controllerHealth && controllerHealth.snifferLastErrorLog,
      memory: mapData(result[0], ['memFree', 'memTotal']),
      storage: mapData(result[0], ['storageTotal', 'storageFree']),
      cpu: mapData(result[0], ['cpuUtilization', 'snifferProcessCPUUtilization', 'mainProcessCPUUtilization']),
    });
  } catch(e) {
    next(e);
  }
}




module.exports = {
  getControllers,
  getController,
  status,
  deleteController,
  updateController,
  downloadUpgrade,
  downloadService,
  generateService,
  addController,
  controllerZipGenerator: generateInstallerZip,
  getControllerHealthStats
};
