var httpErrors = require('http-errors');
var models = require('../models');
var fs = require('fs');
var fsExtra = require('fs-extra');
var csvWriter = require('csv-writer').createObjectCsvWriter;
var sanitize = require('sanitize-filename');
const csv = require('fast-csv');
const {
  promisify
} = require('util');
const readFile = promisify(fs.readFile);
const writeFile = promisify(fs.writeFile);


const getDevices = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchTerm = ''
    } = req.query;

    let options = {
      limit: limit,
      offset: offset,
      order: []
    };

    if (searchTerm) {
      options.where = {
        [Op.or]: {
          userName: {
            [Op.like]: `%${searchTerm.trim()}%`
          },
          deviceSerialNumber: {
            [Op.like]: `%${searchTerm.trim()}%`
          },
          deviceTag: {
            [Op.like]: `%${searchTerm.trim()}%`
          }
        }
      }

    }

    const allowedSortValues = [
      'userName',
      'id'
    ];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    const result = await models.DeviceInventories.paginate(options);
    res.json(result);

  } catch (error) {
    next(error);
  }
}


const deleteDevice = async (req, res, next) => {
  try {
    await models.DeviceInventories.destroy({
      where: {
        id: req.params.id
      }
    });
    res.json('Device inventory deleted successfully');
  } catch (error) {
    next(error);
  }
}


const updateDevice = async (req, res, next) => {
  try {
    const id = req.params.id;
    const _inventory = req.body.inventory;

    const devices = await models.DeviceInventories.findAndCountAll({
      where: {
        $and: {
          id: {
            $ne: id
          },
          [Op.or]: [{
              deviceSerialNumber: _inventory.deviceSerialNumber
            },
            {
              deviceTag: _inventory.deviceTag
            }
          ]
        }
      }
    });

    if (devices.count > 0) {
      throw new httpErrors.Conflict('Device Serial Number or Device Tag already exist.');
    }

    const inventory = await models.DeviceInventories.findOne({
      where: {
        id: id
      }
    });


    if (!inventory) {
      throw new httpErrors.NotFound('No device to update.');
    }

    await inventory.update(_inventory);

    res.json(inventory);

  } catch (error) {
    next(error);
  }
}


const addDevice = async (req, res, next) => {
  try {
    const {
      inventory
    } = req.body;

    const options = {
      where: {
        [Op.or]: [{
            deviceSerialNumber: inventory.deviceSerialNumber
          },
          {
            deviceTag: inventory.deviceTag
          }
        ]
      }
    };

    const devices = await models.DeviceInventories.findAndCountAll(options);

    if (devices.count > 0) {

      throw new httpErrors.Conflict('Device Serial Number or Device Tag already exist.');

    }

    const addedInventory = await models.DeviceInventories.create(inventory);

    res.json(addedInventory);

  } catch (error) {
    next(error);
  }
}


const uploadCsv = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  
  let dataArray = [];
  let file = req.body.fileData;
  let encodedFileData = file.data.split(';base64,').pop();
  let fileName = sanitize(file.name);
  const filePath = '/var/www/html/admin-gui/storage/deviceInventories/' + fileName;

  try {

    await writeFile(filePath, encodedFileData, 'base64');

    await new Promise((resolve, reject) => {
      const namePattern = RegExp(/^[a-zA-Z0-9 .]*$/);
      const deviceSerialNumberPattern = RegExp(/^[0-9a-zA-Z-]{1,64}$/);
      const deviceTagPattern = RegExp(/^[0-9a-zA-Z-]{1,199}$/);
      const unicodeTextExpression = /[^\u0000-\u00ff]/;
      const csvStream = fs.createReadStream(filePath);

      const hashMap = {
        deviceSerialNumber: {},
        deviceTag: {}
      };

      csvStream.pipe(csv.parse({
          headers: false
        })).validate(function(data, cb) {

          if (!(data.length === 3 
              && data[0].length <= 30 
              && data[1].length <= 64
              && data[2].length <= 199
              && namePattern.test(data[0])
              && deviceSerialNumberPattern.test(data[1]) 
              && deviceTagPattern.test(data[2])
            )
          ) {
            return cb(new httpErrors.UnprocessableEntity("Invalid file format."))
          }

          if (unicodeTextExpression.test(data[1]) || unicodeTextExpression.test(data[0]) || unicodeTextExpression.test(data[2])) {
            return cb(new httpErrors.UnprocessableEntity("File contains unicoded character(s)."))
          }


          if (!hashMap.deviceSerialNumber[data[1].trim()]) {
            hashMap.deviceSerialNumber[data[1].trim()] = 1;
          } else {
            return cb(new httpErrors.Conflict("The file contains duplicate Device Serial Number(s)."));
          }

          if(data[0].trim() == ''){
            return cb(new httpErrors.Conflict("The file contains empty Username field(s)."));
          }

          if(data[1].trim() == ''){
            return cb(new httpErrors.Conflict("The file contains empty Device Serial Number field(s)."));
          }

          if(data[2].trim() == ''){
            return cb(new httpErrors.Conflict("The file contains empty Device Tag field(s)."));
          }

          if (!hashMap.deviceTag[data[2].trim()]) {
            hashMap.deviceTag[data[2].trim()] = 1;
          } else {
            return cb(new httpErrors.Conflict("The file contains duplicate device tag(s)."))
          }

          return cb(null, data);
        })
        .on("data", function(data) {
          dataArray.push({
            userName: data[0].trim(),
            deviceSerialNumber: data[1].trim(),
            deviceTag: data[2] ? data[2].trim() : null
          })
        })
        .on("end", function() {
          return resolve();
        })
        .on("error", function(error) {
          reject(error);
        })
    });

    await Promise.all([
      models.DeviceInventories.destroy({
        where: {},
        transaction
      }),
      models.DeviceInventories.bulkCreate(dataArray, {
        transaction
      })
    ]);

    await transaction.commit();

    await fsExtra.remove(filePath);

    res.json('file written finishes locally');

  } catch (error) {
    await transaction.rollback();
    fsExtra.remove(filePath);
    next(error);
  }
}


const downloadCsv = async (req, res, next) => {
  try {

    const filePath = '/var/www/html/admin-gui/storage/deviceInventories/device-inventory.csv';
    const deviceInventories = await models.DeviceInventories.findAll();

    const csvFile = csvWriter({
      path: filePath,
      header: [{
          id: 'userName',
          title: ''
        },
        {
          id: 'deviceSerialNumber',
          title: ''
        },
        {
          id: 'deviceTag',
          title: ''
        }
      ]
    });

    const deviceInventoriesArray = [];
    for (deiviceInventory in deviceInventories) {
      const {
        userName,
        deviceSerialNumber,
        deviceTag
      } = deviceInventories[
        deiviceInventory
      ];
      deviceInventoriesArray.push({
        userName: userName,
        deviceSerialNumber: deviceSerialNumber,
        deviceTag: deviceTag
      });
    }
    await csvFile.writeRecords(deviceInventoriesArray);
    const fileData = await readFile(filePath, 'utf8');

    var linesExceptFirst = fileData.split('\n').slice(1).join('\n');

    await writeFile(filePath, linesExceptFirst);

    res.download(filePath);
    fsExtra.remove(filePath)

  } catch (error) {
    fsExtra.remove(filePath);
    next(error);
  }
}


module.exports = {
  downloadCsv,
  uploadCsv,
  addDevice,
  updateDevice,
  deleteDevice,
  getDevices
};