const router = require('express').Router();
var httpErrors = require('http-errors');
const moment = require('moment');
const crypto = require('crypto');
var models = require('../models');
const LdapManager = require('../routes/common/ldap-manager');
let helper = require('../routes/common/helper');
const resetPasswordQueue = require('../queues/password-reset');
const { generateRandomPassword } = require('../utils/common');
const csv = require('fast-csv');
const { v4: uuidv4 } = require('uuid');
const BulkPasswordChangeQueue = require('../queues/local-user-bulk-password-change');
const Openssl = require('../openssl/openssl.js');
const OpensslObj = new Openssl();
const settings = require('../routes/common/settings');
var projectConfig = require('../config/config')['project'][
  process.env.NODE_ENV
];
const {
  actionType,
  publishLogOut,
  publishClients,
  publishMsgFile,
  clientColumnsToPublish,
  publishUserGroups
} = require('../routes/common/event-publisher');
const { Op } = require('sequelize');
const {
  sendOtpQrBulkUsers,
  installerDescriptiveName,
  generateZip
} = require('./clients');
var { generateLink } = require('../routes/common/content-upload');
const {
  getLocalUserCreateMessage,
  getLocalUserEditMessage,
  getLocalUserDeleteMessage,
  getDevicetoLocalUserDeleted,
  getDevicetoLocalUserAdded,
  getLocalUserSendOtp,
  getLocalUserPassChangeMessage,
  getLocalUserMfaMessage
} = require('../config/adminLogs.config');

const { insertIntoAdminLogs } = require('../routes/common/adminLogs');

var emailLinkExpiryInMinutes = 60;
const argon2 = require('argon2');

const checkValidToken = async (req, res, next) => {
  try {
    const { token } = req.params;

    const localUser = await models.LocalUsers.findOne({
      where: {
        link_key: token
      }
    });

    if (localUser.is_used) {
      throw new httpErrors.Unauthorized(
        'Reset password link is already used, please request admin for a new link.'
      );
    }

    if (!moment().isBefore(moment(localUser.link_expiry))) {
      throw new httpErrors.Unauthorized(
        'The link has expired. Please contact Invisily Admin for a new link.'
      );
    }

    res.json({
      tokenValid: true,
      status: true
    });
  } catch (error) {
    next(error);
  }
};

const token = async (req, res, next) => {
  try {
    const userData = await models.LocalUsers.findOne({
      where: {
        link_key: req.params.token
      },
      attributes: ['id', 'userName', 'email', 'name']
    });

    res.json(userData);
  } catch (error) {
    next(error);
  }
};

const updatePassword = async (req, res, next) => {
  try {
    const { password } = req.body;

    const hash = crypto
      .createHash('sha256')
      .update(password)
      .digest('hex');

    await models.LocalUsers.update(
      {
        password: hash,
        is_used: 1
      },
      {
        where: {
          link_key: req.params.token
        }
      }
    );

    res.json('Password reset successful');
  } catch (error) {
    next(error);
  }
};

const getAvailableLocalUsers = async (req, res, next) => {
  try {
    const { departmentId } = req.params;

    const options = {
      attributes: ['id', 'name'],
      where: {
        department: departmentId
      }
    };
    const result = await models.LocalUsers.findAll(options);

    res.status(200).json({
      data: result
    });
  } catch (e) {
    next(e);
  }
};

const getLocalUsers = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchTerm = '',
      identityProvider = '',
      organization = '',
      department = ''
    } = req.query;

    const options = {
      limit: +limit,
      offset: +offset,
      order: [],
      include: [
        {
          model: models.Departments,
          attributes: ['id', 'name'],
          include: [
            {
              model: models.Organizations,
              attributes: ['name']
            }
          ]
        }
      ]
    };

    if (searchTerm.trim()) {
      options.where = {
        [Op.or]: {
          name: {
            [Op.like]: `%${searchTerm.trim()}%`
          },
          email: {
            [Op.like]: `%${searchTerm.trim()}%`
          },
          userName: {
            [Op.like]: `%${searchTerm.trim()}%`
          }
        }
      };
    }

    if (identityProvider !== '') {
      options.where = {
        ...options.where,
        identity_provider: {
          [Op.like]: `%${identityProvider}%`
        }
      };
    }

    if (organization !== '') {
      const organizationDoc = await models.Organizations.findOne({
        where: {
          name: organization
        }
      });
      if (!organizationDoc) {
        throw new httpErrors.NotFound('Organization not found.');
      }
      const departments = await models.Departments.findAll({
        where: {
          organization_id: organizationDoc.id
        }
      });
      if (!departments) {
        throw new httpErrors.NotFound(
          'No departments associated with the Organization.'
        );
      }
      let deptFiltering = [];
      departments.map(department => {
        deptFiltering.push(department?.toJSON());
      });
      options.where = {
        ...options.where,
        department: {
          [Op.in]: deptFiltering.map(department => department.id)
        }
      };
    }

    if (department !== '') {
      options.where = {
        ...options.where,
        department: {
          [Op.eq]: department
        }
      };
    }

    const allowedSortValues = ['name', 'id'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    let result = await models.LocalUsers.paginate(options);

    const formattedData = result.data.map(user => {
      const userData = user.toJSON();
      const departmentName = user.Department
        ? user.Department.toJSON().name
        : null;
      const organizationName =
        user.Department && user.Department.Organization
          ? user.Department.Organization.toJSON().name
          : null;

      return {
        ...userData,
        departmentName,
        organizationName
      };
    });

    const setting = await models.Settings.findOne({
      attributes: ['is_ldap_syncing', 'biometrics_state'],
      order: [['id', 'DESC']],
      limit: 1
    });

    result = {
      ...result,
      isLdapSyncing: setting.is_ldap_syncing,
      data: formattedData
    };

    res.json(result);
  } catch (error) {
    next(error);
  }
};

const getLocalUsersFromDropDown = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchTerm = '',
      identityProvider = '',
      organization = '',
      department = '',
      allUser
    } = req.query;

    // Validate sort value
    const validSortColumns = ['id', 'name']; // Add other valid columns as needed
    if (!validSortColumns.includes(sort)) {
      sort = 'id'; // Default value
    }

    const whereClauses = [
      `(LocalUsers.name LIKE :searchTerm OR LocalUsers.email LIKE :searchTerm OR LocalUsers.userName LIKE :searchTerm)`,
      `LocalUsers.identity_provider LIKE :identityProvider`
    ];

    if (department) {
      whereClauses.push(`LocalUsers.department = :department`);
    }

    if (organization) {
      whereClauses.push(`Organizations.name = :organization`);
    }

    const whereString = whereClauses.join(' AND ');

    let query = `
      SELECT
          LocalUsers.*,
          Departments.name AS departmentName,
          Organizations.name AS organizationName
      FROM LocalUsers
      LEFT JOIN Departments ON LocalUsers.department = Departments.id
      LEFT JOIN Organizations ON Departments.organization_id = Organizations.id
      WHERE ${whereString}
      ORDER BY LocalUsers.${sort} ${
      ['asc', 'desc'].includes(dir) ? dir : 'desc'
    }`;

    if (!allUser) {
      query += ' LIMIT :offset, :limit';
    }

    const replacements = {
      searchTerm: `%${searchTerm}%`,
      identityProvider: `%${identityProvider}%`,
      department: department,
      organization: organization,
      offset: allUser ? '' : Number(offset),
      limit: allUser ? '' : Number(limit)
    };

    console.log('query ', query);

    const result = await models.sequelize.query(query, {
      replacements: replacements,
      type: models.sequelize.QueryTypes.SELECT
    });

    const setting = await models.Settings.findOne({
      attributes: ['is_ldap_syncing', 'biometrics_state'],
      order: [['id', 'DESC']],
      limit: 1
    });

    const finalResult = {
      data: [...result],
      isLdapSyncing: setting.is_ldap_syncing,
      recordsTotal: result.length
    };

    res.json(finalResult);
  } catch (error) {
    next(error);
  }
};

const getLocalUsersSearch = async (req, res, next) => {
  try {
    const { limit = 10, offset = 0, searchTerm = '', groupId, orderBy = 'asc' } = req.query;

    const validOrderBy = ['asc', 'desc'].includes(orderBy.toLowerCase()) ? orderBy : 'asc';

    let excludedUserIds = [];
    if (groupId) {
      const excludedUsersQuery = `
        SELECT localUsers.id 
        FROM UserGroupsPivot userGroupsPivot
        LEFT JOIN CustomUserGroups customUserGroups ON customUserGroups.id = userGroupsPivot.group_id 
        LEFT JOIN LocalUsers localUsers ON localUsers.id = userGroupsPivot.user_id
        WHERE customUserGroups.id = :groupId
      `;
      const groupData = await models.sequelize.query(excludedUsersQuery, {
        replacements: { groupId },
        type: models.sequelize.QueryTypes.SELECT,
      });
      excludedUserIds = groupData.map((group) => group.id);
    }

    const conditions = [];
    const replacements = {};

    // Process search terms
    searchTerm.split(' And ').forEach((condition) => {
      if (condition.startsWith('Identity Provider = ')) {
        conditions.push(`localUsers.identity_provider = :identityProvider`);
        replacements.identityProvider = decodeURIComponent(condition.split('Identity Provider = ')[1]);
      } else if (condition.startsWith('Organization = ')) {
        conditions.push(`organizations.name = :organization`);
        replacements.organization = decodeURIComponent(condition.split('Organization = ')[1]);
      } else if (condition.startsWith('Department = ')) {
        conditions.push(`departments.name = :department`);
        replacements.department = decodeURIComponent(condition.split('Department = ')[1]);
      } else if (condition.startsWith('Username = ')) {
        conditions.push(`localUsers.userName = :username`);
        replacements.username = decodeURIComponent(condition.split('Username = ')[1]);
      } else if (condition.startsWith('Name = ')) {
        conditions.push(`localUsers.name = :name`);
        replacements.name = decodeURIComponent(condition.split('Name = ')[1]);
      } else if (condition.startsWith('Email = ')) {
        conditions.push(`localUsers.email = :email`);
        replacements.email = decodeURIComponent(condition.split('Email = ')[1]);
      }
    });

    if (excludedUserIds.length > 0) {
      conditions.push(`localUsers.id NOT IN (:excludedUserIds)`);
      replacements.excludedUserIds = excludedUserIds;
    }

    const whereClause = conditions.length > 0 ? `WHERE ${conditions.join(' AND ')}` : '';

    // Main query for user data
    const query = `
      SELECT localUsers.*, 
             departments.id AS departmentId, departments.name AS departmentName, 
             organizations.id AS organizationId, organizations.name AS organizationName
      FROM LocalUsers localUsers
      LEFT JOIN Departments departments ON localUsers.department = departments.id
      LEFT JOIN Organizations organizations ON departments.organization_id = organizations.id
      ${whereClause}
      ORDER BY localUsers.userName ${validOrderBy}
      LIMIT :limit OFFSET :offset;
    `;

    replacements.limit = +limit;
    replacements.offset = +offset * +limit;

    const result = await models.sequelize.query(query, {
      replacements,
      type: models.sequelize.QueryTypes.SELECT,
    });

    // Count for filtered users
    const userCountQuery = `
      SELECT COUNT(localUsers.id) AS local_users_count
      FROM LocalUsers localUsers
      LEFT JOIN Departments departments ON localUsers.department = departments.id
      LEFT JOIN Organizations organizations ON departments.organization_id = organizations.id
      ${whereClause};
    `;
    const userCountResult = await models.sequelize.query(userCountQuery, {
      replacements,
      type: models.sequelize.QueryTypes.SELECT,
    });

    const userCount = userCountResult[0]?.local_users_count || 0;

    // Count for total users
    const totalCountQuery = `
      SELECT COUNT(localUsers.id) AS total_count
      FROM LocalUsers localUsers
      LEFT JOIN Departments departments ON localUsers.department = departments.id
      LEFT JOIN Organizations organizations ON departments.organization_id = organizations.id;
    `;
    const totalCountResult = await models.sequelize.query(totalCountQuery, {
      type: models.sequelize.QueryTypes.SELECT,
    });

    const totalCount = totalCountResult[0]?.total_count || 0;

    // Fetch settings
    const setting = await models.Settings.findOne({
      attributes: ['is_ldap_syncing', 'biometrics_state'],
      order: [['id', 'DESC']],
      limit: 1,
    });

    res.json({
      data: result,
      isLdapSyncing: setting?.is_ldap_syncing || false,
      recordsTotal: searchTerm || excludedUserIds.length > 0 ? userCount : totalCount,
    });
  } catch (error) {
    next(error);
  }
};


const getUsersAndUserGroups = async (req, res, next) => {
  try {
    const localUsersInGroupQuery = `
      SELECT DISTINCT localUsers.id as id, localUsers.name as name, 'user' as type, localUsers.runtime_bind_id 
      FROM LocalUsers localUsers
      LEFT JOIN Departments departments on localUsers.department = departments.id
      LEFT JOIN Organizations organizations ON departments.organization_id = organizations.id
      WHERE  NOT (localUsers.userName LIKE '%gwadmin%')`;

    const localUsersInGroup = (
      await models.sequelize.query(localUsersInGroupQuery)
    )[0];

    const usersGroupQuery = `
      SELECT customUserGroups.id as id, customUserGroups.group_name as name, 'userGroup' as type, 
      customUserGroups.runtime_bind_id, customUserGroups.raw_query_id  
      FROM CustomUserGroups customUserGroups`;

    const usersGroup = (await models.sequelize.query(usersGroupQuery))[0];

    let idCounter = 1;
    let result = [...localUsersInGroup, ...usersGroup];

    res.json({
      data: result
    });
  } catch (error) {
    next(error);
  }
};

const getLocalUser = async (req, res, next) => {
  try {
    const options = {
      where: {
        id: req.params.id
      },
      attributes: [
        'id',
        'userName',
        'identity_provider',
        'name',
        'is_strong_mfa',
        'biometric_state',
        'email',
        'department',
        'client_mfa',
        'contact',
        'reachable_email'
      ]
    };

    const userData = await models.LocalUsers.findOne(options);

    const clientSetting = await models.ClientSetting.findOne({});
    const response = {
      userData,
      global_darkening: clientSetting.global_darkening
    };

    res.json(response);
  } catch (error) {
    next(error);
  }
};

const getLocalUserServices = async (req, res, next) => {
  try {
    let client = await models.Client.findOne({
      where: {
        user_id: req.params.id
      }
    });

    let localUser = await models.LocalUsers.findOne({
      where: {
        id: req.params.id
      }
    });

    if (!localUser) {
      throw httpErrors.NotFound('Client or User not found.');
    }

    // Code releated to Entitlements (Users/Clients)
    let query = `CALL GetEntitlementsForUser(${localUser.dataValues.id})`;

    let results = await models.sequelize.query(query);

    let entitlementOrganizedData = {};
    results.map(value => {
      if (!entitlementOrganizedData[value?.policy_id]) {
        entitlementOrganizedData[value?.policy_id] = {
          policy_id: value?.policy_id,
          name: value?.entitlement_name,
          timeRestricted: value?.time_restricted,
          services: []
        };

        let services = results
          .filter(obj => obj.policy_id === value.policy_id)
          .map(obj => obj.name);

        entitlementOrganizedData[value?.policy_id].services = services.join(
          ', '
        );
      }
    });

    let entitlementData = Object.values(entitlementOrganizedData);

    // Code releated to Services (Users/Clients)
    let all_services = false;

    for (let data of results) {
      if (data.all_services) {
        all_services = true;
      }
    }

    let serviceQuery, serviceResults;
    if (all_services) {
      serviceQuery = `
        SELECT services.id, services.Name as name, services.ipv4_address, services.type, services.mappedurl
        FROM Services services
      `;
      serviceResults = (await models.sequelize.query(serviceQuery))[0];
    } else {
      serviceQuery = `CALL GetServicesForUser(${localUser.dataValues.id})`;
      serviceResults = await models.sequelize.query(serviceQuery);
    }

    // Code releated to Service Groups (Users/Clients)
    let serviceGroupQuery = `CALL GetServiceGroupsForUser(${localUser.dataValues.id})`;

    let serviceGroupResults = await models.sequelize.query(serviceGroupQuery);

    let groupOrganizedData = {};
    serviceGroupResults.map((value, index) => {
      if (!groupOrganizedData[value?.id]) {
        groupOrganizedData[value?.id] = {
          id: value.id,
          name: value.name,
          createdAt: value.created_at,
          description: value.description,
          services: []
        };

        let services = serviceGroupResults
          .filter(obj => obj.id === value.id)
          .map(obj => obj.service_name);

        groupOrganizedData[value?.id].services = services.join(', ');
      }
    });

    let userServicesGroup = Object.values(groupOrganizedData);

    res.json({
      entitlements: entitlementData,
      userServices: serviceResults,
      userServicesGroup: userServicesGroup,
      client: client?.dataValues
    });
  } catch (error) {
    next(error);
  }
};

const resetPassword = (req, res, next) => {
  try {
    const token = req.query.token;
    res.redirect(`/web/localUserResetPassword/${token}`);
  } catch (error) {
    next(error);
  }
};

const sendResetLink = async (req, res, next) => {
  try {
    const { role, id: userId } = req.user;

    const { id, email } = req.body;

    let localUser;
    if (id) {
      localUser = await models.LocalUsers.findOne({
        where: {
          id,
          userName: {
            [Op.notLike]: '%gwadmin%'
          }
        }
      });

      insertIntoAdminLogs(getLocalUserPassChangeMessage(localUser.name), req);
    } else if (email) {
      localUser = await models.LocalUsers.findOne({
        where: {
          email,
          userName: {
            [Op.notLike]: '%gwadmin%'
          }
        }
      });
    }

    if (!localUser) {
      throw httpErrors.NotFound('User not found.');
    }

    const clients = await models.Client.findAll({
      attributes: ['id'],
      where: {
        user: localUser.userName
      },
      raw: true
    });

    await Promise.all([
      localUser.update({
        change_pass: true,
        password_updated_at: moment()
      }),

      models.Client.update(
        {
          is_active: false
        },
        {
          where: {
            client_type: 'Server',
            user_id: localUser.id,
            device_id: null
          }
        }
      ),

      clients.length
        ? publishLogOut(actionType.UPDATED, {
            code: 63,
            ids: clients.map(c => c.id)
          })
        : []
    ]);

    resetPasswordQueue.addtoQueue({
      user: localUser.userName
    });

    res.json('Email sent to client.');
  } catch (error) {
    next(error);
  }
};

const makeLDAPChanges = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    await models.sequelize.query(
      `UPDATE Clients c
      INNER JOIN LocalUsers u ON c.Email = u.email
      SET
        c.user_id = u.id,
        c.device_id = c.id + 1000
      WHERE c.Email IS NOT NULL AND c.user_id IS NULL`,
      {
        transaction
      }
    );
    await models.sequelize.query(
      `INSERT INTO DeviceInventories (
          userName, createdAt, updatedAt,
          device_name, user_id, has_agent,
          darkening_state, os, os_version,
          risk_score, geoip_track, id
        )
        SELECT
          \`user\`, UTC_TIMESTAMP(), UTC_TIMESTAMP(),
          Name, user_id, 1,
          client_darkening, device_os, client_os_version,
          risk_score, geoip_track, device_id
        FROM Clients
        WHERE Name NOT LIKE '%}@{%' AND Email IS NOT NULL`,
      {
        transaction
      }
    );
    await models.sequelize.query(
      `UPDATE Clients c
      INNER JOIN DeviceInventories d ON c.Name = d.device_name
      SET c.device_id = d.id
      WHERE c.Name NOT LIKE '%}@{%' AND c.Email IS NOT NULL`,
      {
        transaction
      }
    );
    await models.sequelize.query(
      `UPDATE Clients c
      INNER JOIN LocalUsers u ON c.user_id = u.id
      INNER JOIN DeviceInventories d ON c.device_id = d.id
      SET c.Name = CONCAT("{", u.name, "}@{", d.device_name, "}")
      WHERE c.Name NOT LIKE '%}@{%'`,
      {
        transaction
      }
    );
    await transaction.commit();
    next();
  } catch (error) {
    await transaction.rollback();
    next();
  }
};

const createLocalUser = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    let {
      name,
      email,
      userName,
      department = '',
      client_mfa = false,
      is_strong_mfa = false,
      biometric_state = false,
      contact = '',
      identity_provider,
      reachable_email,
    } = req.body.localUser;

    if (department != '') {
      department = department.deptId;
    }

    const isLdapUserFound = await checkUserExistInLdapOrNot(userName, email);

    if (isLdapUserFound) {
      throw new httpErrors.Conflict('User Already exists in LDAP.');
    }

    const localUser = await models.LocalUsers.findOne({
      where: {
        $or: [
          {
            userName: {
              $eq: userName
            }
          },
          {
            email: {
              $eq: email
            }
          }
        ]
      }
    });

    if (localUser) {
      throw new httpErrors.Conflict(
        `Local User with this username or email already exists.`
      );
    }

    const password = generateRandomPassword();
    const hashPassword = await crypto
      .createHash('sha256')
      .update(password)
      .digest('hex');

    const createdUser = await models.LocalUsers.create(
      {
        name,
        email,
        userName,
        department,
        client_mfa,
        is_strong_mfa,
        biometric_state,
        contact,
        password: hashPassword,
        change_pass: true,
        password_updated_at: moment(),
        runtime_bind_id: Math.floor(Date.now() / 1000),
        reachable_email : email
      },
      {
        transaction
      }
    );

    /*await models.Client.update(
        {
            is_local_deleted: 0
        },
        {
            where: {
                user: createdUser.userName
            },
            transaction
        }
    );*/

    await transaction.commit();

    insertIntoAdminLogs(getLocalUserCreateMessage(name), req);

    helper.sendNewLocalUserMail(
      createdUser.email,
      createdUser.name,
      createdUser.userName,
      password
    );

    await updateUserGroupFromUserId(createdUser.id, department);

    res.json(createdUser);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const searchClient = async (req, res, next) => {
  try {
    const data = await models.LocalUsers.findAll({
      where: {
        userName: {
          $like: `%${req.query.search.trim()}%`
        }
      }
    });

    res.json(data);
  } catch (error) {
    //'something went wrong in fetching local users'
    next(error);
  }
};

const updateLocalUser = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    let {
      name,
      email,
      userName,
      department = '',
      client_mfa = false,
      is_strong_mfa = false,
      biometric_state = false,
      contact = '',
      reachable_email,
      identity_provider
    } = req.body.localUser;

    if (department != '') {
      department = department.deptId;
    }

    const localUserUsername = await getUserNameFromId(req.params.id);

    if (localUserUsername === process.env.LOCAL_USER) {
      throw new httpErrors.Conflict("Can't edit this user");
    }

    /*
    const isLdapUserFound= await checkUserExistInLdapOrNot(userName, email);
    if(isLdapUserFound){
        throw new httpErrors.Conflict('User Already exists in LDAP.');
    }*/

    const localUser = await models.LocalUsers.findOne({
      where: {
        id: {
          $ne: req.params.id
        },
        $or: [
          {
            userName: {
              $eq: userName
            }
          },
          {
            email: {
              $eq: email
            }
          }
        ]
      }
    });

    if (localUser) {
      throw new httpErrors.Conflict(
        'Local User with this username or email already exists.'
      );
    }

    // const hashPassword = password ? await crypto.createHash('sha256').update(password).digest('hex') : '' ;
    if(identity_provider != 'Active Directory'){
      reachable_email = email
    }

    const updateClause = {
      name,
      email,
      department,
      client_mfa,
      is_strong_mfa,
      biometric_state,
      contact,
      reachable_email
    };

    const userToUpdate = await models.LocalUsers.findOne(
      {
        where: {
          id: req.params.id
        }
      },
      { transaction }
    );

    const previousUserDepartment = userToUpdate.dataValues.department;

    if (!userToUpdate) {
      throw new httpErrors.NotFound('No found to udpated');
    }

    // if (!password) {
    //     delete updateClause.password;
    // }

    const updatedUSer = await userToUpdate.update(updateClause, {
      transaction
    });

    // await models.LocalUsers.update(updateClause, {
    //     where: {
    //       id: req.params.id
    //     },
    //     transaction
    // });

    const relevantClients = await models.Client.findAll({
      where: {
        user_id: updatedUSer.id
      }
    });

    await Promise.all(
      relevantClients.map(_c => {
        let arr = _c.Name.split('}@{');
        _c.Name = '{' + name + '}@{' + arr[1];
        _c.Email = reachable_email;
        _c.user = userName;
        _c.client_mfa = client_mfa;
        _c.is_strong_mfa = is_strong_mfa;
        _c.biometric_state = biometric_state;
        return _c.save({ transaction });
      })
    );

    let changedColumns = Object.keys(updatedUSer._changed);
    changedColumns = changedColumns.filter(c =>
      clientColumnsToPublish.includes(c)
    );

    if (changedColumns.includes('client_mfa') && updatedUSer.client_mfa) {
      sendOtpToUsers([updatedUSer.id]);
    }

    if (changedColumns.length) {
      const columns = {};
      changedColumns.forEach(column => {
        columns[column] = updatedUSer[column];
      });

      const otherClients = relevantClients.filter(
        x => x.client_type == 'Client'
      );
      const serverClients = relevantClients.filter(
        x => x.client_type == 'Server'
      );

      await Promise.all([
        otherClients.length
          ? publishClients(actionType.UPDATED, {
              ids: otherClients.map(x => x.id),
              data: columns,
              type: 'Client'
            })
          : [],

        serverClients.length
          ? publishClients(actionType.UPDATED, {
              ids: serverClients.map(x => x.id),
              data: columns,
              type: 'Server'
            })
          : []
      ]);
    }

    await transaction.commit();

    await updateUserGroupFromUserId(
      req.params.id,
      department,
      previousUserDepartment
    );

    insertIntoAdminLogs(getLocalUserEditMessage(name), req);

    res.json('User updated successfully');
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const createNewClient = async (userRecord, deviceRecord, transaction) => {
  try {
    const [encKey, hmacKey] = await Promise.all([
      OpensslObj.generateEncKey(),
      OpensslObj.generateHMacKey()
    ]);
    if (!encKey.error && !hmacKey.error) {
      const enc_key = encKey.key;
      const HMacKey = hmacKey.key;
      const isServerType =
        [
          'IoT Bridge Client',
          'Physical Server',
          'Thin Client (Bridge Client)',
          'Server Bridge Client',
          'Virtual Server',
          'ARM Platform'
        ].indexOf(deviceRecord.device_type) != -1;
      const createdClient = await models.Client.create(
        {
          user: userRecord.userName,
          device_hardware_id: crypto
            .createHash('md5')
            .update(deviceRecord.device_id + '')
            .digest('hex'),
          device_os: deviceRecord.os,
          enc_key,
          HMacKey,
          agent_id: 0,
          is_local: userRecord.identity_provider == 'Local User',
          client_type: isServerType ? 'Server' : 'Client',
          Name: '{' + userRecord.name + '}@{' + deviceRecord.device_name + '}',
          Email: userRecord.reachable_email,
          client_status: 1,
          access_restrictions: 0,
          device_os_full_name: deviceRecord.os,
          user_id: userRecord.id,
          device_id: deviceRecord.id,
          geoip_track: deviceRecord.geoip_track,
          serial_number: deviceRecord.device_id + '',
          client_mfa: userRecord.client_mfa,
          biometric_state: userRecord.biometric_state,
          is_strong_mfa: userRecord.is_strong_mfa,
          client_darkening: deviceRecord.darkening_state,
          HMAC_KEY_LEN: 24,
          ENC_KEY_LEN: 32
        },
        {
          transaction
        }
      );

      await models.ClientSoftwareInventory.create(
        {
          client_id: createdClient.id
        },
        { transaction }
      );

      await Promise.all([
        models.sequelize.query(
          'UPDATE Clients SET agent_id = id where agent_id = 0;',
          {
            transaction
          }
        ),
        publishMsgFile(actionType.ADDED, {
          id: createdClient.id,
          data: {
            enc_key: createdClient.enc_key,
            ENC_KEY_LEN: createdClient.ENC_KEY_LEN,
            HMacKey: createdClient.HMacKey,
            HMAC_KEY_LEN: createdClient.HMAC_KEY_LEN,
            device_hardware_id: createdClient.device_hardware_id
          },
          type: createdClient.client_type
        })
      ]);
      await transaction.commit();
      if (deviceRecord.os === 'android' || deviceRecord.os === 'ios') {
        const packageClient = await models.Client.findById(
          createdClient.dataValues.id
        );
        generateZip(packageClient || createdClient, 1);
      } else {
        if (createdClient.device_os) {
          if (createdClient.device_os !== 'ubuntuserver20armv8') {
            emailLinkExpiryInMinutes = await settings.getExpirySettings();
            const link = await generateLink({
              os: createdClient.device_os,
              component: 'client',
              componentId: createdClient.id,
              expiry_time: emailLinkExpiryInMinutes,
              filename: 'installer'
            });

            // const link = {}

            let onuri = process.env.CONTENT_URL + '/api/v1/links';
            helper.sendInstallerEmail({
              recipient: createdClient.Email,
              emailSubject: 'Invisily Client Software Installation',
              recipientName: userRecord.name,
              onPremiseDownloadLink: `${onuri}/download/${link.key}`,
              offPremiseDownloadLink: `${onuri}/download/${link.key}`,
              templateFileName:
                projectConfig.onPremApiUrl !== projectConfig.offPremApiUrl
                  ? 'client_onboarding_email.html'
                  : 'client_onboarding_email_single.html',
              device_os: installerDescriptiveName[createdClient.device_os]
                ? installerDescriptiveName[createdClient.device_os]
                : createdClient.device_os,
              package_expiry_time: moment(link.link_expired_at).fromNow(true)
            });
          }
        }
      }
    }
  } catch (error) {
    throw new Error(error.message);
  }
};

const deviceAssociationWithUser = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    const userId = req.params.id;
    const { operation = '' } = req.body;

    if (operation == 'add') {
      let device = req.body.deviceName;
      const deviceRecord = await models.DeviceInventories.findOne({
        where: {
          device_name: device
        }
      });

      device_type = ['Physical Server', 'Virtual Server'].includes(
        deviceRecord.device_type
      )
        ? 'Server'
        : 'Client';
      if (device_type == 'Server') {
        const clientsTypeMismatch = await models.Client.findOne({
          where: {
            user_id: userId,
            client_type: 'Client'
          }
        });

        if (clientsTypeMismatch) {
          throw new httpErrors(
            400,
            'Can not add server device when desktop device is already added.'
          );
        }
      } else if (device_type == 'Client') {
        const clientsTypeMismatch = await models.Client.findOne({
          where: {
            user_id: userId,
            client_type: 'Server'
          }
        });
        if (clientsTypeMismatch) {
          throw new httpErrors(
            400,
            'Can not add desktop device when server device is already added.'
          );
        }
      }

      if (!deviceRecord) {
        throw new httpErrors.NotFound('Device Not Found.');
      }

      const query =  `
        Select di.device_id, di.device_name, c.Name, c.client_status 
        from DeviceInventories di 
        INNER JOIN Clients c ON c.device_id = di.id
        WHERE (di.device_type = 'Physical Server' OR di.device_type = 'Virtual Server') AND c.user IS NULL AND c.client_status = 4 AND di.device_name = :device
      `;

      const [onboardeds2sClient] = await models.sequelize.query(query, {
        replacements: { device: device },
        type: models.Sequelize.QueryTypes.SELECT
      });

      if (onboardeds2sClient) {
        throw new httpErrors.BadRequest('Cannot add already onboarded S2S client.');
      }

      const user = await models.LocalUsers.findById(userId);
      if (!user) {
        throw new httpErrors.NotFound('User Not Found.');
      }
      if (user.department !== deviceRecord.department) {
        throw new httpErrors(400, 'Cannot add device from another department');
      }

      const associatedUser = await models.LocalUsers.findOne({
        where: {
          id: deviceRecord.user_id ? deviceRecord.user_id : userId
        }
      });

      if (deviceRecord.user_id) {
        throw new httpErrors.NotFound(
          `Device already assigned to user '${
            associatedUser ? associatedUser.name : deviceRecord.user_id
          }'.`
        );
      }
      deviceRecord.user_id = userId;
      deviceRecord.userName = associatedUser ? associatedUser.name : '';
      deviceRecord.has_agent = true;
      await deviceRecord.save({ transaction });
      const userRecord = await models.LocalUsers.findOne({
        where: {
          id: userId
        }
      });
      if (userRecord) {
        const clientCount = await models.Client.count({});
        const license = await models.License.findOne({
          order: [['id', 'DESC']],
          limit: 1
        });

        if (!license) {
          throw new httpErrors.Conflict('License not found.');
        }

        if (clientCount == license.max_number_of_clients) {
          throw new httpErrors.Conflict('Client limit already reached.');
        }

        insertIntoAdminLogs(getDevicetoLocalUserAdded(user.name, device), req);

        await createNewClient(userRecord, deviceRecord, transaction);
      }
      res.json({
        success: true,
        message: 'Device added successfully.'
      });
    } else if (operation == 'remove') {
      let device = req.body.deviceId;
      const userId = req.params.id;
      const user = await models.LocalUsers.findById(userId);
      const deviceRecord = await models.DeviceInventories.findOne({
        where: {
          id: device
        }
      });
      if (!deviceRecord) {
        throw new httpErrors.NotFound('Device Not Found.');
      }
      if (!deviceRecord.user_id) {
        throw new httpErrors.NotFound('Device is not assigned to any user.');
      }

      const clientRecord = await models.Client.findOne({
        where: {
          device_id: device,
          user: {
            [Op.ne]: null,
          },
        }
      });

      if (clientRecord) {
        // Check if device is asigned to bridge
        const bridge = await models.ServerBridge.findOne({
          where: {
            clientId: clientRecord.id
          },
          include: [
            {
              model: models.ServerBridgeDeviceId,
              as: 'serverBridgeIds'
            }
          ]
        });
        if (bridge && bridge.serverBridgeIds) {
          throw new httpErrors.Conflict(
            'Cannot Unbind device. Delete associated Bridge and try again.'
          );
        }
      }

      deviceRecord.user_id = null;
      deviceRecord.userName = '';
      deviceRecord.has_agent = false;
      await deviceRecord.save({ transaction });

      if (clientRecord) {
        const excludePolicies = await models.sequelize.query(`
          SELECT id FROM Entitlements WHERE id IN (
            SELECT policy_id AS id
            FROM ClientGroupPolicies
            WHERE ClientGroupPolicies.policy_type_id = 1
            GROUP BY ClientGroupPolicies.policy_id
            HAVING COUNT(ClientGroupPolicies.id) > 0
            UNION
            SELECT id FROM Entitlements WHERE all_clients = true
          )
        `);

        let policiesToExclude = [];

        if (excludePolicies[0] && excludePolicies[0].length > 0) {
          policiesToExclude = excludePolicies[0].map(policy => policy.id);
        }

        let Entitlements = await models.sequelize.query(`
          SELECT ClientPolicies.policy_id AS id
          FROM ClientPolicies
          WHERE ClientPolicies.policy_type_id = 1 AND ClientPolicies.policy_id IN (
            SELECT ClientPolicies.policy_id
            FROM ClientPolicies
            WHERE ClientPolicies.policy_type_id = 1 AND ClientPolicies.client_id IN (${[
              clientRecord.id
            ].toString()})
          )
          GROUP BY ClientPolicies.policy_id
          HAVING count(ClientPolicies.id) < 2
        `);
        Entitlements = Entitlements[0].map(policy => {
          return policy.id;
        });

        if (Entitlements.length != 0) {
          let query1 = `
            DELETE FROM ClientPolicies WHERE ClientPolicies.policy_type_id = 1 AND ClientPolicies.policy_id IN (${Entitlements.toString()})
          `;
          let query2 = `
            DELETE FROM Entitlements WHERE Entitlements.id IN (${Entitlements.toString()})
          `;
          if (policiesToExclude.length > 0) {
            query1 =
              query1 +
              ` and policy_id not in (${policiesToExclude.toString()}) `;
            query2 =
              query2 + ` and id not in (${policiesToExclude.toString()}) `;
          }
          await models.sequelize.query(query1, { transaction });
          await models.sequelize.query(query2, { transaction });
        }

        //Security (Device) Policies removal
        let devicePolicies = await models.sequelize.query(`
          SELECT ClientPolicies.policy_id AS id
          FROM ClientPolicies
          WHERE ClientPolicies.policy_type_id = 2 AND ClientPolicies.policy_id IN (
            SELECT ClientPolicies.policy_id
            FROM ClientPolicies
            WHERE ClientPolicies.policy_type_id = 2 AND ClientPolicies.client_id IN (${[
              clientRecord.id
            ].toString()})
          )
          GROUP BY ClientPolicies.policy_id
          HAVING COUNT(ClientPolicies.id) < 2;
        `);
        devicePolicies = devicePolicies[0].map(policy => {
          return policy.id;
        });

        if (devicePolicies.length != 0) {
          await models.sequelize.query(
            `
            DELETE FROM ClientPolicies WHERE policy_type_id = 2 AND policy_id IN (${devicePolicies.toString()});
          `,
            { transaction }
          );
          await models.sequelize.query(
            `
            DELETE FROM DevicePolicies WHERE id IN (${devicePolicies.toString()});
          `,
            { transaction }
          );
        }

        // Software Policies removal
        let softwarePolicies = await models.sequelize.query(`
          SELECT ClientPolicies.policy_id AS id
          FROM ClientPolicies
          WHERE ClientPolicies.policy_type_id = 4 AND ClientPolicies.policy_id IN (
            SELECT ClientPolicies.policy_id
            FROM ClientPolicies
            WHERE ClientPolicies.policy_type_id = 4 AND ClientPolicies.client_id IN (${[
              clientRecord.id
            ].toString()})
          )
          GROUP BY ClientPolicies.policy_id
          HAVING COUNT(client_id) < 2;
        `);
        softwarePolicies = softwarePolicies[0].map(policy => {
          return policy.id;
        });

        if (softwarePolicies.length != 0) {
          await models.sequelize.query(
            `
            DELETE FROM ClientPolicies WHERE policy_type_id = 4 AND policy_id IN (${softwarePolicies.toString()});
          `,
            { transaction }
          );
          await models.sequelize.query(
            `
            DELETE FROM SoftwarePolicies WHERE id IN (${softwarePolicies.toString()});
          `,
            { transaction }
          );
        }

        // Software Process Policies removal
        let softwareProcessPolicies = await models.sequelize.query(`
          SELECT ClientPolicies.policy_id AS id
          FROM ClientPolicies
          WHERE ClientPolicies.policy_type_id = 5 AND ClientPolicies.policy_id IN (
            SELECT ClientPolicies.policy_id
            FROM ClientPolicies
            WHERE ClientPolicies.policy_type_id = 5 AND ClientPolicies.client_id IN (${[
              clientRecord.id
            ].toString()})
          )
          GROUP BY ClientPolicies.policy_id
          HAVING COUNT(client_id) < 2;
        `);
        softwareProcessPolicies = softwareProcessPolicies[0].map(policy => {
          return policy.id;
        });

        if (softwareProcessPolicies.length != 0) {
          await models.sequelize.query(
            `
            DELETE FROM ClientPolicies WHERE policy_type_id = 5 AND policy_id IN (${softwareProcessPolicies.toString()});
          `,
            { transaction }
          );
          await models.sequelize.query(
            `
            DELETE FROM SoftwareProcessPolicies WHERE id IN (${softwareProcessPolicies.toString()});
          `,
            { transaction }
          );
        }
        // await clientRecord.destroy({ transaction });
        await models.sequelize.query(
          `DELETE FROM Clients WHERE device_id = ${device} and user is not null`,
          { transaction }
        );

        await publishClients(actionType.DELETED, {
          ids: [clientRecord.id],
          data: {},
          type: clientRecord.client_type
        });
      }
      await transaction.commit();

      insertIntoAdminLogs(
        getDevicetoLocalUserDeleted(user.name, deviceRecord.device_name),
        req
      );

      res.json({
        success: true,
        message: 'Device removed successfully.'
      });
    } else {
      throw new httpErrors.NotFound('Invalid Operation.');
    }
  } catch (error) {
    if (error.name === 'SequelizeForeignKeyConstraintError') {
      res
        .status(500)
        .json({ message: 'Remove client from Entitlement(s) first' });
    } else {
      next(error);
    }
  }
};

const getUserDeviceRelatedPolicies = async (req, res, next) => {
  try {
    const { deviceId } = req.params;

    let policiesCount = 0;

    const clientRecord = await models.Client.findOne({
      where: {
        device_id: deviceId
      }
    });

    if (clientRecord && clientRecord.id) {
      const devicePolicyCount = await models.sequelize.query(
        `
      SELECT count(id) as count from DevicePolicies where id in
      (SELECT policy_id as id
        FROM
            ClientPolicies
        WHERE
            ClientPolicies.policy_type_id = 2 AND ClientPolicies.policy_id IN (SELECT ClientPolicies.policy_id FROM ClientPolicies WHERE ClientPolicies.policy_type_id = 2 AND ClientPolicies.client_id IN (${[
              clientRecord.id
            ].toString()}))
        Group by
            ClientPolicies.policy_id
        HAVING
            count(ClientPolicies.id) < 2)

      `
      );

      if (devicePolicyCount[0].length > 0) {
        policiesCount += devicePolicyCount[0][0].count;
      }

      const accessPolicyCount = await models.sequelize.query(
        `
        SELECT count(id) as count FROM Entitlements WHERE Entitlements.id IN (SELECT ClientPolicies.policy_id AS id
          FROM
              ClientPolicies
          WHERE
              ClientPolicies.policy_type_id = 1 AND ClientPolicies.policy_id IN (SELECT ClientPolicies.policy_id FROM ClientPolicies WHERE ClientPolicies.policy_type_id = 1 AND ClientPolicies.client_id IN (${[
                clientRecord.id
              ].toString()}))
          Group by
              ClientPolicies.policy_id
          HAVING
              count(ClientPolicies.id) < 2
            )
            and  Entitlements.id NOT IN (
          SELECT id from Entitlements where id in (
            SELECT policy_id AS id
            FROM
                ClientGroupPolicies
            WHERE
                ClientGroupPolicies.policy_type_id = 1
            Group by
                ClientGroupPolicies.policy_id
            HAVING
                count(ClientGroupPolicies.id) > 0
            UNION
              SELECT id from Entitlements where all_clients = true
        )
      )

      `
      );

      if (accessPolicyCount[0].length > 0) {
        policiesCount += accessPolicyCount[0][0].count;
      }

      res.status(200).json({
        data: policiesCount
      });
    }
  } catch (e) {
    next(e);
  }
};

const deleteLocalUser = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    const user = await models.LocalUsers.findOne({
      where: {
        id: req.params.id
      }
    });

    if (!user) {
      throw new httpErrors.NotFound('User not found');
    }

    if (user.userName === process.env.LOCAL_USER) {
      throw new httpErrors.Conflict("Can't delete this user");
    }
    // User is Only created as a client Either that user is attached with Device or B2B Bridge
    // Here we check either given user is a client if yes than don't delete it
    const clients = await models.Client.findAll({
      where: {
        user: user.userName
      }
    });

    if (clients.length > 0) {
      throw new httpErrors.Conflict(
        'Cannot delete User. Delete bound device(s) first.'
      );
    }

    await user.destroy({ transaction });

    await models.Client.update(
      {
        is_local_deleted: 1
      },
      {
        where: {
          user: req.params.userName
        },
        transaction
      }
    );

    insertIntoAdminLogs(getLocalUserDeleteMessage(user.name), req);

    await transaction.commit();

    res.json('User deleted successfully');
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const bulkPasswordChange = async (req, res, next) => {
  try {
    const {
      selectAll = false,
      selectAllExcept = [],
      selectedUsers = [],
      search = ''
    } = req.body;

    BulkPasswordChangeQueue.addtoQueue({
      selectAll,
      selectAllExcept,
      selectedUsers,
      search
    });

    res.send('Password change process initiated.');
  } catch (error) {
    next(error);
  }
};

const bulkDeleteUser = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    const { selectAll, selectedIds } = req.body;

    let options = {
      where: {
        id: {
          ...(selectAll
            ? {
                [Op.notIn]: selectedIds
              }
            : {
                [Op.in]: selectedIds
              })
        }
      },
      transaction
    };

    options = {
      ...options,
      where: {
        ...options.where,
        userName: {
          ...options.where.userName,
          [Op.notLike]: `%gwadmin%`
        }
      }
    };

    const localUsers = await models.LocalUsers.findAll(options, {
      transaction
    });

    const clientOptions = {
      where: {
        user: {
          [Op.in]: localUsers.map(_u => _u.userName)
        }
      }
    };

    let clients = await models.Client.findAll(clientOptions, { transaction });

    if (clients.length != 0) {
      throw new httpErrors.Conflict(
        'One or more users have devices associated with them.'
      );
    }

    await models.LocalUsers.destroy(options, { transaction });

    await transaction.commit();

    return res.json({ message: 'User(s) deleted successfully.' });
  } catch (error) {
    next(error);
  }
};

const getCustomUserGroups = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();

  try {
    const customUserGroupsQuery = `
    SELECT * FROM CustomUserGroups customUserGroups LEFT JOIN 
    (SELECT group_id, COUNT(*) AS record_count FROM UserGroupsPivot
    GROUP BY group_id) AS group_counts ON
    customUserGroups.id = group_counts.group_id;
    `;
    const customUserGroups = await models.sequelize.query(
      customUserGroupsQuery
    );

    await transaction.commit();
    res.json(customUserGroups[0]);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const getCustomUserGroupById = async (req, res, next) => {
  try {
    const id = req.params.id;

    const groupQuery = `
      SELECT * FROM CustomUserGroups WHERE id = ${id}
    `;

    const [group] = await models.sequelize.query(groupQuery);

    const query = `SELECT * FROM UserGroupsPivot userGroupsPivot
    LEFT JOIN CustomUserGroups customUserGroups on customUserGroups.id = userGroupsPivot.group_id 
    LEFT JOIN LocalUsers localUsers on localUsers.id = userGroupsPivot.user_id
    WHERE customUserGroups.id = :customUserGroupID
    `;

    const data = await models.sequelize.query(query, {
          replacements: { customUserGroupID: id },
          type: models.sequelize.QueryTypes.SELECT,
    })

    const totalCountQuery = 'SELECT COUNT(*) as totalCount FROM LocalUsers;';
    const totalCount = (await models.sequelize.query(totalCountQuery))[0];

    console.log(
      'Comparing the Users Length ',
      totalCount[0].totalCount,
      group[0].length
    );

    res.status(200).json({
      group: group[0],
      data: data,
      count: totalCount[0].totalCount,
      all_selected: totalCount[0].totalCount == group[0].length
    });
  } catch (error) {
    next(error);
  }
};

const updateUserGroup = async (req, res, next) => {
  const { localUser } = req.body;
  const transaction = await models.sequelize.transaction();
  try {
    var conditions = localUser.rawQuery.split(' And ');
    var identityProviders = null;
    var organizations = null;
    var departments = null;
    var orderBy = null;
    var usernames = null;
    var names = null;
    var emails = null;
    let exisitngUsers = [];
    let exisitngUsersState = {};
    let sqlQuery = '';

    if (localUser.selectedIds.length > 0) conditions = [];

    const customUserGroup = await models.CustomUserGroups.findOne({
      where: {
        group_name: localUser.groupName,
        group_description: localUser.groupDesc
      }
    });

    if (localUser.allUser) {
      console.log('Going to execute');
      const query = `DELETE FROM sdp.UserGroupsPivot WHERE group_id = ${customUserGroup.dataValues.id};`;
      await models.sequelize.query(query);
    }

    // Loop through conditions and extract values
    if (localUser.rawQuery && localUser.rawQuery.length > 0) {
      for (var i = 0; i < conditions.length; i++) {
        var condition = conditions[i];
        if (condition.startsWith('Identity Provider = ')) {
          identityProviders = condition.split('Identity Provider = ')[1];
        } else if (condition.startsWith('Organization = ')) {
          organizations = condition.split('Organization = ')[1];
        } else if (condition.startsWith('Department = ')) {
          if (departments === null) {
            departments = [];
          }
          departments.push(
            decodeURIComponent(condition.split('Department = ')[1])
          );
        } else if (condition.startsWith('Username = ')) {
          if (usernames === null) {
            usernames = [];
          }
          usernames.push(decodeURIComponent(condition.split('Username = ')[1]));
        } else if (condition.startsWith('Name = ')) {
          if (names === null) {
            names = [];
          }
          names.push(decodeURIComponent(condition.split('Name = ')[1]));
        } else if (condition.startsWith('Email = ')) {
          if (emails === null) {
            emails = [];
          }
          emails.push(decodeURIComponent(condition.split('Email = ')[1]));
        } else if (condition.startsWith('OrderBy = ')) {
          orderBy = condition.split('OrderBy = ')[1];
        }
      }

      const queryIdentityProvider = identityProviders
        ? `WHERE localUsers.identity_provider = '${identityProviders}'`
        : '';
      const queryOrganizations = organizations
        ? `${
            identityProviders ? ' AND ' : 'WHERE'
          } organizations.name = '${organizations}'`
        : '';
      const queryDepartment = departments
        ? `${
            identityProviders || organizations ? ' AND ' : 'WHERE'
          } departments.name = '${departments}'`
        : '';
      const queryUsernames = usernames
        ? `${
            organizations || identityProviders || departments
              ? ' AND '
              : 'WHERE'
          } localUsers.userName IN ('${usernames.join("','")}')`
        : '';
      const queryNames = names
        ? `${
            organizations || identityProviders || departments || usernames
              ? ' AND '
              : 'WHERE'
          } localUsers.name IN ('${names.join("','")}')`
        : '';
      const queryEmails = emails
        ? `${
            organizations ||
            identityProviders ||
            departments ||
            usernames ||
            names
              ? ' AND '
              : 'WHERE'
          } localUsers.email IN ('${emails.join("','")}')`
        : '';

      sqlQuery = `
      SELECT localUsers.*, departments.id as departmentId, departments.name as departmentName, 
      organizations.id as organizationId, organizations.name as organizationName 
      FROM LocalUsers localUsers
      LEFT JOIN Departments departments on localUsers.department = departments.id
      LEFT JOIN Organizations organizations ON departments.organization_id = organizations.id
      ${queryIdentityProvider} ${queryOrganizations} ${queryDepartment} ${queryUsernames} ${queryNames} ${queryEmails}
      ORDER BY localUsers.userName ${orderBy == 'Ascending' ? 'asc' : 'desc'}`;
    } else {
      if (localUser.allUser) {
        console.log('Ah Shit lets debug');

        sqlQuery = `
          SELECT localUsers.*, departments.id as departmentId, departments.name as departmentName, 
          organizations.id as organizationId, organizations.name as organizationName 
          FROM LocalUsers localUsers
          LEFT JOIN Departments departments on localUsers.department = departments.id
          LEFT JOIN Organizations organizations ON departments.organization_id = organizations.id
          ORDER BY localUsers.userName desc`;
      } else {
        sqlQuery = `
          SELECT userGroupSQLQuery.query as query 
          FROM CustomUserGroups customUserGroups
          LEFT JOIN UserGroupSQLQuery userGroupSQLQuery ON userGroupSQLQuery.id = customUserGroups.sql_query_id 
          WHERE customUserGroups.id  = ${customUserGroup?.dataValues?.id}`;
        let result = (await models.sequelize.query(sqlQuery))[0];
        sqlQuery = result[0]?.query;
        console.log('Let check out ', sqlQuery);
      }
    }

    let result = (await models.sequelize.query(sqlQuery))[0];
    let ids = result.map(item => item.id);
    let userId =
      localUser?.selectedIds.length > 0
        ? localUser.selectedIds
        : [...new Set(ids)];

    const userExist = await models.UserGroupsPivot.findOne({
      where: {
        user_id: { [Op.in]: localUser.selectedIds },
        group_id: customUserGroup.id
      }
    });

    if (userExist) {
      throw new httpErrors.Conflict('User(s) already exist in this group');
    }

    let isDeclarativeIds = localUser.selectedIds;

    const userPivot = await models.UserGroupsPivot.findAll({
      where: {
        group_id: customUserGroup.dataValues.id
      }
    });

    userPivot.forEach(value => {
      exisitngUsers.push(value.dataValues.user_id);

      if (!exisitngUsersState[value?.user_id]) {
        exisitngUsersState[value?.user_id] = {
          user_id: value?.user_id,
          is_declarative: value?.is_declarative
        };
      }
    });

    let staticCheck = false;
    if (localUser.rawQuery.length == 0) {
      userId = [...userId, ...exisitngUsers];
      localUser.selectedIds = [...localUser.selectedIds, ...exisitngUsers];
      staticCheck = true;
    } else if (
      !localUser.rawQuery.length == 0 &&
      localUser.selectedIds.length > 0
    ) {
      localUser.selectedIds = [...localUser.selectedIds, ...exisitngUsers];
      staticCheck = true;
    } else {
      localUser.selectedIds = [...new Set(ids)];
    }

    await models.UserGroupRawQuery.update(
      { query: localUser.rawQuery },
      { where: { id: customUserGroup.dataValues.raw_query_id } }
    );

    await models.UserGroupSQLQuery.update(
      { query: sqlQuery },
      { where: { id: customUserGroup.dataValues.sql_query_id } }
    );

    await models.UserGroupsPivot.destroy({
      where: {
        group_id: customUserGroup.dataValues.id
      }
    });

    localUser.allUser
      ? userId.map(async (userId, index) => {
          await models.UserGroupsPivot.create({
            group_id: customUserGroup.dataValues.id,
            user_id: userId
          });
        })
      : localUser.selectedIds.map(async (userId, index) => {
          await models.UserGroupsPivot.create({
            group_id: customUserGroup.dataValues.id,
            user_id: userId
          });
        });

    if (staticCheck) {
      let userState = Object.values(exisitngUsersState);
      let usersToUpdate = [...isDeclarativeIds];

      // only add those users if the previous state was static
      userState.forEach(value => {
        if (value?.is_declarative === false) {
          usersToUpdate.push(value?.user_id);
        }
      });

      usersToUpdate.map(async (userId, index) => {
        await models.UserGroupsPivot.update(
          {
            is_declarative: false
          },
          {
            where: {
              group_id: customUserGroup.dataValues.id,
              user_id: userId
            }
          }
        );
      });
    }

    await publishUserGroups(actionType.UPDATED, {
      id: +customUserGroup.dataValues.id,
      new_users: localUser.allUser
        ? [...new Set(userId)]
        : [...new Set(localUser.selectedIds)],
      del_users: exisitngUsers
    });

    await transaction.commit();
    res.json('user(s) in group updated successfully');
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const createCustomUserGroups = async (req, res, next) => {
  const { localUser } = req.body;
  const transaction = await models.sequelize.transaction();

  try {
    var conditions = localUser.rawQuery.split(' And ');
    var identityProviders = null;
    var organizations = null;
    var departments = null;
    var orderBy = null;
    var usernames = null;
    var names = null;
    var emails = null;

    const isUserGroupNameExists = await models.CustomUserGroups.findOne({
      where: {
        group_name: localUser.groupName
      }
    });

    console.log('isUserGroupNameExists ');

    if (isUserGroupNameExists?.dataValues?.id) {
      throw new httpErrors.Conflict(
        'User group with the same name already exits.'
      );
    }

    if (localUser.selectedIds.length > 0) conditions = [];

    // Loop through conditions and extract values
    for (var i = 0; i < conditions.length; i++) {
      var condition = conditions[i];
      if (condition.startsWith('Identity Provider = ')) {
        if (identityProviders === null) {
          identityProviders = [];
        }
        identityProviders.push(
          decodeURIComponent(condition.split('Identity Provider = ')[1])
        );
      } else if (condition.startsWith('Organization = ')) {
        if (organizations === null) {
          organizations = [];
        }
        organizations.push(
          decodeURIComponent(condition.split('Organization = ')[1])
        );
      } else if (condition.startsWith('Department = ')) {
        if (departments === null) {
          departments = [];
        }
        departments.push(
          decodeURIComponent(condition.split('Department = ')[1])
        );
      } else if (condition.startsWith('Username = ')) {
        if (usernames === null) {
          usernames = [];
        }
        usernames.push(decodeURIComponent(condition.split('Username = ')[1]));
      } else if (condition.startsWith('Name = ')) {
        if (names === null) {
          names = [];
        }
        names.push(decodeURIComponent(condition.split('Name = ')[1]));
      } else if (condition.startsWith('Email = ')) {
        if (emails === null) {
          emails = [];
        }
        emails.push(decodeURIComponent(condition.split('Email = ')[1]));
      } else if (condition.startsWith('OrderBy = ')) {
        orderBy = condition.split('OrderBy = ')[1];
      }
    }

    const queryIdentityProvider = identityProviders
      ? `WHERE localUsers.identity_provider = '${identityProviders}'`
      : '';
    const queryOrganizations = organizations
      ? `${
          identityProviders ? ' AND ' : 'WHERE'
        } organizations.name = '${organizations}'`
      : '';
    const queryDepartment = departments
      ? `${
          organizations || identityProviders ? ' AND ' : 'WHERE'
        } departments.name = '${departments}'`
      : '';
    const queryUsernames = usernames
      ? `${
          organizations || identityProviders || departments ? ' AND ' : 'WHERE'
        } localUsers.userName IN ('${usernames.join("','")}')`
      : '';
    const queryNames = names
      ? `${
          organizations || identityProviders || departments || usernames
            ? ' AND '
            : 'WHERE'
        } localUsers.name IN ('${names.join("','")}')`
      : '';
    const queryEmails = emails
      ? `${
          organizations ||
          identityProviders ||
          departments ||
          usernames ||
          names
            ? ' AND '
            : 'WHERE'
        } localUsers.email IN ('${emails.join("','")}')`
      : '';

    const sqlQuery = `
      SELECT localUsers.*, departments.id as departmentId, departments.name as departmentName, 
      organizations.id as organizationId, organizations.name as organizationName 
      FROM LocalUsers localUsers
      LEFT JOIN Departments departments on localUsers.department = departments.id
      LEFT JOIN Organizations organizations ON departments.organization_id = organizations.id
      ${queryIdentityProvider} ${queryOrganizations} ${queryDepartment} ${queryUsernames} ${queryNames} ${queryEmails}
      ORDER BY localUsers.userName ${orderBy == 'Ascending' ? 'asc' : 'desc'}`;

    console.log(sqlQuery);

    const sqlQueryForController = `
    SELECT localUsers.id as user_id
    FROM LocalUsers localUsers
    LEFT JOIN Departments departments on localUsers.department = departments.id
    LEFT JOIN Organizations organizations ON departments.organization_id = organizations.id
    ${queryIdentityProvider} ${queryOrganizations} ${queryDepartment} ${queryUsernames} ${queryNames} ${queryEmails}
    ORDER BY localUsers.userName ${orderBy == 'Ascending' ? 'asc' : 'desc'}`;

    let result = (await models.sequelize.query(sqlQuery))[0];
    let ids = result.map(item => item.id);
    console.log('ids ', ids);
    const userId =
      localUser?.selectedIds.length > 0
        ? localUser.selectedIds
        : [...new Set(ids)];

    console.log('userId ', userId);

    const userGroupRawQuery = await models.UserGroupRawQuery.create(
      {
        query: localUser.rawQuery
      },
      { transaction }
    );

    const userGroupSQLQuery = await models.UserGroupSQLQuery.create(
      {
        query: sqlQuery,
        user_id_query: sqlQueryForController
      },
      { transaction }
    );

    const customUserGroup = await models.CustomUserGroups.create(
      {
        group_name: localUser.groupName,
        group_description: localUser.groupDesc,
        raw_query_id: userGroupRawQuery?.id,
        sql_query_id: userGroupSQLQuery?.id,
        runtime_bind_id: Date.now()
      },
      { transaction }
    );

    userId.map(async (userId, index) => {
      await models.UserGroupsPivot.create({
        group_id: customUserGroup.id,
        user_id: userId,
        is_declarative: localUser?.selectedIds.length > 0 ? 0 : 1
      });
    });

    if (identityProviders) {
      identityProviders?.map(
        async (value, index) => {
          await models.CustomUserGroupsDetail.create({
            group_id: customUserGroup.id,
            label_name: 'Identity Provider',
            label_value: value
          });
        },
        { transaction }
      );
    }

    if (organizations) {
      organizations?.map(
        async (value, index) => {
          await models.CustomUserGroupsDetail.create({
            group_id: customUserGroup.id,
            label_name: 'Organization',
            label_value: value
          });
        },
        { transaction }
      );
    }

    if (departments) {
      departments?.map(
        async (value, index) => {
          await models.CustomUserGroupsDetail.create({
            group_id: customUserGroup.id,
            label_name: 'Department',
            label_value: value
          });
        },
        { transaction }
      );
    }
    await publishUserGroups(actionType.ADDED, {
      id: customUserGroup.id
    });

    await transaction.commit();
    res.json('custom user group added');
  } catch (error) {
    console.log(error.message);
    await transaction.rollback();
    next(error);
  }
};

const deleteUserFromCustomGoup = async (req, res, next) => {
  try {
    const userId = req.params.userId;
    const groupId = req.params.groupId;

    const userRemove = await models.UserGroupsPivot.destroy({
      where: {
        user_id: userId,
        group_id: groupId
      }
    });

    if (!userRemove) {
      throw new httpErrors.NotFound('Failed to remove user from group');
    }

    res.status(200).json({ message: 'User removed from group successfully' });
  } catch (error) {
    next(error);
  }
};

const deleteCustomGroup = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    const id = req.params.id;

    const userGroupDetail = await models.CustomUserGroups.find({
      where: {
        id: id
      },
      transaction: transaction
    });

    console.log('userGroupDetail');

    await models.CustomUserGroupsDetail.destroy({
      where: {
        group_id: id
      },
      transaction: transaction
    });
    console.log('CustomUserGroupsDetail');

    await models.UserGroupsPivot.destroy({
      where: {
        group_id: id
      },
      transaction: transaction
    });

    console.log('UserGroupsPivot');

    const userGroup = await models.CustomUserGroups.destroy({
      where: {
        id: id
      },
      transaction: transaction
    });
    console.log('CustomUserGroups');

    await models.UserGroupRawQuery.destroy({
      where: {
        id: userGroupDetail?.raw_query_id
      },
      transaction: transaction
    });
    console.log('UserGroupRawQuery');

    await models.UserGroupSQLQuery.destroy({
      where: {
        id: userGroupDetail?.sql_query_id
      },
      transaction: transaction
    });
    console.log('UserGroupSQLQuery');

    await transaction.commit();

    if (userGroup) {
      await publishUserGroups(actionType.DELETED, {
        id: +id
      });

      res.status(200).json({ message: 'User group deleted successfully' });
    } else {
      await transaction.rollback();
      res.status(500).json({ message: 'Faileed to delete user group' });
    }
  } catch (error) {
    if (error.name === 'SequelizeForeignKeyConstraintError') {
      await transaction.rollback();
      res.status(500).json({ message: 'Failed to delete user group' });
    } else {
      await transaction.rollback();
      next(error);
    }
  }
};

const updateCustomGroup = async (req, res, next) => {
  try {
    const { editGroupName, editGroupDesc } = req.body.data;
    const { id } = req.params;
    const userGroup = await models.CustomUserGroups.update(
      {
        group_name: editGroupName,
        group_description: editGroupDesc
      },
      {
        where: {
          id: id
        }
      }
    );

    if (userGroup) {
      await publishUserGroups(actionType.UPDATED, {
        id: +id,
        new_users: [],
        del_users: []
      });

      res.status(200).json({ message: 'User group updated successfully' });
    } else {
      res.status(500).json({ message: 'Failed to update user group' });
    }
  } catch (error) {
    next(error);
  }
};

const customGroupData = async (req, res, next) => {
  try {
    let deptartments = [];
    let organizations = [];
    let localUsers = [];

    const department = await models.Departments.findAll();
    if (!department) {
      throw new httpErrors.NotFound('Departments not found');
    }
    department.forEach(dept => deptartments.push(dept.dataValues.name));

    const organization = await models.Organizations.findAll();
    if (!organization) {
      throw new httpErrors.NotFound('Organizations not found');
    }
    organization.forEach(org => organizations.push(org.dataValues.name));

    const localUser = await models.LocalUsers.findAll();

    deptartments = [...new Set(deptartments)];
    organizations = [...new Set(organizations)];
    const userNames = localUser.map(item => item.userName);
    const emails = localUser.map(item => item.email);
    const names = localUser.map(item => item.name);

    res.status(200).send({
      organization: organizations,
      departments: deptartments,
      userNames,
      emails,
      names
    });
  } catch (error) {
    next(error);
  }
};

const handleBulkdelete = async (req, res, next) => {
  try {
    const { allUser, selectedGroups, allUsersExcept } = req.body;
    let groups;
    if (allUser) {
      if (allUsersExcept.length > 0) {
        groups = await models.CustomUserGroups.findAll({
          where: {
            id: { [Op.notIn]: allUsersExcept }
          }
        });
      } else {
        groups = await models.CustomUserGroups.findAll();
      }
    } else if (!allUser && selectedGroups.length > 0) {
      groups = await models.CustomUserGroups.findAll({
        where: {
          id: { [Op.in]: selectedGroups }
        }
      });
    } else {
      res.status(400).send('Invalid request. Provide valid input data.');
    }

    const groupsToDeleteIds = groups.map(group => group.id);
    const groupsRawQueryIds = groups.map(group => group.raw_query_id);
    const groupsSqlQueryIds = groups.map(group => group.sql_query_id);

    await models.CustomUserGroupsDetail.destroy({
      where: {
        group_id: { [Op.in]: groupsToDeleteIds }
      }
    });

    await models.UserGroupsPivot.destroy({
      where: {
        group_id: { [Op.in]: groupsToDeleteIds }
      }
    });

    await models.CustomUserGroups.destroy({
      where: {
        id: { [Op.in]: groupsToDeleteIds }
      }
    });

    await models.UserGroupRawQuery.destroy({
      where: {
        id: { [Op.in]: groupsRawQueryIds }
      }
    });

    await models.UserGroupSQLQuery.destroy({
      where: {
        id: { [Op.in]: groupsSqlQueryIds }
      }
    });

    res.status(200).json({ msg: 'success' });
  } catch (error) {
    if (error.name === 'SequelizeForeignKeyConstraintError') {
      res
        .status(500)
        .json({ message: 'Remove user group(s) from entitlement(s) first' });
    } else {
      next(error);
    }
  }
};

const handleBulkdeleteCurrentPage = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    const selected = req.body;
    if (selected.length > 0) {
      groups = await models.CustomUserGroups.findAll({
        where: {
          id: { [Op.in]: selected }
        },
        transaction
      });
    } else {
      res.status(400).send('Invalid request. Provide valid input data.');
    }

    const groupsToDeleteIds = groups.map(group => group.id);
    const groupsRawQueryIds = groups.map(group => group.raw_query_id);
    const groupsSqlQueryIds = groups.map(group => group.sql_query_id);

    await models.CustomUserGroupsDetail.destroy({
      where: {
        group_id: { [Op.in]: groupsToDeleteIds }
      },
      transaction
    });

    await models.UserGroupsPivot.destroy({
      where: {
        group_id: { [Op.in]: groupsToDeleteIds }
      },
      transaction
    });

    await models.CustomUserGroups.destroy({
      where: {
        id: { [Op.in]: groupsToDeleteIds }
      },
      transaction
    });

    await models.UserGroupRawQuery.destroy({
      where: {
        id: { [Op.in]: groupsRawQueryIds }
      },
      transaction
    });

    await models.UserGroupSQLQuery.destroy({
      where: {
        id: { [Op.in]: groupsSqlQueryIds }
      },
      transaction
    });

    await transaction.commit();
    res.status(200).json({ msg: 'success' });
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};
const paginationGroups = async (req, res, next) => {
  const { limit, sort, dir, offset } = req.query;
  const sanitizedDir = dir && dir.toLowerCase() === 'asc' ? 'ASC' : 'DESC';
  const sanitizedLimit = parseInt(limit, 10) || 10;
  const sanitizedOffset = parseInt(offset, 10) || 0;

  const transaction = await models.sequelize.transaction();
  try {
    const customUserGroupsQuery = `
    SELECT * FROM CustomUserGroups customUserGroups LEFT JOIN 
    (SELECT group_id, COUNT(*) AS record_count FROM UserGroupsPivot
    GROUP BY group_id) AS group_counts ON
    customUserGroups.id = group_counts.group_id
    ORDER BY :querySort ${sanitizedDir}
    LIMIT :limit
    OFFSET :offset
    `;

    const customUserGroups = await models.sequelize.query(customUserGroupsQuery, {
      replacements: {
        querySort : sort,
        limit: sanitizedLimit,
        offset: sanitizedOffset * sanitizedLimit, // Calculate the actual offset
      },
      type: models.Sequelize.QueryTypes.SELECT,
    });
  
    const countQuery = `
      SELECT COUNT(customUserGroups.id) AS users_groups_count
      FROM CustomUserGroups customUserGroups
      LEFT JOIN (
        SELECT group_id, COUNT(*) AS record_count
        FROM UserGroupsPivot
        GROUP BY group_id
      ) AS group_counts ON customUserGroups.id = group_counts.group_id
    `;
    const count = await models.sequelize.query(countQuery, {
      type: models.Sequelize.QueryTypes.SELECT,
    });

    await transaction.commit();
    res.json({
      groups: customUserGroups,
      count: count[0]?.users_groups_count
    });
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const bulkActions = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    const {
      selectAll,
      selectedIds,
      organization,
      department,
      searchText,
      actionType: type,
      status,
      idp,
      verificationPassword,
      mfa
    } = req.body;
    const { id } = req.user;

    let options = {
      where: {
        id: {
          ...(selectAll
            ? {
                [Op.notIn]: selectedIds
              }
            : {
                [Op.in]: selectedIds
              })
        },
        [Op.or]: {
          name: {
            [Op.like]: `%${searchText.trim()}%`
          },
          email: {
            [Op.like]: `%${searchText.trim()}%`
          },
          userName: {
            [Op.like]: `%${searchText.trim()}%`
          }
        },
        ...(idp ? { identity_provider: idp } : {})
      },
      transaction
    };

    if (organization) {
      const myOrganization = await models.Organizations.findOne({
        where: {
          name: organization
        },
        attributes: ['id']
      });
      if (myOrganization) {
        const departmentOptions = {
          organization_id: myOrganization.id
        };
        if (department) {
          departmentOptions.id = department;
        }
        const myDepartments = await models.Departments.findAll({
          where: departmentOptions
        });
        if (myDepartments && myDepartments.length > 0) {
          options.where.department = {
            [Op.in]: myDepartments.map(department => department.id)
          };
        }
      }
    }

    if (type == 'send_qr_code') {
      const localUsers = await models.LocalUsers.findAll(options, {
        transaction
      });

      const clientOptions = {
        attributes: [
          'Email',
          'Name',
          'user',
          [
            models.sequelize.fn('max', models.sequelize.col('linOTP_token')),
            'linOTP_token'
          ],
          [
            models.sequelize.fn('max', models.sequelize.col('OTP_image_data')),
            'OTP_image_data'
          ],
          [models.sequelize.fn('max', models.sequelize.col('id')), 'id']
        ],
        where: {
          user: {
            [Op.in]: localUsers.map(_u => _u.userName)
          }
        },
        group: ['Email', 'Name', 'user']
      };

      let users = await models.Client.findAll(clientOptions, { transaction });
      users = await Promise.all(
        users.map(user => {
          if (user.OTP_image_data) return user;
          user.skip_otp_check = true;
          return user.generateOtp();
        })
      );

      sendOtpQrBulkUsers(users);

      for (const user of users) {
        insertIntoAdminLogs(getLocalUserSendOtp(user.user), req);
      }

      await transaction.commit();

      return res.json({
        message: 'User(s) will receive OTP Code within 2 minutes.'
      });
    } else if (type == 'deleteUser') {
      if (verificationPassword) {
        const user = await models.User.findOne({
          attributes: ['password'],
          where: {
            id
          }
        });
        let match;
        if (user) {
          if (!mfa) {
            match = await argon2.verify(user.password, verificationPassword);
          }
          if (match || mfa) {
            options = {
              ...options,
              where: {
                ...options.where,
                userName: {
                  ...options.where.userName,
                  [Op.notLike]: `%gwadmin%`
                }
              }
            };

            console.log('object :>> ');
            const localUsers = await models.LocalUsers.findAll(options, {
              transaction
            });

            const clientOptions = {
              where: {
                user: {
                  [Op.in]: localUsers.map(_u => _u.userName)
                }
              }
            };

            let clients = await models.Client.findAll(clientOptions, {
              transaction
            });

            if (clients.length != 0) {
              throw new httpErrors.Conflict(
                'One or more users have devices associated with them.'
              );
            }

            await models.LocalUsers.destroy(options, { transaction });
            bulkDeleteUser, await transaction.commit();

            for (const name of localUsers) {
              insertIntoAdminLogs(getLocalUserDeleteMessage(name.name), req);
            }

            return res.json({ message: 'User(s) deleted successfully.' });
          } else {
            throw new Error('Passwords do not match');
          }
        } else {
          throw new Error('User not found');
        }
      }
    } else if (type == 'enforcePasswordChange') {
      options = {
        ...options,
        where: {
          ...options.where,
          userName: {
            ...options.where.userName,
            [Op.notLike]: `%gwadmin%`
          }
        }
      };

      const localUsers = await models.LocalUsers.findAll(options, {
        transaction
      });

      BulkPasswordChangeQueue.addtoQueue({
        selectAll,
        selectedIds,
        organization,
        department,
        searchText,
        idp
      });

      for (const name of localUsers) {
        insertIntoAdminLogs(getLocalUserPassChangeMessage(name.name), req);
      }

      res.json({ message: 'Password change request sent.' });
    } else {
      const localUsers = await models.LocalUsers.findAll(options, {
        transaction
      });

      let updateUserObj = {};

      if (type == 'client_mfa') {
        updateUserObj = {
          ...updateUserObj,
          client_mfa: status,
          biometric_state: false,
          is_strong_mfa: false
        };
      }

      if (type == 'biometric_state') {
        const biometric_state_flag = localUsers.reduce((acc, cur) => {
          console.log('Client MFA: ', cur.client_mfa);
          return acc && cur.client_mfa;
        }, true);
        if (!biometric_state_flag) {
          throw new httpErrors.Conflict(
            'Client MFA not enabled for one or more users.'
          );
        }
        updateUserObj = {
          ...updateUserObj,
          biometric_state: status,
          is_strong_mfa: false
        };
      }

      if (type == 'is_strong_mfa') {
        const is_strong_mfa_flag = localUsers.reduce((acc, cur) => {
          console.log('Biometric MFA: ', cur.biometric_state);
          return acc && cur.biometric_state;
        }, true);
        if (!is_strong_mfa_flag) {
          throw new httpErrors.Conflict(
            'Biometric MFA not enabled for one or more users.'
          );
        }
        updateUserObj = {
          ...updateUserObj,
          is_strong_mfa: status
        };
      }

      const actions = {
        client_mfa: status => ({ client_mfa: status }),
        biometric_state: status => ({ biometric_state: status }),
        is_strong_mfa: status => ({ is_strong_mfa: status })
      };

      await models.LocalUsers.update({ ...updateUserObj }, options);

      if (type == 'client_mfa' && status) {
        const userIds = localUsers.map(localUser => localUser.id);
        sendOtpToUsers(userIds);
      }

      const [, affectedClients] = await Promise.all([
        models.Client.update(
          { ...actions[type](status) },
          {
            where: {
              user_id: {
                [Op.in]: localUsers.map(_u => _u.id)
              }
            },
            transaction
          }
        ),
        models.Client.findAll({
          attributes: ['id', 'client_type'],
          where: {
            user_id: {
              [Op.in]: localUsers.map(_u => _u.id)
            }
          },
          transaction
        })
      ]);

      const otherClients = affectedClients.filter(
        c => c.client_type == 'Client'
      );

      if (otherClients.length) {
        publishClients(actionType.UPDATED, {
          ids: otherClients.map(_ac => _ac.id),
          data: {
            ...actions[type](status)
          },
          type: 'Client'
        });
      }

      await transaction.commit();

      for (const name of localUsers) {
        insertIntoAdminLogs(getLocalUserMfaMessage(name.name), req);
      }

      res.json({ message: 'User(s) updated' });
    }
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const deleteCurrentPage = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    const { selectedIds } = req.body;

    let options = {
      where: {
        id: {
          [Op.in]: selectedIds
        }
      },
      transaction
    };

    options = {
      ...options,
      where: {
        ...options.where,
        userName: {
          ...options.where.userName,
          [Op.notLike]: `%gwadmin%`
        }
      }
    };

    const localUsers = await models.LocalUsers.findAll(options, {
      transaction
    });

    const clientOptions = {
      where: {
        user: {
          [Op.in]: localUsers.map(_u => _u.userName)
        }
      }
    };

    let clients = await models.Client.findAll(clientOptions, { transaction });

    if (clients.length != 0) {
      throw new httpErrors.Conflict(
        'One or more users have devices associated with them.'
      );
    }

    await models.LocalUsers.destroy(options, { transaction });
    await transaction.commit();

    for (const name of localUsers) {
      insertIntoAdminLogs(getLocalUserDeleteMessage(name.name), req);
    }

    return res.json({ message: 'User(s) deleted successfully.' });
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

const sendOtpToUsers = async userIds => {
  const setting = await models.ClientSetting.findOne({
    limit: 1,
    order: [['id', 'DESC']],
    attributes: { exclude: ['createdAt', 'updatedAt'] }
  });

  if (setting.mfa_server_state) {
    const clients = await models.Client.findAll({
      where: {
        user_id: userIds
      }
    });

    sendOtpQrBulkUsers(clients);
  }
};

const updateUserGroupFromUserId = async (
  userId,
  department,
  previousUserDepartment = ''
) => {
  const transaction = await models.sequelize.transaction();
  try {
    let userIds = [];
    let updateGroupResult;
    const userGroupsPivot = await models.UserGroupsPivot.findAll(
      {
        where: {
          user_id: userId,
          is_declarative: 1
        }
      },
      { transaction }
    );

    const userGroupsIds = userGroupsPivot.map(value => {
      return value.dataValues?.group_id;
    });

    const nonDeclarativeUsers = await models.UserGroupsPivot.findOne(
      {
        where: {
          user_id: userId,
          is_declarative: false
        }
      },
      { transaction }
    );

    await models.UserGroupsPivot.destroy(
      {
        where: {
          group_id: userGroupsIds,
          is_declarative: true
        }
      },
      { transaction }
    );

    const userGroupDetail = await models.CustomUserGroups.findAll(
      {
        where: {
          id: { [Op.in]: userGroupsIds }
        }
      },
      { transaction }
    );

    userGroupDetail.forEach(async (value, index) => {
      const sqlQuery = await models.UserGroupSQLQuery.findOne({
        where: {
          id: value?.dataValues?.sql_query_id
        }
      });

      let result = (
        await models.sequelize.query(sqlQuery?.dataValues?.query)
      )[0];
      console.log('qwerty', sqlQuery?.dataValues?.query);
      let ids = result.map(item => item.id);
      userIds = [...new Set(ids)];
      userIds.map(async (userId, index) => {
        await models.UserGroupsPivot.create({
          group_id: value?.dataValues?.id,
          user_id: userId
        });
      });
    });

    const departmentQuery = `
      SELECT departments.name  from Departments departments
      where departments.id = '${department}' `;

    let departmentResult = (await models.sequelize.query(departmentQuery))[0];

    console.log('departmentResult ', departmentResult[0]?.name);

    if (departmentResult[0]?.name) {
      const updateGroupQuery = `
      SELECT DISTINCT customUserGroupsDetail.group_id FROM CustomUserGroupsDetail customUserGroupsDetail
      INNER JOIN UserGroupsPivot userGroupsPivot On userGroupsPivot.group_id = customUserGroupsDetail.group_id 
      WHERE customUserGroupsDetail.label_value = '${departmentResult[0].name}' AND userGroupsPivot.is_declarative = True;
      `;

      console.log('updateGroupQuery ', updateGroupQuery);
      updateGroupResult = (await models.sequelize.query(updateGroupQuery))[0];

      console.log('updateGroupResult ', updateGroupResult);

      updateGroupResult.map(async (result, index) => {
        await models.UserGroupsPivot.create({
          group_id: result?.group_id,
          user_id: userId
        });
      });
    }

    await transaction.commit();

    if (
      (userGroupsIds?.length > 0 || updateGroupResult?.length > 0) &&
      department != previousUserDepartment
    ) {
      await publishUserGroups(actionType.UPDATED, {
        id: userGroupsIds[0] || updateGroupResult[0]?.group_id,
        new_users: userIds,
        del_users: [parseInt(userId)]
      });
    }
  } catch (error) {
    await transaction.rollback();
    throw new Error(error?.message);
  }
};

/*
    ############################## Utility Functions ##################################
*/

const getUserNameFromId = id => {
  return new Promise((resolve, reject) => {
    models.LocalUsers.findOne({
      where: {
        id
      },
      attributes: ['userName']
    })
      .then(userData => {
        if (userData) {
          resolve(userData.dataValues.userName);
        }
      })
      .catch(error => {
        reject(error);
      });
  });
};

const checkUserExistInLdapOrNot = async (userName, email) => {
  const filter = `(|(uid=${userName})(samAccountName=${userName}))`;
  const searchFilter = `(&(objectClass=person)${filter})`;
  let isUserFoundInAnyLdap = false;

  try {
    const ldapConfigs = await models.Ldap.findAll();

    for (const ldap of ldapConfigs) {
      const ldapManager = new LdapManager();
      const ldapConnection = await ldapManager.reuseOrCreateLdapConnection(ldap);
      const options = {
        scope: 'sub',
        filter: searchFilter,
        attributes: ['cn', 'mail', 'uid', 'userPrincipalName']
      };

      const isLdapUserFound = await new Promise((resolve, reject) => {
        ldapManager.do('search', ldapManager.settings.LDAP_dc, options, (err, res) => {
          if (err) {
            console.log(err && (err.message || err));
            return reject(err);
          }

          let isUserFound = false;
          res.on('searchEntry', entry => {
            if (entry) {
              isUserFound = true;
            }
          });

          res.on('error', err => {
            console.log(err && (err.message || err));
            return reject(err);
          });

          res.on('end', result => {
            return resolve(isUserFound);
          });
        });
      });

      if (isLdapUserFound) {
        isUserFoundInAnyLdap = true;
        break; // Exit loop if user found in any LDAP
      }
    }
  } catch (err) {
    console.log(err && (err.message || err));
  }

  return isUserFoundInAnyLdap;
};

const uploadCsv = async (req, res) => {
  try {
    const REQUIRED_COLUMNS = ['Name', 'Username', 'Email', 'Department', 'Contact'];
    const fileData = req.body.fileData;

    let file = fileData;
    let encodedFileData = file.data.split(';base64,').pop();

    const csvData = Buffer.from(encodedFileData, 'base64');

      let rows = [];
      let headers = [];
      let missingColumns = [];
      let usernames = new Set();
      let emails = new Set();
      await new Promise((resolve, reject) => {
        csv.parseString(csvData, { headers: true })
        .on('headers', receivedHeaders => {
          headers = receivedHeaders;
          missingColumns = REQUIRED_COLUMNS.filter(col => !headers.includes(col));
          if (missingColumns.length > 0) {
            reject(`Missing required columns. ${missingColumns.join(', ')}. Required columns are: ${REQUIRED_COLUMNS.join(', ')}`);
          }
        })
        .on('data', row => {
          if (!row.Email) {
            reject(`Missing email field in row with username: ${row.Username}`);
          }
          if (usernames.has(row.Username)) {
            reject(`Duplicate username found: ${row.Username}`);
          } else {
            usernames.add(row.Username);
          }
          if (emails.has(row.Email)) {
            reject(`Duplicate email found: ${row.Email}`);
          } else {
            emails.add(row.Email);
          }
          rows.push(row);
        })
          .on('end', resolve) 
          .on('error', reject);
      });
  
      for (let i = 0; i < rows.length; i++) {
        try {
          await sleep(1000);
          await validateRow(rows[i]);
        } catch (error) {
          return res.status(400).json({ error: `Row ${i + 1}: ${error.message}` });
        }
      }
      for (const row of rows) {
        await createUser(row);
      }
  
      res.status(200).json({ message: 'CSV upload and user creation successful.' });
    } catch (error) {
      res.status(400).send(error);
    }
};

async function validateRow(row) {
  const contactPattern = /^(\+[0-9]{1,3}\s?\([0-9]{1,4}\)\s?[0-9]{3,4}\-?\–?[0-9]{4}|(\+[0-9]{1,3}|0)\d{10})$/;

  if (!row.Name) throw new Error('Name is missing');
  if (!row.Username) throw new Error('Username is missing');
  if (!row.Email) throw new Error('Email is missing');
  if (!isValidEmail(row.Email)) throw new Error('Invalid email format');

  const emailDomain = row.Email.split('@')[1];
  if (!emailDomain) throw new Error('Invalid email domain');

  const validDomain = await models.ValidDomain.findOne({ where: { domain: emailDomain } });
  if (!validDomain) throw new Error(`Email domain '${emailDomain}' is not valid`);

  if (!row.Department) throw new Error('Department is missing');
  if (!row.Contact) throw new Error('Contact is missing');
  if (!contactPattern.test(padContactNumber(row.Contact))){
    throw new Error('Invalid contact number format');
  } else{
    row.Contact = padContactNumber(row.Contact)
  }

  const existingUser = await models.LocalUsers.findOne({ where: { email: row.Email } });
  if (existingUser) throw new Error('User with this email already exists');

  const existingUsername = await models.LocalUsers.findOne({ where: { userName: row.Username } });
  if (existingUsername) throw new Error('User with this username already exists');

  const department = await models.Departments.findOne({ where: { name: row.Department } });
  if (!department) throw new Error(`Department '${row.Department}' not found`);

  const organization = await models.Organizations.findOne({
    where: { name: row.Organization },
    include: {
      model: models.Departments,
      where: { name: row.Department }
    }
  });

  if (!organization) {
    throw new Error(`Organization '${row.Organization}' does not contain department '${row.Department}'`);
  }
  
}

function padContactNumber(contact) {
  contact = contact.toString();
  if (!contact.startsWith('+')) {
    if (contact.length === 12) {
      contact = '+' + contact;
    } else if (!contact.startsWith('0') && contact.length === 10) { 
      contact = contact.padStart(11, '0');
    }
  }
  console.log('contact :>> ', contact);
  return contact;
}

function isValidEmail(email) {
  const emailRegex = /\S+@\S+\.\S+/;
  return emailRegex.test(email);
}

function sleep(ms) {
  return new Promise(resolve => setTimeout(resolve, ms));
}

async function createUser(row) {
  const password = generateRandomPassword();
  const hashedPassword = crypto.createHash('sha256').update(password).digest('hex');
  
  const department = await models.Departments.findOne({ where: { name: row.Department } });
  if (!department) throw new Error(`Department '${row.Department}' not found`);

  await models.LocalUsers.create({
      name: row.Name,
      email: row.Email,
      userName: row.Username,
      contact: row.Contact,
      department: department.id,
      password: hashedPassword,
      change_pass: true,
      password_updated_at: moment().format(),
      identity_provider: 'Local User',
      runtime_bind_id: Math.floor(Date.now() / 1000) * 1000 + Math.floor(Math.random() * 1000)
  });

  helper.sendNewLocalUserMail(
    row.Email, 
    row.Name, 
    row.Username, 
    password
  );
}


const getUserGroupSearch = async (req, res, next) => {
  try {
    let {
      limit = 10,
      offset = 0,
      sort = 'group_name',
      dir = 'asc',
      searchTerm = ''
    } = req.query;

    const sanitizedDir = dir.toLowerCase() === 'asc' ? 'ASC' : 'DESC';
    const sanitizedLimit = parseInt(limit, 10) || 10;
    const sanitizedOffset = parseInt(offset, 10) || 0;
 
    const query = `
      SELECT cu.*, IFNULL(gc.record_count, 0) AS totalRecordsCount
      FROM CustomUserGroups cu
      LEFT JOIN (
          SELECT group_id, COUNT(*) AS record_count
          FROM UserGroupsPivot
          GROUP BY group_id
      ) AS gc ON cu.id = gc.group_id
      ${searchTerm
        ? `WHERE cu.group_name LIKE :searchTerm`
        : ''}
      ORDER BY :querySort ${sanitizedDir}
      LIMIT :limit OFFSET :offset
    `;

    const userGroups = await models.sequelize.query(query, {
      replacements: {
        searchTerm: `%${searchTerm}%`, // Use wildcards for LIKE clause
        limit: sanitizedLimit,
        offset: sanitizedOffset,
        querySort: sort
      },
      type: models.Sequelize.QueryTypes.SELECT
    });

    res.json(userGroups);
  } catch (error) {
    next(error);
  }
};

module.exports = {
  getLocalUsers,
  getLocalUser,
  getLocalUserServices,
  resetPassword,
  createLocalUser,
  searchClient,
  updateLocalUser,
  deleteLocalUser,
  sendResetLink,
  token,
  checkValidToken,
  updatePassword,
  bulkPasswordChange,
  bulkDeleteUser,
  deviceAssociationWithUser,
  bulkActions,
  createCustomUserGroups,
  getCustomUserGroups,
  makeLDAPChanges,
  getUserDeviceRelatedPolicies,
  getAvailableLocalUsers,
  createNewClient,
  getLocalUsersSearch,
  customGroupData,
  updateCustomGroup,
  deleteCustomGroup,
  deleteUserFromCustomGoup,
  getCustomUserGroupById,
  getUsersAndUserGroups,
  updateUserGroup,
  handleBulkdelete,
  handleBulkdeleteCurrentPage,
  paginationGroups,
  getUserGroupSearch,
  getLocalUsersFromDropDown,
  deleteCurrentPage,
  uploadCsv
};
