var httpErrors = require('http-errors');
var models = require('../models');

const { 
  getOrganizationUpdateMessage,
  getOrganizationCreateMessage,
  getOrganizationDeleteMessage,
  getDepartmentAddMessage,
  getDepartmentDeleteMessage
} = require('../config/adminLogs.config');

const { insertIntoAdminLogs } = require('../routes/common/adminLogs');

const getOrganizationNames = async (req, res, next) => {
  const organizations = await models.Organizations.findAll();
  let nameArr = [];
  if(organizations) {
    organizations.map(org => {
      nameArr.push(org.name);
    })
    res.json(nameArr);
  } 
}

const getOrganizations = async (req, res, next) => {
    try {
        const {
          limit,
          offset,
          sort = 'id',
          dir = 'desc',
          searchTerm='',
        } = req.query;
      
        const options = {
          limit: +limit,
          offset: +offset,
          order: []
        };
    
        if(searchTerm.trim()){
          options.where = {
            [Op.or]: {
              name: {
                [Op.like]: `%${searchTerm.trim()}%`
              }
            }
          }
        }

        const allowedSortValues = [
          'name',
          'id'
        ];
      
        if (allowedSortValues.includes(sort)) {
          options.order.push([models.sequelize.literal(sort), dir]);
        }
      
        const result = await models.Organizations.paginate(options);
        res.json(result);
    
    } catch (error) {
        next(error);
    }
}

const getOrganization = async (req, res, next) => {
  try {
      const { id } = req.params;
     
      const organization = await models.Organizations.findOne({
        where: {
          id
        }
      });

      if (!organization) {
        throw new httpErrors.NotFound('Organization not found');
      }

      res.json(organization);
  
  } catch (error) {
      next(error);
  }
}

const getOrganizationByName = async (req, res, next) => {
  try {
      const { name } = req.params;

      const organization = await models.Organizations.findOne({
        where: {
          name
        }
      });

      if (!organization) {
        throw new httpErrors.NotFound('Organization not found');
      }

      res.json(organization);
  
  } catch (error) {
      next(error);
  }
}

const getDepartments = async (req, res, next) => {
  try {
      const {
        organization = 0,
        limit,
        offset,
        sort = 'id',
        dir = 'desc',
        searchTerm=''
      } = req.query;
    
      const options = {
        limit: +limit,
        offset: +offset,
        order: []
      };
  
      if(searchTerm.trim()){
        options.where = {
          [Op.or]: {
            name: {
              [Op.like]: `%${searchTerm.trim()}%`
            }
          }
        }
      }

      if(organization != 0) {
        options.where = {
          ...options.where,
          organization_id: organization
        }
      }

      const allowedSortValues = [
        'name',
        'id'
      ];
    
      if (allowedSortValues.includes(sort)) {
        options.order.push([models.sequelize.literal(sort), dir]);
      }
    
      const result = await models.Departments.paginate(options);
      res.json(result);
  
  } catch (error) {
      next(error);
  }
}

const getAllDepartments = async (req, res, next) => {
  try {
      const {
        organization = 0,
      } = req.query;
    
      const options = {};

      if(organization != 0) {
        options.where = {
          ...options.where,
          organization_id: organization
        }
      }
    
      const result = await models.Departments.findAll(options);
      res.json({ data: result});
  
  } catch (error) {
      next(error);
  }
}

const getDepartment  = async (req, res, next) => {
  try {
      const department = await models.Departments.findOne({
        where: {
          id: req.params.id
        }
      });

      if(!department) {
        throw new httpErrors.NotFound('Department not found');
      }

      res.json(department);
  
  } catch (error) {
      next(error);
  }
}

const createDepartment = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {

      const { 
          organizationId, 
          name
      } = req.body.department;

      const duplicateDepartment = await models.Departments.findOne({
        where: {
          organization_id: organizationId,
          name
        }
      });

      if(duplicateDepartment) {
        throw new httpErrors.NotFound('Department already exists in this organization.');
      }

      const createdDepartment = await models.Departments
      .create({ 
          name,
          organization_id: organizationId
        },{ 
            transaction
        }
      );

      const organization = await models.Organizations.findOne({
        where: {
          id: organizationId
        }
      });

      organization.department_count = +organization.department_count + 1;

      await organization.save({transaction});

      await transaction.commit();

      insertIntoAdminLogs(
        getDepartmentAddMessage(name, organization.name),
        req
      )

      res.json(createdDepartment);

  } catch (error) {
      await transaction.rollback();
      next(error);
  }
} 

const createOrganization = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {

    const { departments } = req.body

      const { 
          department_count, 
          name
      } = req.body.organization;

      const organization = await models.Organizations.findOne({
        where: {
          name
        }
      });

      if (organization) {
        throw new httpErrors.NotFound('Organization name already in use.');
      }

      const createdOrganization = await models.Organizations.create({ 
          name,
          department_count
        },{ 
            transaction
        }
      );

      await createdOrganization.save({transaction});

      if(departments && Array.isArray(departments)) {
        let departmentsArray = []
        departments.forEach(department => {
          if(department.name) {
            departmentsArray.push({
              name: department.name,
              organization_id: createdOrganization.id
            })
          }
        })

        if(departmentsArray.length > 0) {
          const departments = await models.Departments.bulkCreate(departmentsArray, { transaction })
        }
      }

      await transaction.commit();

      insertIntoAdminLogs(
        getOrganizationCreateMessage(name),
        req
      )

      res.json(createdOrganization);

  } catch (error) {
      await transaction.rollback();
      next(error);
  }
}

const deleteOrganization = async (req, res, next) => {
  const transaction =  await models.sequelize.transaction();
  try {
      const organization = await models.Organizations.findOne({
        where: {
          id: req.params.id,
        }
      });

      if (!organization) {
        throw new httpErrors.NotFound('Organization not found');
      }

      const departments = await models.Departments.findAll({
        where: {
          organization_id: req.params.id,
        }
      });

      if(departments) {
        const departmentIds = departments.map(department => department.id)
        const users = await models.LocalUsers.findAll({
          where: {
            department: departmentIds
          }
        });

        const message = 'There are users/devices belonging to this organization'

        if(users.length > 0) {
          throw new httpErrors(400, message)
        }

        const devices = await models.DeviceInventories.findAll({
          where: {
            department: departmentIds
          }
        });

        if(devices.length > 0) {
          throw new httpErrors(400, message)
        }
      }

      await Promise.all(departments.map(department => department.destroy()))

      await organization.destroy({transaction});

      await transaction.commit();

      insertIntoAdminLogs(
        getOrganizationDeleteMessage(organization.name),
        req
      )

      res.json('Organization deleted successfully');

  } catch (error) {
      await transaction.rollback();
      next(error);
  }
}

const deleteDepartment = async (req, res, next) => {
  const transaction =  await models.sequelize.transaction();
  try {
      const dept = await models.Departments.findOne({
        where: {
          id: req.params.id,
        }
      });

      if (!dept) {
        throw new httpErrors.NotFound('Department not found');
      }

      const org = await models.Organizations.findOne({
        where: {
          id: dept.organization_id
        }
      });

      org.department_count = +org.department_count - 1;

      const users = await models.LocalUsers.findAll({
        where: {
          department: dept.id
        }
      });

      const message = 'There are users/devices belonging to this department'

      if(users.length > 0) {
        throw new httpErrors(400, message)
      }
      
      const devices = await models.DeviceInventories.findAll({
        where: {
          department: dept.id
        }
      });

      if(devices.length > 0) {
        throw new httpErrors(400, message)
      }
      

      await org.save({transaction});

      await dept.destroy({transaction});

      await transaction.commit();

      insertIntoAdminLogs(
        getDepartmentDeleteMessage(dept.name, org.name),
        req
      )

      res.json('Department deleted successfully');

  } catch (error) {
      await transaction.rollback();
      next(error);
  }
}

const updateDepartment = async (req, res, next) => {
  const transaction =  await models.sequelize.transaction();
  try {
      const department = await models.Departments.findOne({
        where: {
          id: req.params.id,
        }
      });

      if (!department) {
        throw new httpErrors.NotFound('Department not found');
      }

      const { name } = req.body;

      department.name = name;

      await department.save({transaction});

      await transaction.commit();

      res.json('Department updated successfully');

  } catch (error) {
      await transaction.rollback();
      next(error);
  }
}

const updateOrganization = async (req, res, next) => {
  const transaction =  await models.sequelize.transaction();
  try {
      const organization = await models.Organizations.findOne({
        where: {
          id: req.params.id,
        }
      });

      if (!organization) {
        throw new httpErrors.NotFound('Organization not found');
      }

      const { name } = req.body;

      organization.name = name;

      await organization.save({transaction});

      await transaction.commit();

      insertIntoAdminLogs(
        getOrganizationUpdateMessage(name),
        req
      )

      res.json('Organization updated successfully');

  } catch (error) {
      await transaction.rollback();
      next(error);
  }
}

module.exports = {
  getOrganizations,
  deleteOrganization,
  getDepartments,
  createDepartment,
  deleteDepartment,
  updateDepartment,
  getOrganization,
  updateOrganization,
  createOrganization,
  getDepartment,
  getOrganizationNames,
  getOrganizationByName,
  getAllDepartments
};
