const models = require('../models');
const { convertToArray } = require('../utils/common');
const {
  actionType,
  publishRoutingPolicy,
  publishOsPolicy,
  publishProcessPolicy,
  publishAccessPolicy,
  publishDataVolumePolicy,
  publishSecurityPolicy,
  publishSoftwarePolicy
} = require('../routes/common/event-publisher');

const httpErrors = require('http-errors');
const _ = require('lodash');
const { Op, cls } = require('sequelize');
const createHttpError = require('http-errors');
const { insertIntoAdminLogs } = require('../routes/common/adminLogs');
const {
  getPolicyCreateMessage,
  policyTypes,
  getPolicyEditMessage,
  getPolicyDeleteMessage,
  getHostDeleteMessage,
  getPolicyEnableMessage,
  getPolicyDisableMessage
} = require('../config/adminLogs.config');
const DevicePolicies = require('sequelize/lib/model');
const Client = require('../models/client');
const queryTypes = require('sequelize/lib/query-types');
const { job } = require('../cronjobs/device_validation');

/**
 * Get the list of polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getPolicies = async (req, res, next) => {
  try {
    let {
      limit,
      offset,
      sort = 'name',
      dir = 'asc',
      name,
      clients = [],
      allClientPolicies,
      policy = ''
    } = req.query;

    const options = {
      where: {},
      limit,
      offset,
      order: []
      // distinct: true
    };

    const allowedSortValues = [
      'name',
      'access_restriction',
      'access_start_time',
      'access_end_time'
    ];

    if (allowedSortValues.includes(sort)) {
      options.order.push([sort, dir]);
    }

    if (name) {
      options.where = {
        ...options.where,
        name: {
          [Op.like]: '%' + name + '%'
        }
      };
    }

    if (policy != '') {
      if (policy == 'device') policy = 'Device';
      else if (policy == 'user-agent') policy = 'User Agent';

      options.where = {
        ...options.where,
        policy_type: {
          [Op.like]: '%' + policy + '%'
        }
      };
    }

    let clientIds = [];
    if (clients && clients.length) {
      clientIds = convertToArray(clients).map(id => parseInt(id, 10)).filter(id => !isNaN(id));
      
      options.where[Op.or] = [
        { all_clients: true },
        models.sequelize.literal(`
          EXISTS (
            SELECT 1 
            FROM ClientPolicies 
            INNER JOIN Clients ON ClientPolicies.client_id = Clients.id
            WHERE ClientPolicies.policy_id = Entitlements.id 
            AND Clients.id IN (?)
            AND ClientPolicies.policy_type_id = 1
          ) OR EXISTS (
            SELECT 1 
            FROM GroupClients 
            INNER JOIN ClientGroupPolicies ON ClientGroupPolicies.client_group_id = GroupClients.groupId
            INNER JOIN Clients ON GroupClients.clientId = Clients.id
            WHERE ClientGroupPolicies.policy_id = Entitlements.id 
            AND Clients.id IN (?)
            AND ClientGroupPolicies.policy_type_id = 1
          )
        `)
      ];

      // When executing, provide replacements
      options.replacements = [clientIds, clientIds];
    }

    if (allClientPolicies == 'true') {
      options.where.all_clients = true;
    }

    const result = await models.Entitlements.paginate({
      ...options
    });

    const enrichedData = await Promise.all(result.data.map(async (item) => {
      const [clients, services, users, groups, serviceGroups, userGroups] = await Promise.all([
        models.sequelize.query(
          'SELECT Name FROM Clients WHERE id IN (SELECT client_id FROM ClientPolicies WHERE policy_id = ? AND policy_type_id = 1)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT Name FROM Services WHERE id IN (SELECT service_id FROM ServicePolicies WHERE policy_id = ? AND policy_type_id = 1)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name as Name FROM LocalUsers WHERE id IN (SELECT user_id FROM UserPolicies WHERE policy_id = ? AND policy_type_id = 1)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name FROM `Groups` WHERE id IN (SELECT client_group_id FROM ClientGroupPolicies WHERE policy_id = ? AND policy_type_id = 1)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name FROM ServiceGroups WHERE id IN (SELECT service_group_id FROM ServiceGroupPolicies WHERE policy_id = ? AND policy_type_id = 1)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT group_name as name FROM CustomUserGroups WHERE id IN (SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ? AND policy_type_id = 1)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        )
      ]);
    
      return {
        ...item.dataValues,
        Clients: clients || [],
        Services: services || [],
        Groups: groups || [],
        ServiceGroups: serviceGroups || [],
        UserGroups: userGroups || [],
        Users: users || []
      };
    }));
    
    result.data = enrichedData;

    res.json(result);
  } catch (error) {
    next(error);
  }
};

module.exports.verifyPolicies = async (req, res, next) => {
  try {
    const {
      userId,
      serviceGroupId,
      selectedIds,
      gatewayGroupId,
      gatewayId,
      clientGroupId,
      allServices
    } = req.query;

    let policyExist = false;
    let servicesIds = [];
    if (selectedIds.length > 0) {
      servicesIds = selectedIds.split(',').map(str => parseInt(str, 10));
    }

    if (userId) {
      const userPolicies = await models.sequelize
        .query(`SELECT * FROM LocalUsers lu 
      INNER JOIN UserPolicies up on up.user_id = lu.id 
      WHERE lu.id = ${userId}`);

      if (userPolicies[0].length > 0) {
        policyExist = true;
      }
    } else if (serviceGroupId) {
      const serviceGroupPolicies = await models.sequelize
        .query(`SELECT * FROM ServiceGroups sg 
      INNER JOIN ServiceGroupPolicies sgp on sgp.service_group_id = sg.id 
      WHERE sg.id = ${serviceGroupId}`);

      if (serviceGroupPolicies[0].length > 0) {
        policyExist = true;
      }
    } else if (servicesIds.length > 0) {
      const servicesPolicies = await models.sequelize
        .query(`SELECT * FROM Services s 
      INNER JOIN ServicePolicies sp on sp.service_id = s.id 
      WHERE s.id IN (${servicesIds.join(',')})`);

      if (servicesPolicies[0].length > 0) {
        policyExist = true;
      }
    } else if (gatewayGroupId) {
      const gatewayGroupPolicies = await models.sequelize
        .query(`SELECT * FROM GatewayGroups gg 
      INNER JOIN GatewayGroupPolicies ggp on gg.id = ggp.gateway_group_id 
      WHERE gg.id = ${gatewayGroupId}`);

      if (gatewayGroupPolicies[0].length > 0) {
        policyExist = true;
      }
    } else if (gatewayId) {
      const gatewayPolicies = await models.sequelize
        .query(`SELECT * FROM Gateways g 
      INNER JOIN GatewayPolicies gp on gp.gateway_id = g.id 
      WHERE g.id = ${gatewayId}`);

      if (gatewayPolicies[0].length > 0) {
        policyExist = true;
      }
    } else if (clientGroupId) {
      const clientGroupPolicies = await models.sequelize
        .query(`SELECT  * from \`Groups\` g 
      INNER JOIN ClientGroupPolicies cgp on cgp.client_group_id = g.id 
      WHERE g.id =  ${clientGroupId}`);

      if (clientGroupPolicies[0].length > 0) {
        policyExist = true;
      }
    }else if(allServices){
      const servicesPolicies = await models.sequelize
        .query(`SELECT * FROM Services s 
      INNER JOIN ServicePolicies sp on sp.service_id = s.id 
      `);

      if (servicesPolicies[0].length > 0) {
        policyExist = true
      }
    }

    res.status(200).json(policyExist);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getPolicyById = async (req, res, next) => {
  const { id } = req.params;
  try {
    let policy = await models.Entitlements.findOne({
      where: {
        id
      }
    });
    if (!policy) throw new httpErrors.NotFound('Policy not found');
    const clients = await models.sequelize.query(`
      SELECT id, Name, device_os FROM Clients
      WHERE id IN (
        SELECT client_id FROM ClientPolicies
        WHERE policy_id = ${id} AND policy_type_id = 1
      );
    `);
    const services = await models.sequelize.query(`
      SELECT id, Name FROM Services
      WHERE id IN (
        SELECT service_id FROM ServicePolicies
        WHERE policy_id = ${id} AND policy_type_id = 1
      );
    `);
    const groups = await models.sequelize.query(`
      SELECT id, name FROM \`Groups\`
      WHERE id IN (
        SELECT client_group_id FROM ClientGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 1
      );
    `);
    const serviceGroups = await models.sequelize.query(`
      SELECT id, name FROM ServiceGroups
      WHERE id IN (
        SELECT service_group_id FROM ServiceGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 1
      );
    `);
    const users = await models.sequelize.query(`
      SELECT id, Name, runtime_bind_id FROM LocalUsers
      WHERE id IN (
        SELECT user_id FROM UserPolicies
        WHERE policy_id = ${id} AND policy_type_id = 1
      );
    `);
    const userGroups = await models.sequelize.query(`
      SELECT id, group_name as name, runtime_bind_id, raw_query_id FROM CustomUserGroups
      WHERE id IN (
        SELECT user_group_id FROM UserGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 1
      );
    `);

    policy.dataValues = {
      ...policy.dataValues,
      Clients: clients[0],
      Services: services[0],
      Groups: groups[0],
      ServiceGroups: serviceGroups[0],
      Users: users[0],
      UserGroups: userGroups[0]
    };
    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

module.exports.getUserPolicyById = async (req, res, next) => {
  const { id } = req.params;
  const transaction = await models.sequelize.transaction();
  try {
    const policyIds = await models.sequelize.query(
      `SELECT e.id from LocalUsers lu 
      INNER JOIN UserGroupsPivot ugp on ugp.user_id = lu.id 
      INNER JOIN UserGroupPolicies ugp2 on ugp.group_id = ugp2.user_group_id 
      INNER JOIN UserPolicies up on up.user_id = lu.id 
      INNER JOIN Entitlements e on e.id = ugp2.policy_id 
      WHERE lu.id = ${id} OR e.all_users = 1
      UNION 
      SELECT e.id from LocalUsers lu 
      INNER JOIN UserGroupsPivot ugp on ugp.user_id = lu.id 
      INNER JOIN UserPolicies up on up.user_id = lu.id 
      INNER JOIN Entitlements e on e.id = up.policy_id
      WHERE lu.id = ${id} OR e.all_users = 1 `,
      { transaction }
    );

    await transaction.commit();
    return res.json(policyIds[0]);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};
/**
 * Create a policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.createPolicy = async (req, res, next) => {
  const {
    name,
    access_restriction,
    time_zone,
    access_start_time,
    access_end_time,
    months,
    days,
    is_enabled,
    client,
    clientGroup,
    service,
    serviceGroup,
    user,
    userGroup,
    all_users,
    all_clients,
    all_services,
    policy_type,
    event
  } = req.body;

  const policyObject = {
    name,
    access_restriction,
    time_zone,
    access_start_time,
    access_end_time,
    months: months.join(','),
    days: days.join(','),
    is_enabled,
    all_clients,
    all_services,
    all_users,
    policy_type,
    event
  };

  try {
    // Step 1: To check Either the Given Policy with name already Exist.
    const result = await models.Entitlements.findOne({
      where: {
        name
      }
    });
    if (result) {
      return res
        .status(400)
        .json({ message: 'Entitlement with this name already exists' });
    }
    // Step 2: IF Does not Exist create New one
    const transaction = await models.sequelize.transaction();
    const policy = await models.Entitlements.create(policyObject, {
      transaction
    });
    const CPolicies = client.map(c => {
      return {
        client_id: c,
        policy_type_id: 1,
        policy_id: policy.id
      };
    });
    const CGroupPolicies = clientGroup.map(c => {
      return {
        client_group_id: c,
        policy_type_id: 1,
        policy_id: policy.id
      };
    });
    const SPolicies = service.map(c => {
      return {
        service_id: c,
        policy_type_id: 1,
        policy_id: policy.id
      };
    });
    const SGroupPolicies = serviceGroup.map(c => {
      return {
        service_group_id: c,
        policy_type_id: 1,
        policy_id: policy.id
      };
    });
    const UPolicies = user.map(c => {
      return {
        user_id: c,
        policy_type_id: 1,
        policy_id: policy.id
      };
    });
    const UGroupPolicies = userGroup.map(c => {
      return {
        user_group_id: c,
        policy_type_id: 1,
        policy_id: policy.id,
        user_group_type: 1
      };
    });

    if (!all_clients) {
      await models.ClientPolicies.bulkCreate(CPolicies, { transaction });
      await models.ClientGroupPolicies.bulkCreate(CGroupPolicies, {
        transaction
      });
    }
    if (!all_services) {
      await models.ServicePolicies.bulkCreate(SPolicies, { transaction });
      await models.ServiceGroupPolicies.bulkCreate(SGroupPolicies, {
        transaction
      });
    }
    if (!all_users) {
      await models.UserPolicies.bulkCreate(UPolicies, { transaction });
      await models.UserGroupPolicies.bulkCreate(UGroupPolicies, {
        transaction
      });
    }
    await transaction.commit();

    await publishAccessPolicy(actionType.ADDED, {
      id: policy.id,
      mode: 2,
      data: {
        name: policy.name,
        access_restriction: +policy.access_restriction,
        access_start_time: policy.access_start_time,
        access_end_time: policy.access_end_time,
        months: policy.months,
        days: policy.days,
        all_clients: +policy.all_clients,
        all_services: +policy.all_services,
        all_users: +policy.all_users,
        policy_type: policy.policy_type,
        event : policy.event
      }
    });

    insertIntoAdminLogs(
      getPolicyCreateMessage(policy.name, policyTypes.ACCESS_POLICY),
      req
    );

    return res.json(policy);
  } catch (error) {
    console.log('ERROR : ', error);
    return error;
  }
};

/**
 * Create a policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.updatePolicy = async (req, res, next) => {
  const { id } = req.params;
  const {
    name,
    access_restriction,
    time_zone,
    access_start_time,
    access_end_time,
    months,
    days,
    is_enabled,
    client,
    clientGroup,
    service,
    serviceGroup,
    all_clients,
    all_services,
    policy_type,
    all_users,
    userGroup,
    user,
    event
  } = req.body;

  const transaction = await models.sequelize.transaction();
  try {
    const policy = await models.Entitlements.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');

    if(!(policy.name === name)) {
      const duplicateName = await models.Entitlements.findOne({
        where: {
          name
        }
      });

      if (duplicateName) {
        throw new httpErrors.BadRequest('Entitlement with this name already exists');
      }
    }

    let CPolicies = await models.sequelize.query(
      `select * from ClientPolicies where policy_id = ${id} and policy_type_id = 1`
    );

    let SPolicies = await models.sequelize.query(
      `select * from ServicePolicies where policy_id = ${id} and policy_type_id = 1`
    );

    let CGroupPolicies = await models.sequelize.query(
      `select * from ClientGroupPolicies where policy_id = ${id} and policy_type_id = 1`
    );

    let SGroupPolicies = await models.sequelize.query(
      `select * from ServiceGroupPolicies where policy_id = ${id} and policy_type_id = 1`
    );

    let UPolicies = await models.sequelize.query(
      `SELECT * FROM UserPolicies WHERE policy_id = ${id} AND policy_type_id = 1`
    );

    let UGroupPolicies = await models.sequelize.query(
      `SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ${id} AND policy_type_id = 1`
    );

    CPolicies = (CPolicies[0] || []).map(v => v.client_id);
    CGroupPolicies = (CGroupPolicies[0] || []).map(v => v.client_group_id);
    SPolicies = (SPolicies[0] || []).map(v => v.service_id);
    SGroupPolicies = (SGroupPolicies[0] || []).map(v => v.service_group_id);
    UPolicies = (UPolicies[0] || []).map(v => v.user_id);
    UGroupPolicies = (UGroupPolicies[0] || []).map(v => v.user_group_id);

    const clientToAdd = all_clients ? [] : _.difference(client, CPolicies);
    const clientToDelete = all_clients
      ? CPolicies
      : _.difference(CPolicies, client);
    const clientGroupToAdd = all_clients
      ? []
      : _.difference(clientGroup, CGroupPolicies);
    const clientGroupToDelete = all_clients
      ? CGroupPolicies
      : _.difference(CGroupPolicies, clientGroup);
    const serviceToAdd = all_services ? [] : _.difference(service, SPolicies);
    const serviceToDelete = all_services
      ? SPolicies
      : _.difference(SPolicies, service);
    const serviceGroupToAdd = all_services
      ? []
      : _.difference(serviceGroup, SGroupPolicies);
    const serviceGroupToDelete = all_services
      ? SGroupPolicies
      : _.difference(SGroupPolicies, serviceGroup);

    const userToAdd = all_users ? [] : _.difference(user, UPolicies);
    const userToDelete = all_users ? UPolicies : _.difference(UPolicies, user);
    const userGroupToAdd = all_users
      ? []
      : _.difference(userGroup, UGroupPolicies);
    const userGroupToDelete = all_users
      ? UGroupPolicies
      : _.difference(UGroupPolicies, userGroup);

    const [, , , , , , , , , , , , updatedPolicy] = await Promise.all([
      models.ClientPolicies.bulkCreate(
        clientToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 1,
          client_id: v
        })),
        { transaction }
      ),
      models.ClientPolicies.destroy({
        where: {
          client_id: {
            [Op.in]: clientToDelete
          },
          policy_id: policy.id,
          policy_type_id: 1
        },
        transaction
      }),
      models.ClientGroupPolicies.bulkCreate(
        clientGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 1,
          client_group_id: v
        })),
        { transaction }
      ),
      models.ClientGroupPolicies.destroy({
        where: {
          client_group_id: {
            [Op.in]: clientGroupToDelete
          },
          policy_id: policy.id,
          policy_type_id: 1
        },
        transaction
      }),

      models.ServicePolicies.bulkCreate(
        serviceToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 1,
          service_id: v
        })),
        { transaction }
      ),
      models.ServicePolicies.destroy({
        where: {
          service_id: {
            [Op.in]: serviceToDelete
          },
          policy_type_id: 1,
          policy_id: policy.id
        },
        transaction
      }),
      models.ServiceGroupPolicies.bulkCreate(
        serviceGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 1,
          service_group_id: v
        })),
        { transaction }
      ),
      models.ServiceGroupPolicies.destroy({
        where: {
          service_group_id: {
            [Op.in]: serviceGroupToDelete
          },
          policy_type_id: 1,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserPolicies.bulkCreate(
        userToAdd.map(c => {
          return {
            user_id: c,
            policy_type_id: 1,
            policy_id: policy.id
          };
        }, transaction)
      ),
      models.UserGroupPolicies.bulkCreate(
        userGroupToAdd.map(c => {
          return {
            user_group_id: c,
            policy_type_id: 1,
            policy_id: policy.id,
            user_group_type: 1
          };
        }, transaction)
      ),
      models.UserPolicies.destroy({
        where: {
          user_id: {
            [Op.in]: userToDelete
          },
          policy_type_id: 1,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserGroupPolicies.destroy({
        where: {
          user_group_id: {
            [Op.in]: userGroupToDelete
          },
          policy_type_id: 1,
          policy_id: policy.id
        },
        transaction
      }),

      await policy.update(
        {
          name,
          access_restriction,
          time_zone,
          access_start_time,
          access_end_time,
          months: months.join(','),
          days: days.join(','),
          is_enabled,
          all_clients,
          all_services,
          all_users,
          event
        },
        {
          transaction,
          updateData: {
            client,
            clientGroup,
            service,
            serviceGroup,
            all_clients,
            all_services,
            all_users
          }
        }
      )
    ]);

    await transaction.commit();

    await publishAccessPolicy(actionType.UPDATED, {
      id: updatedPolicy.id,
      data: {
        name: updatedPolicy.name,
        access_restriction: +updatedPolicy.access_restriction,
        access_start_time: updatedPolicy.access_start_time,
        access_end_time: updatedPolicy.access_end_time,
        months: updatedPolicy.months,
        days: updatedPolicy.days,
        all_clients: +updatedPolicy.all_clients,
        all_services: +updatedPolicy.all_services,
        all_users: +updatedPolicy.all_users
      },
      mode:
        (clientToAdd.length ||
          clientToDelete.length ||
          clientGroupToAdd.length ||
          clientGroupToDelete.length) &&
        (serviceToAdd.length ||
          serviceToDelete.length ||
          serviceGroupToAdd.length ||
          serviceGroupToDelete.length)
          ? 3
          : clientToAdd.length ||
            clientToDelete.length ||
            clientGroupToAdd.length ||
            clientGroupToDelete.length
          ? 1
          : 2
    });

    insertIntoAdminLogs(
      getPolicyEditMessage(policy.name, policyTypes.ACCESS_POLICY),
      req
    );
    return res.json(policy);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/**
 * Delete a policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.deletePolicy = async (req, res, next) => {
  const { id } = req.params;
  const transaction = await models.sequelize.transaction();
  try {
    const policy = await models.Entitlements.findOne({
      where: { id },
      transaction
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');

    await Promise.all([
      models.ClientPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 1 },
        transaction
      }),
      models.ServicePolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 1 },
        transaction
      }),
      models.ClientGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 1 },
        transaction
      }),
      models.ServiceGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 1 },
        transaction
      }),
      models.UserPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 1 },
        transaction
      }),
      models.UserGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 1 },
        transaction
      }),
      policy.destroy({ transaction })
    ]);

    await transaction.commit();

    await publishAccessPolicy(actionType.DELETED, {
      id: id,
      data: {},
      mode: 2
    });
    
    insertIntoAdminLogs(
      getPolicyDeleteMessage(policy.name, policyTypes.ACCESS_POLICY),
      req
    );
    return res.json(policy);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/**
 * Enable/Disable policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.setEnableValue = async (req, res, next) => {
  const { id, value } = req.params;
  try {
    const policy = await models.Entitlements.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');

    await policy.update(
      { is_enabled: value },
      {
        hooks: false
      }
    );

    let logMessage;
    if (value === 'true') {
      logMessage = getPolicyEnableMessage(
        policy.name,
        policyTypes.ACCESS_POLICY
      );
    } else {
      logMessage = getPolicyDisableMessage(
        policy.name,
        policyTypes.ACCESS_POLICY
      );
    }

    insertIntoAdminLogs(logMessage, req);

    await publishAccessPolicy(
      value == 'true' ? actionType.ADDED : actionType.DELETED,
      {
        id,
        mode: 3,
        data: {
          name: policy.name,
          access_restriction: +policy.access_restriction,
          access_start_time: policy.access_start_time,
          access_end_time: policy.access_end_time,
          months: policy.months,
          days: policy.days,
          all_clients: +policy.all_clients,
          all_services: +policy.all_services
        }
      }
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/** get all clients
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getAllClientsByName = async (req, res, next) => {
  try {
    const {
      limit = 10,
      offset = 0,
      dir = 'asc',
      sort = 'Name',
      name
    } = req.query;

    const options = {
      where: {},
      order: [],
      attributes: ['id', 'Name', 'device_os'],
      limit: +limit,
      offset: +offset * +limit || 0
    };

    const allowedSortValues = ['Name'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    if (name) {
      options.where = {
        ...options.where,
        name: {
          [Op.like]: '%' + name + '%'
        }
      };
    }

    const clients = await models.Client.findAll(options);

    return res.json(clients);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the services, clients, client groups and service groups
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getAllClientAndServices = async (req, res, next) => {
  const policyId = verifyPolicyId(req.query.policyId)
    ? req.query.policyId
    : false;
  const clientName = isSafeString(req.query.clientName)
    ? req.query.clientName
    : null;
  const serviceName = isSafeString(req.query.serviceName)
    ? req.query.serviceName
    : null;
  const clientType = req.query.clientType;
  let clientOffset = +req.query.clientOffset || 0;
  let serviceOffset = +req.query.serviceOffset || 0;
  const only_non_persistent = req.query.only_non_persistent;

  let withUser = req.query.withUser;
  const limit = 10;
  clientOffset = clientOffset * limit;
  serviceOffset = serviceOffset * limit;
  let where = 'where 1=1 ';
  // let clientWhere = where + ` and client_type like '%${clientType}%'`;
  let clientWhere = where;
  let clientGroupWhere = where;
  let serviceWhere = where;
  if (clientName && clientName !== 'null') {
    clientWhere += ` and (c.name LIKE '%${clientName}%' or c.id like '%${clientName}%') `;
    clientGroupWhere += ` and (g.name LIKE '%${clientName}%' or g.id like '%${clientName}%') `;
  }
  if (serviceName && serviceName !== 'null') {
    serviceWhere += ` and name LIKE '%${serviceName}%' `;
  }

  let clientLeftJoinQuery = '';
  let serviceLeftJoinQuery = '';
  let clientGroupLeftJoinQuery = '';
  let serviceGroupLeftJoinQuery = '';
  let userType = clientType == 'User' ? 'not' : '';
  let devicesJoinQuery = `left join DeviceInventories di on di.id = c.device_id`;
  if (clientType != '')
    clientWhere += ` and ifnull(di.device_type, "") ${userType} in ('IoT Bridge Client', 'Server Bridge Client', 'Thin Client (Bridge Client)') `;

  if (policyId) {
    clientLeftJoinQuery = `left join ClientPolicies cp on cp.policy_type_id = 1 and cp.client_id = c.id and cp.policy_id  = ${policyId}`;
    serviceLeftJoinQuery = `left join ServicePolicies sp on sp.policy_type_id = 1 and sp.service_id = s.id and sp.policy_id = ${policyId}`;
    serviceGroupLeftJoinQuery = `left join ServiceGroupPolicies sgp on sgp.policy_type_id = 1 and sgp.service_group_id = sg.id and sgp.policy_id = ${policyId}`;
    clientGroupLeftJoinQuery = `left join ClientGroupPolicies cgp on cgp.policy_type_id = 1 and cgp.client_group_id = g.id and cgp.policy_id = ${policyId}`;
  }

  const persistentWhere =
    only_non_persistent === 'false' ? '' : " and is_persistent ='false'";

  try {
    const query = `
    (SELECT c.id ,c.name,c.device_os,c.user_id, "client" as type from Clients c ${clientLeftJoinQuery} ${devicesJoinQuery} ${clientWhere} LIMIT ${clientOffset},${limit})
    UNION ALL
    (SELECT s.id ,s.name,s.port,"service" as temp, "service" as type from Services s ${serviceLeftJoinQuery} ${serviceWhere} ${persistentWhere} LIMIT ${serviceOffset},${limit})
    UNION ALL
    (SELECT g.id ,g.name, 0 as is_deleted,"clientGroup" as temp, "clientGroup" as type from \`Groups\` g ${clientGroupLeftJoinQuery} ${clientGroupWhere} LIMIT ${clientOffset},${limit})
    UNION ALL
    (SELECT sg.id ,sg.name, 0 as is_deleted, "serviceGroup" as temp, "serviceGroup" as type from ServiceGroups sg ${serviceGroupLeftJoinQuery} ${serviceWhere} ${persistentWhere} LIMIT ${serviceOffset},${limit})
    `;

    const result = await models.sequelize.query(query, {
      type: models.sequelize.QueryTypes.SELECT
    });
    return res.json(result);
  } catch (error) {
    next(error);
  }
};

/**
 * Its Temporary Fix by Bilal Akbar to remove onScrollPagination
 * Get the services, clients, client groups and service groups
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getAllClientAndServicesWithoutPagination = async (
  req,
  res,
  next
) => {
  const policyId = verifyPolicyId(req.query.policyId)
    ? req.query.policyId
    : false;
  const clientName = isSafeString(req.query.clientName)
    ? req.query.clientName
    : null;
  const serviceName = isSafeString(req.query.serviceName)
    ? req.query.serviceName
    : null;
  const clientType = req.query.clientType;
  let clientOffset = +req.query.clientOffset || 0;
  let serviceOffset = +req.query.serviceOffset || 0;
  let withUser = req.query.withUser;
  const limit = 10;
  clientOffset = clientOffset * limit;
  serviceOffset = serviceOffset * limit;
  let where = 'where 1=1 ';
  // let clientWhere = where + ` and client_type like '%${clientType}%'`;
  let clientWhere = where;
  let clientGroupWhere = where;
  let serviceWhere = where;
  if (clientName && clientName !== 'null') {
    clientWhere += ` and (c.name LIKE '%${clientName}%' or c.id like '%${clientName}%') `;
    clientGroupWhere += ` and (g.name LIKE '%${clientName}%' or g.id like '%${clientName}%') `;
  }
  if (serviceName && serviceName !== 'null') {
    serviceWhere += ` and name LIKE '%${serviceName}%' `;
  }

  let clientLeftJoinQuery = '';
  let serviceLeftJoinQuery = '';
  let clientGroupLeftJoinQuery = '';
  let serviceGroupLeftJoinQuery = '';
  let userType = clientType == 'User' ? 'not' : '';
  let devicesJoinQuery = `left join DeviceInventories di on di.id = c.device_id`;
  if (clientType != '')
    clientWhere += ` and ifnull(di.device_type, "") ${userType} in ('IoT Bridge Client', 'Server Bridge Client', 'Thin Client (Bridge Client)') `;

  if (policyId) {
    clientLeftJoinQuery = `left join ClientPolicies cp on cp.policy_type_id = 1 and cp.client_id = c.id and cp.policy_id  = ${policyId}`;
    serviceLeftJoinQuery = `left join ServicePolicies sp on sp.policy_type_id = 1 and sp.service_id = s.id and sp.policy_id = ${policyId}`;
    serviceGroupLeftJoinQuery = `left join ServiceGroupPolicies sgp on sgp.policy_type_id = 1 and sgp.service_group_id = sg.id and sgp.policy_id = ${policyId}`;
    clientGroupLeftJoinQuery = `left join ClientGroupPolicies cgp on cgp.policy_type_id = 1 and cgp.client_group_id = g.id and cgp.policy_id = ${policyId}`;
  }

  try {
    const query = `
   (SELECT c.id ,c.name,c.device_os,c.user_id, "client" as type from Clients c ${clientLeftJoinQuery} ${devicesJoinQuery} ${clientWhere})
   UNION ALL
   (SELECT s.id ,s.name,s.port,"service" as temp, "service" as type from Services s ${serviceLeftJoinQuery} ${serviceWhere})
   UNION ALL
   (SELECT g.id ,g.name, 0 as is_deleted,"clientGroup" as temp, "clientGroup" as type from \`Groups\` g ${clientGroupLeftJoinQuery} ${clientGroupWhere})
   UNION ALL
   (SELECT sg.id ,sg.name, 0 as is_deleted, "serviceGroup" as temp, "serviceGroup" as type from ServiceGroups sg ${serviceGroupLeftJoinQuery} ${serviceWhere})
   `;

    const result = await models.sequelize.query(query, {
      type: models.sequelize.QueryTypes.SELECT
    });
    return res.json(result);
  } catch (error) {
    next(error);
  }
};

/**
 * Enable/disable device policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */

module.exports.setDeviceEnableValue = async (req, res, next) => {
  const { id, value } = req.params;
  try {
    const policy = await models.DevicePolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');
    await policy.update(
      { is_enabled: value },
      {
        hooks: false
      }
    );

    let logMessage;
    if (value === 'true') {
      logMessage = getPolicyEnableMessage(
        policy.name,
        policyTypes.SECURITY_POLICY
      );
    } else {
      logMessage = getPolicyDisableMessage(
        policy.name,
        policyTypes.SECURITY_POLICY
      );
    }

    insertIntoAdminLogs(logMessage, req);

    await publishSecurityPolicy(
      value == 'true' ? actionType.ADDED : actionType.DELETED,
      {
        id,
        mode: 3,
        data: {
          name: policy.name,
          is_epp_client: +policy.is_epp_client,
          risk_score_threshold: +policy.risk_score_threshold,
          all_clients: +policy.all_clients,
          all_services: +policy.all_services,
          is_blocked: +policy.is_blocked,
          compliance: policy.compliance,
        }
      }
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Enable/disable software process policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */

module.exports.setSoftwareProcessEnableValue = async (req, res, next) => {
  const { id, value } = req.params;
  try {
    const policy = await models.SoftwareProcessPolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');
    await policy.update(
      { is_enabled: value },
      {
        hooks: false
      }
    );

    let logMessage;
    if (value === 'true') {
      logMessage = getPolicyEnableMessage(
        policy.name,
        policyTypes.PROCESS_POLICY
      );
    } else {
      logMessage = getPolicyDisableMessage(
        policy.name,
        policyTypes.PROCESS_POLICY
      );
    }

    insertIntoAdminLogs(logMessage, req);

    await publishProcessPolicy(
      value == 'true' ? actionType.ADDED : actionType.DELETED,
      {
        id,
        mode: 3,
        data: {
          name: policy.name,
          all_clients: +policy.all_clients,
          all_services: +policy.all_services,
          is_blocked: +policy.is_blocked
        }
      }
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the all os that exists
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */

module.exports.getAllOs = async (req, res, next) => {
  try {
    const { id } = req.query;
    let query = 'SELECT os from OsPolicies dp';
    const result = await models.sequelize
      .query(query, {
        type: models.sequelize.QueryTypes.SELECT
      })
      .map(o => o.os);
    return res.json(result);
  } catch (error) {
    next(error);
  }
};

module.exports.setOSEnableValue = async (req, res, next) => {
  const { id, value } = req.params;
  try {
    const policy = await models.OsPolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');
    await policy.update(
      { is_enabled: value },
      {
        hooks: false
      }
    );

    let logMessage;
    if (value === 'true') {
      logMessage = getPolicyEnableMessage(policy.name, policyTypes.OS_POLICY);
    } else {
      logMessage = getPolicyDisableMessage(policy.name, policyTypes.OS_POLICY);
    }

    insertIntoAdminLogs(logMessage, req);

    await publishOsPolicy(
      value == 'true' ? actionType.ADDED : actionType.DELETED,
      {
        id,
        mode: 3,
        data: {
          os: policy.os,
          os_version: policy.os_version,
          is_blocked: +policy.is_blocked,
          is_enabled: +policy.is_enabled,
          os_display_name: policy.os_display_name
        }
      }
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

module.exports.setSoftwareEnableValue = async (req, res, next) => {
  const { id, value } = req.params;
  try {
    const policy = await models.SoftwarePolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');
    await policy.update(
      { is_enabled: value },
      {
        hooks: false
      }
    );

    let logMessage;
    if (value === 'true') {
      logMessage = getPolicyEnableMessage(
        policy.name,
        policyTypes.SOFTWARE_POLICY
      );
    } else {
      logMessage = getPolicyDisableMessage(
        policy.name,
        policyTypes.SOFTWARE_POLICY
      );
    }

    insertIntoAdminLogs(logMessage, req);

    await publishSoftwarePolicy(
      value == 'true' ? actionType.ADDED : actionType.DELETED,
      {
        id,
        mode: 3,
        data: {
          name: policy.name,
          grace_period: policy.grace_period,
          block_services: policy.block_services,
          all_clients: +policy.all_clients,
          all_services: +policy.all_services,
          is_enabled: +policy.is_enabled,
          all_users: +policy.all_users
        }
      }
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the list of device polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getDevicePolicies = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchValue,
      clients = [],
      allClientPolicies = false
    } = req.query;

    const options = {
      where: {},
      limit,
      offset,
      order: []
      // distinct: true
    };

    const allowedSortValues = [
      'id',
      'antivirus',
      'is_epp_client',
      'is_blocked'
    ];

    if (allowedSortValues.includes(sort)) {
      options.order.push([sort, dir]);
    }

    if (searchValue) {
      options.where = {
        ...options.where,
        name: {
          [Op.like]: '%' + searchValue + '%'
        }
      };
    }

    let clientWhereClause = {};
   let clientIds = [];
    if (clients && clients.length) {
      clientIds = convertToArray(clients).map(id => parseInt(id, 10)).filter(id => !isNaN(id));
      
      options.where[Op.or] = [
        { all_clients: true },
        models.sequelize.literal(`
          EXISTS (
            SELECT 1 
            FROM ClientPolicies 
            INNER JOIN Clients ON ClientPolicies.client_id = Clients.id
            WHERE ClientPolicies.policy_id = DevicePolicies.id 
            AND Clients.id IN (?)
            AND ClientPolicies.policy_type_id = 2
          ) OR EXISTS (
            SELECT 1 
            FROM GroupClients 
            INNER JOIN ClientGroupPolicies ON ClientGroupPolicies.client_group_id = GroupClients.groupId
            INNER JOIN Clients ON GroupClients.clientId = Clients.id
            WHERE ClientGroupPolicies.policy_id = DevicePolicies.id 
            AND Clients.id IN (?)
            AND ClientGroupPolicies.policy_type_id = 2
          )
        `)
      ];

      // When executing, provide replacements
      options.replacements = [clientIds, clientIds];
    }

    if (allClientPolicies == 'true') {
      options.where.all_clients = true;
    }

    const result = await models.DevicePolicies.paginate({
      ...options
    });
    
    const enrichedData = await Promise.all(result.data.map(async (item) => {
      const [
        userGroups, 
        users, 
        clients, 
        services, 
        groups, 
        serviceGroups, 
        securityEndpoint
      ] = await Promise.all([
        models.sequelize.query(
          'SELECT group_name as name FROM CustomUserGroups WHERE id IN (SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ? AND policy_type_id = 2)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name as Name FROM LocalUsers WHERE id IN (SELECT user_id FROM UserPolicies WHERE policy_id = ? AND policy_type_id = 2)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT Name FROM Clients WHERE id IN (SELECT client_id FROM ClientPolicies WHERE policy_id = ? AND policy_type_id = 2)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT Name FROM Services WHERE id IN (SELECT service_id FROM ServicePolicies WHERE policy_id = ? AND policy_type_id = 2)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name FROM `Groups` WHERE id IN (SELECT client_group_id FROM ClientGroupPolicies WHERE policy_id = ? AND policy_type_id = 2)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name FROM ServiceGroups WHERE id IN (SELECT service_group_id FROM ServiceGroupPolicies WHERE policy_id = ? AND policy_type_id = 2)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          `SELECT GROUP_CONCAT(DISTINCT CONCAT(si.binding_name, ', ') SEPARATOR '') AS Name 
           FROM DevicePolicies dp 
           INNER JOIN SecurityPolicyPivots spp ON spp.policy_id = dp.id 
           INNER JOIN SoftwareInstances si ON si.id = spp.software_id 
           WHERE dp.id = ?`,
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        )
      ]);
    
      return {
        ...item.dataValues,
        UserGroups: userGroups || [],
        Users: users || [],
        Clients: clients,
        Services: services,
        Groups: groups,
        ServiceGroups: serviceGroups,
        securityEndpoint: securityEndpoint
      };
    }));
    
    result.data = enrichedData;

    res.json(result);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the list of device polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getDevicePolicyById = async (req, res, next) => {
  const { id } = req.params;
  try {
    let policy = await models.DevicePolicies.findOne({
      where: {
        id
      }
    });
    if (!policy) throw new httpErrors.NotFound('Device policy not found');
    const clients = await models.sequelize.query(`
      SELECT id, Name, device_os FROM Clients
      WHERE id IN (
        SELECT client_id FROM ClientPolicies
        WHERE policy_id = ${id} AND policy_type_id = 2
      );
    `);
    const services = await models.sequelize.query(`
      SELECT id, Name FROM Services
      WHERE id IN (
        SELECT service_id FROM ServicePolicies
        WHERE policy_id = ${id} AND policy_type_id = 2
      );
    `);
    const groups = await models.sequelize.query(`
      SELECT id, name FROM \`Groups\`
      WHERE id IN (
        SELECT client_group_id FROM ClientGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 2
      );
    `);
    const serviceGroups = await models.sequelize.query(`
      SELECT id, name FROM ServiceGroups
      WHERE id IN (
        SELECT service_group_id FROM ServiceGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 2
      );
    `);
    const userGroups = await models.sequelize.query(`
      SELECT id, group_name as name, runtime_bind_id, raw_query_id FROM CustomUserGroups
      WHERE id IN (
        SELECT user_group_id FROM UserGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 2
      );
    `);
    const users = await models.sequelize.query(`
      SELECT id, Name, runtime_bind_id FROM LocalUsers
      WHERE id IN (
        SELECT user_id FROM UserPolicies
        WHERE policy_id = ${id} AND policy_type_id = 2
      );
    `);

    let endpointSecurity = await models.SecurityPolicyPivots.findAll({
      where: {
        policy_id: id,
      }
    });

    endpointSecurity = endpointSecurity?.map(software => software?.software_id);

    policy.dataValues = {
      ...policy.dataValues,
      Clients: clients[0],
      Services: services[0],
      Groups: groups[0],
      ServiceGroups: serviceGroups[0],
      Users: users[0],
      UserGroups: userGroups[0],
      EndpointSecurity: endpointSecurity
    };
    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the list of device polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.createDevicePolicy = async (req, res, next) => {
  const {
    name,
    antivirus,
    is_epp_client,
    is_blocked,
    policy,
    client,
    clientGroup,
    service,
    serviceGroup,
    all_clients,
    all_services,
    user,
    userGroup,
    all_users,
    risk_score_threshold,
    compliance
  } = req.body;

  const policyObject = {
    name,
    antivirus,
    is_epp_client,
    is_blocked,
    policy,
    compliance,
    all_clients,
    all_services,
    all_users,
    risk_score_threshold
  };

  try {
    const transaction = await models.sequelize.transaction();

    // check if policy with same name already exists

    const result = await models.DevicePolicies.count({
      where: { name }
    });

    if (result)
      throw new httpErrors.Conflict('Policy with this name already exists');

    const policy = await models.DevicePolicies.create(policyObject, {
      transaction
    });

    if (policy) {
      antivirus?.forEach(async software => {
        await models.SecurityPolicyPivots.create(
          {
            policy_id: policy?.id,
            software_id: software
          },
          { transaction }
        );
      });
    }

    const CDPolicies = client.map(c => {
      return {
        client_id: c,
        policy_type_id: 2,
        policy_id: policy.id
      };
    });
    const CGroupDPolicies = clientGroup.map(c => {
      return {
        client_group_id: c,
        policy_type_id: 2,
        policy_id: policy.id
      };
    });
    const SDPolicies = service.map(c => {
      return {
        service_id: c,
        policy_type_id: 2,
        policy_id: policy.id
      };
    });
    const SGroupDPolicies = serviceGroup.map(c => {
      return {
        service_group_id: c,
        policy_type_id: 2,
        policy_id: policy.id
      };
    });
    const UPolicies = user.map(c => {
      return {
        user_id: c,
        policy_type_id: 2,
        policy_id: policy.id
      };
    });
    const UGroupPolicies = userGroup.map(c => {
      return {
        user_group_id: c,
        policy_type_id: 2,
        policy_id: policy.id,
        user_group_type: 1
      };
    });

    if (!all_clients) {
      await models.ClientPolicies.bulkCreate(CDPolicies, { transaction });
      await models.ClientGroupPolicies.bulkCreate(CGroupDPolicies, {
        transaction
      });
    }
    if (!all_services) {
      await models.ServicePolicies.bulkCreate(SDPolicies, { transaction });
      await models.ServiceGroupPolicies.bulkCreate(SGroupDPolicies, {
        transaction
      });
    }
    if (!all_users) {
      await models.UserPolicies.bulkCreate(UPolicies, { transaction });
      await models.UserGroupPolicies.bulkCreate(UGroupPolicies, {
        transaction
      });
    }
    await transaction.commit();

    await publishSecurityPolicy(actionType.ADDED, {
      id: policy.id,
      mode: 2,
      data: {
        name: policy.name,
        is_epp_client: +policy.is_epp_client,
        risk_score_threshold: +policy.risk_score_threshold,
        all_clients: +policy.all_clients,
        all_services: +policy.all_services,
        all_users: +policy.all_users,
        is_blocked: +policy.is_blocked,
        compliance: policy.compliance
      }
    });

    insertIntoAdminLogs(
      getPolicyCreateMessage(policy.name, policyTypes.SECURITY_POLICY),
      req
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the list of device polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.updateDevicePolicy = async (req, res, next) => {
  const { id } = req.params;

  const {
    name,
    antivirus,
    is_epp_client,
    is_blocked,
    policy,
    client,
    clientGroup,
    service,
    serviceGroup,
    all_clients,
    all_services,
    all_users,
    user,
    userGroup,
    risk_score_threshold,
    compliance
  } = req.body;

  const transaction = await models.sequelize.transaction();
  try {
    const policy = await models.DevicePolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Device policy not found');

    if(!(policy.name === name)) {
      const duplicateName = await models.DevicePolicies.findOne({
        where: {
          name
        }
      });

      if (duplicateName) {
        throw new httpErrors.BadRequest('Policy with this name already exists');
      }
    }

    let CDPolicies = await models.sequelize.query(
      `select * from ClientPolicies where policy_id = ${id} and policy_type_id = 2`
    );

    let SDPolicies = await models.sequelize.query(
      `select * from ServicePolicies where policy_id = ${id} and policy_type_id = 2`
    );

    let CGroupDPolicies = await models.sequelize.query(
      `select * from ClientGroupPolicies where policy_id = ${id} and policy_type_id = 2`
    );

    let SGroupDPolicies = await models.sequelize.query(
      `select * from ServiceGroupPolicies where policy_id = ${id} and policy_type_id = 2`
    );

    let UPolicies = await models.sequelize.query(
      `SELECT * FROM UserPolicies WHERE policy_id = ${id} AND policy_type_id = 2`
    );

    let UGroupPolicies = await models.sequelize.query(
      `SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ${id} AND policy_type_id = 2`
    );

    CDPolicies = (CDPolicies[0] || []).map(v => v.client_id);
    CGroupDPolicies = (CGroupDPolicies[0] || []).map(v => v.client_group_id);
    SDPolicies = (SDPolicies[0] || []).map(v => v.service_id);
    SGroupDPolicies = (SGroupDPolicies[0] || []).map(v => v.service_group_id);
    UPolicies = (UPolicies[0] || []).map(v => v.user_id);
    UGroupPolicies = (UGroupPolicies[0] || []).map(v => v.user_group_id);

    const clientToAdd = all_clients ? [] : _.difference(client, CDPolicies);
    const clientToDelete = all_clients
      ? CDPolicies
      : _.difference(CDPolicies, client);
    const clientGroupToAdd = all_clients
      ? []
      : _.difference(clientGroup, CGroupDPolicies);
    const clientGroupToDelete = all_clients
      ? CGroupDPolicies
      : _.difference(CGroupDPolicies, clientGroup);
    const serviceToAdd = all_services ? [] : _.difference(service, SDPolicies);
    const serviceToDelete = all_services
      ? SDPolicies
      : _.difference(SDPolicies, service);
    const serviceGroupToAdd = all_services
      ? []
      : _.difference(serviceGroup, SGroupDPolicies);
    const serviceGroupToDelete = all_services
      ? SGroupDPolicies
      : _.difference(SGroupDPolicies, serviceGroup);
    const userToAdd = all_users ? [] : _.difference(user, UPolicies);
    const userToDelete = all_users ? UPolicies : _.difference(UPolicies, user);
    const userGroupToAdd = all_users
      ? []
      : _.difference(userGroup, UGroupPolicies);
    const userGroupToDelete = all_users
      ? UGroupPolicies
      : _.difference(UGroupPolicies, userGroup);

    const [, , , , , , , , , , , , updatedPolicy] = await Promise.all([
      models.ClientPolicies.bulkCreate(
        clientToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 2,
          client_id: v
        })),
        { transaction }
      ),
      models.ClientPolicies.destroy({
        where: {
          client_id: {
            [Op.in]: clientToDelete
          },
          policy_type_id: 2,
          policy_id: policy.id
        },
        transaction
      }),
      models.ClientGroupPolicies.bulkCreate(
        clientGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 2,
          client_group_id: v
        })),
        { transaction }
      ),
      models.ClientGroupPolicies.destroy({
        where: {
          client_group_id: {
            [Op.in]: clientGroupToDelete
          },
          policy_type_id: 2,
          policy_id: policy.id
        },
        transaction
      }),

      models.ServicePolicies.bulkCreate(
        serviceToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 2,
          service_id: v
        })),
        { transaction }
      ),
      models.ServicePolicies.destroy({
        where: {
          service_id: {
            [Op.in]: serviceToDelete
          },
          policy_type_id: 2,
          policy_id: policy.id
        },
        transaction
      }),
      models.ServiceGroupPolicies.bulkCreate(
        serviceGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 2,
          service_group_id: v
        })),
        { transaction }
      ),
      models.ServiceGroupPolicies.destroy({
        where: {
          service_group_id: {
            [Op.in]: serviceGroupToDelete
          },
          policy_type_id: 2,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserPolicies.bulkCreate(
        userToAdd.map(c => {
          return {
            user_id: c,
            policy_type_id: 2,
            policy_id: policy.id
          };
        }, transaction)
      ),
      models.UserGroupPolicies.bulkCreate(
        userGroupToAdd.map(c => {
          return {
            user_group_id: c,
            policy_type_id: 2,
            policy_id: policy.id,
            user_group_type: 1
          };
        }, transaction)
      ),
      models.UserPolicies.destroy({
        where: {
          user_id: {
            [Op.in]: userToDelete
          },
          policy_type_id: 2,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserGroupPolicies.destroy({
        where: {
          user_group_id: {
            [Op.in]: userGroupToDelete
          },
          policy_type_id: 2,
          policy_id: policy.id
        },
        transaction
      }),
      await policy.update(
        {
          name,
          antivirus,
          is_epp_client,
          is_blocked,
          all_clients,
          all_services,
          all_users,
          risk_score_threshold,
          compliance
        },
        {
          transaction,
          updateData: {
            client,
            clientGroup,
            service,
            serviceGroup,
            all_clients,
            all_services,
            all_users
          }
        }
      )
    ]);

    await models.SecurityPolicyPivots.destroy(
      {
        where: {
          policy_id: policy?.id
        }
      },
      { transaction }
    );

    antivirus?.forEach(async software => {
      await models.SecurityPolicyPivots.create(
        {
          policy_id: policy?.id,
          software_id: software
        }
      );
    });
    await transaction.commit();

    await publishSecurityPolicy(actionType.UPDATED, {
      id: updatedPolicy.id,
      data: {
        name: updatedPolicy.name,
        is_epp_client: +updatedPolicy.is_epp_client,
        risk_score_threshold: +updatedPolicy.risk_score_threshold,
        all_clients: +updatedPolicy.all_clients,
        all_services: +updatedPolicy.all_services,
        all_users: +updatedPolicy.all_users,
        is_blocked: +updatedPolicy.is_blocked,
        compliance: updatedPolicy.compliance,
      },
      mode:
        (clientToAdd.length ||
          clientToDelete.length ||
          clientGroupToAdd.length ||
          clientGroupToDelete.length) &&
        (serviceToAdd.length ||
          serviceToDelete.length ||
          serviceGroupToAdd.length ||
          serviceGroupToDelete.length)
          ? 3
          : clientToAdd.length ||
            clientToDelete.length ||
            clientGroupToAdd.length ||
            clientGroupToDelete.length
          ? 1
          : 2
    });

    
    insertIntoAdminLogs(
      getPolicyEditMessage(policy.name, policyTypes.SECURITY_POLICY),
      req
    );
    return res.json(policy);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/**
 * Device device policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.deleteDevicePolicy = async (req, res, next) => {
  const { id } = req.params;
  //const transaction = await models.sequelize.transaction();
  try {
    const policy = await models.DevicePolicies.findOne({
      where: { id }
      //transaction
    });

    if (!policy) throw new httpErrors.NotFound('Device policy not found');
    // const transaction = await models.sequelize.transaction();
    await Promise.all([
      models.ClientPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 2 }
      }),
      models.ServicePolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 2 }
      }),
      models.ClientGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 2 }
      }),
      models.ServiceGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 2 }
      }),
      models.UserPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 2 }
      }),
      models.UserGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 2 }
      }),
      models.SecurityPolicyPivots.destroy({
        where: { policy_id: policy.id }
      }),
      policy.destroy()
    ]);

    await publishSecurityPolicy(actionType.DELETED, {
      id: id,
      data: {},
      mode: 2
    });

    insertIntoAdminLogs(
      getPolicyDeleteMessage(policy.name, policyTypes.SECURITY_POLICY),
      req
    );

    // await transaction.commit();
    return res.json(policy);
  } catch (error) {
    // await transaction.rollback();
    next(error);
  }
};

// SOFTWARE POLICIES

/**
 * Get the list of Software polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getSoftwarePolicies = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchValue,
      clients = [],
      allClientPolicies = false
    } = req.query;

    const options = {
      where: {},
      limit,
      offset,
      order: []
      // distinct: true
    };

    const allowedSortValues = ['id', 'name', 'grace_period', 'is_blocked'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([sort, dir]);
    }

    if (searchValue) {
      options.where = {
        [Op.or]: {
          name: {
            [Op.like]: `%${searchValue.trim()}%`
          }
        }
      };
    }

    let clientIds = [];
    if (clients && clients.length) {
      clientIds = convertToArray(clients).map(id => parseInt(id, 10)).filter(id => !isNaN(id));
      
      options.where[Op.or] = [
        { all_clients: true },
        models.sequelize.literal(`
          EXISTS (
            SELECT 1 
            FROM ClientPolicies 
            INNER JOIN Clients ON ClientPolicies.client_id = Clients.id
            WHERE ClientPolicies.policy_id = SoftwarePolicies.id 
            AND Clients.id IN (?)
            AND ClientPolicies.policy_type_id = 4
          ) OR EXISTS (
            SELECT 1 
            FROM GroupClients 
            INNER JOIN ClientGroupPolicies ON ClientGroupPolicies.client_group_id = GroupClients.groupId
            INNER JOIN Clients ON GroupClients.clientId = Clients.id
            WHERE ClientGroupPolicies.policy_id = SoftwarePolicies.id 
            AND Clients.id IN (?)
            AND ClientGroupPolicies.policy_type_id = 4
          )
        `)
      ];

      // When executing, provide replacements
      options.replacements = [clientIds, clientIds];
    }

    if (allClientPolicies == 'true') {
      options.where.all_clients = true;
    }

    const result = await models.SoftwarePolicies.paginate({
      ...options
    });
  
    const enrichedData = await Promise.all(result.data.map(async (item) => {
      const [userGroups, users, clients, services, groups, serviceGroups] = await Promise.all([
        models.sequelize.query(
          'SELECT group_name as name FROM CustomUserGroups WHERE id IN (SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ? AND policy_type_id = 4)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name as Name FROM LocalUsers WHERE id IN (SELECT user_id FROM UserPolicies WHERE policy_id = ? AND policy_type_id = 4)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT Name FROM Clients WHERE id IN (SELECT client_id FROM ClientPolicies WHERE policy_id = ? AND policy_type_id = 4)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT Name FROM Services WHERE id IN (SELECT service_id FROM ServicePolicies WHERE policy_id = ? AND policy_type_id = 4)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name FROM `Groups` WHERE id IN (SELECT client_group_id FROM ClientGroupPolicies WHERE policy_id = ? AND policy_type_id = 4)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name FROM ServiceGroups WHERE id IN (SELECT service_group_id FROM ServiceGroupPolicies WHERE policy_id = ? AND policy_type_id = 4)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        )
      ]);
    
      return {
        ...item.dataValues,
        UserGroups: userGroups || [],
        Users: users || [],
        Clients: clients,
        Services: services,
        Groups: groups,
        ServiceGroups: serviceGroups
      };
    }));
    
    result.data = enrichedData;

    res.json(result);
  } catch (error) {
    next(error);
  }
};

module.exports.createSoftwarePolicy = async (req, res, next) => {
  const {
    name,
    grace_period,
    client,
    clientGroup,
    service,
    serviceGroup,
    all_clients,
    all_services,
    block_services,
    user,
    userGroup,
    all_users
  } = req.body;

  const policyObject = {
    name,
    grace_period,
    client,
    clientGroup,
    service,
    block_services,
    all_clients,
    all_services,
    all_users
  };

  try {
    const transaction = await models.sequelize.transaction();

    // Check whether given policy name already exists.
    const result = await models.SoftwarePolicies.findOne({
      where: {
        name
      }
    });

    if (result) {
      return res
        .status(400)
        .json({ message: 'Policy with this name already exists' });
    }

    const policy = await models.SoftwarePolicies.create(policyObject, {
      transaction
    });
    const CSPolicies = client.map(c => {
      return {
        client_id: c,
        policy_type_id: 4,
        policy_id: policy.id
      };
    });
    const CGroupSPolicies = clientGroup.map(c => {
      return {
        client_group_id: c,
        policy_type_id: 4,
        policy_id: policy.id
      };
    });
    const SSPolicies = service.map(c => {
      return {
        service_id: c,
        policy_type_id: 4,
        policy_id: policy.id
      };
    });
    const SGroupSPolicies = serviceGroup.map(c => {
      return {
        service_group_id: c,
        policy_type_id: 4,
        policy_id: policy.id
      };
    });
    const UPolicies = user.map(c => {
      return {
        user_id: c,
        policy_type_id: 4,
        policy_id: policy.id
      };
    });
    const UGroupPolicies = userGroup.map(c => {
      return {
        user_group_id: c,
        policy_type_id: 4,
        policy_id: policy.id,
        user_group_type: 1
      };
    });

    if (!all_clients) {
      await models.ClientPolicies.bulkCreate(CSPolicies, { transaction });
      await models.ClientGroupPolicies.bulkCreate(CGroupSPolicies, {
        transaction
      });
    }
    if (!all_services) {
      await models.ServicePolicies.bulkCreate(SSPolicies, { transaction });
      await models.ServiceGroupPolicies.bulkCreate(SGroupSPolicies, {
        transaction
      });
    }
    if (!all_users) {
      await models.UserPolicies.bulkCreate(UPolicies, { transaction });
      await models.UserGroupPolicies.bulkCreate(UGroupPolicies, {
        transaction
      });
    }
    await transaction.commit();

    insertIntoAdminLogs(
      getPolicyCreateMessage(policy.name, policyTypes.SOFTWARE_POLICY),
      req
    );

    await publishSoftwarePolicy(actionType.ADDED, {
      id: policy.id,
      data: {}
    });

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the software policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getSoftwarePolicyById = async (req, res, next) => {
  const { id } = req.params;
  try {
    let policy = await models.SoftwarePolicies.findOne({
      where: {
        id
      }
    });
    if (!policy) throw new httpErrors.NotFound('Software policy not found');
    const clients = await models.sequelize.query(`
      SELECT id, Name, device_os FROM Clients
      WHERE id IN (
        SELECT client_id FROM ClientPolicies
        WHERE policy_id = ${id} AND policy_type_id = 4
      );
    `);
    const services = await models.sequelize.query(`
      SELECT id, Name FROM Services
      WHERE id IN (
        SELECT service_id FROM ServicePolicies
        WHERE policy_id = ${id} AND policy_type_id = 4
      );
    `);
    const groups = await models.sequelize.query(`
      SELECT id, name FROM \`Groups\`
      WHERE id IN (
        SELECT client_group_id FROM ClientGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 4
      );
    `);
    const serviceGroups = await models.sequelize.query(`
      SELECT id, name FROM ServiceGroups
      WHERE id IN (
        SELECT service_group_id FROM ServiceGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 4
      );
    `);
    const users = await models.sequelize.query(`
      SELECT id, Name, runtime_bind_id FROM LocalUsers
      WHERE id IN (
        SELECT user_id FROM UserPolicies
        WHERE policy_id = ${id} AND policy_type_id = 4
      );
    `);
    const userGroups = await models.sequelize.query(`
      SELECT id, group_name as name, runtime_bind_id, raw_query_id FROM CustomUserGroups
      WHERE id IN (
        SELECT user_group_id FROM UserGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 4
      );
    `);

    policy.dataValues = {
      ...policy.dataValues,
      Clients: clients[0],
      Services: services[0],
      Groups: groups[0],
      ServiceGroups: serviceGroups[0],
      Users: users[0],
      UserGroups: userGroups[0]
    };
    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * update software polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.updateSoftwarePolicy = async (req, res, next) => {
  const { id } = req.params;

  const {
    name,
    grace_period,
    client,
    clientGroup,
    all_clients,
    all_services,
    service,
    serviceGroup,
    block_services,
    user,
    userGroup,
    all_users
  } = req.body;

  const transaction = await models.sequelize.transaction();
  try {
    let allClients;
    const policy = await models.SoftwarePolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Software policy not found');

    if(!(policy.name === name)) {
      const duplicateName = await models.SoftwarePolicies.findOne({
        where: {
          name
        }
      });

      if (duplicateName) {
        throw new httpErrors.BadRequest('Policy with this name already exists');
      }
    }

    let CSPolicies = await models.sequelize.query(
      `select * from ClientPolicies where policy_id = ${id} and policy_type_id = 4`
    );

    let SSPolicies = await models.sequelize.query(
      `select * from ServicePolicies where policy_id = ${id} and policy_type_id = 4`
    );

    let CGroupSPolicies = await models.sequelize.query(
      `select * from ClientGroupPolicies where policy_id = ${id} and policy_type_id = 4`
    );

    let SGroupSPolicies = await models.sequelize.query(
      `select * from ServiceGroupPolicies where policy_id = ${id} and policy_type_id = 4`
    );

    let UPolicies = await models.sequelize.query(
      `SELECT * FROM UserPolicies WHERE policy_id = ${id} AND policy_type_id = 4`
    );

    let UGroupPolicies = await models.sequelize.query(
      `SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ${id} AND policy_type_id = 4`
    );

    CSPolicies = (CSPolicies[0] || []).map(v => v.client_id);
    CGroupSPolicies = (CGroupSPolicies[0] || []).map(v => v.client_group_id);
    SSPolicies = (SSPolicies[0] || []).map(v => v.service_id);
    SGroupSPolicies = (SGroupSPolicies[0] || []).map(v => v.service_group_id);
    UPolicies = (UPolicies[0] || []).map(v => v.user_id);
    UGroupPolicies = (UGroupPolicies[0] || []).map(v => v.user_group_id);

    const clientToAdd = all_clients ? [] : _.difference(client, CSPolicies);
    const clientToDelete = all_clients
      ? CSPolicies
      : _.difference(CSPolicies, client);
    const clientGroupToAdd = all_clients
      ? []
      : _.difference(clientGroup, CGroupSPolicies);
    const clientGroupToDelete = all_clients
      ? CSPolicies
      : _.difference(CGroupSPolicies, clientGroup);

    const serviceToAdd = all_services ? [] : _.difference(service, SSPolicies);
    const serviceToDelete = all_services
      ? SSPolicies
      : _.difference(SSPolicies, service);
    const serviceGroupToAdd = all_services
      ? []
      : _.difference(serviceGroup, SGroupSPolicies);
    const serviceGroupToDelete = all_services
      ? SGroupSPolicies
      : _.difference(SGroupSPolicies, serviceGroup);

    const userToAdd = all_users ? [] : _.difference(user, UPolicies);
    const userToDelete = all_users ? UPolicies : _.difference(UPolicies, user);
    const userGroupToAdd = all_users
      ? []
      : _.difference(userGroup, UGroupPolicies);
    const userGroupToDelete = all_users
      ? UGroupPolicies
      : _.difference(UGroupPolicies, userGroup);

    const [, , , , , , , , updatedPolicy] = await Promise.all([
      models.ClientPolicies.bulkCreate(
        clientToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 4,
          client_id: v
        })),
        { transaction }
      ),
      models.ClientPolicies.destroy({
        where: {
          client_id: {
            [Op.in]: clientToDelete
          },
          policy_type_id: 4,
          policy_id: policy.id
        },
        transaction
      }),
      models.ClientGroupPolicies.bulkCreate(
        clientGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 4,
          client_group_id: v
        })),
        { transaction }
      ),
      models.ClientGroupPolicies.destroy({
        where: {
          client_group_id: {
            [Op.in]: clientGroupToDelete
          },
          policy_type_id: 4,
          policy_id: policy.id
        },
        transaction
      }),
      models.ServicePolicies.bulkCreate(
        serviceToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 4,
          service_id: v
        })),
        { transaction }
      ),
      models.ServicePolicies.destroy({
        where: {
          service_id: {
            [Op.in]: serviceToDelete
          },
          policy_type_id: 4,
          policy_id: policy.id
        },
        transaction
      }),
      models.ServiceGroupPolicies.bulkCreate(
        serviceGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 4,
          service_group_id: v
        })),
        { transaction }
      ),
      models.ServiceGroupPolicies.destroy({
        where: {
          service_group_id: {
            [Op.in]: serviceGroupToDelete
          },
          policy_type_id: 4,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserPolicies.bulkCreate(
        userToAdd.map(c => {
          return {
            user_id: c,
            policy_type_id: 4,
            policy_id: policy.id
          };
        }, transaction)
      ),
      models.UserGroupPolicies.bulkCreate(
        userGroupToAdd.map(c => {
          return {
            user_group_id: c,
            policy_type_id: 4,
            policy_id: policy.id,
            user_group_type: 1
          };
        }, transaction)
      ),
      models.UserPolicies.destroy({
        where: {
          user_id: {
            [Op.in]: userToDelete
          },
          policy_type_id: 4,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserGroupPolicies.destroy({
        where: {
          user_group_id: {
            [Op.in]: userGroupToDelete
          },
          policy_type_id: 4,
          policy_id: policy.id
        },
        transaction
      }),

      await policy.update(
        {
          name,
          grace_period,
          block_services,
          all_clients,
          all_services,
          all_users
        },
        {
          transaction,
          updateData: {
            client,
            clientGroup,
            all_clients,
            all_users
          }
        }
      )
    ]);

    if (all_clients) {
      allClients = await models.Client.findAll({
        attributes: ['id']
      });

      let allClientsIdArrayFormatted = [];
      allClients.forEach(element => {
        allClientsIdArrayFormatted.push(element.id);
      });

      await models.sequelize
        .query(`UPDATE DeviceInventories d INNER JOIN Clients c ON c.device_id = d.id
          SET d.last_sync_time = NOW() WHERE c.id in(${allClientsIdArrayFormatted})`);
    } else if (client.length > 0) {
      await models.sequelize
        .query(`UPDATE DeviceInventories d INNER JOIN Clients c ON c.device_id = d.id
          SET d.last_sync_time = NOW() WHERE c.id in(${client})`);
    } else if (clientGroup.length > 0) {
      await models.sequelize.query(`UPDATE DeviceInventories d
        INNER JOIN Clients c on c.device_id = d.id 
        INNER JOIN GroupClients gc on gc.clientId = c.id 
        INNER JOIN \`Groups\` g on g.id = gc.groupId 
        SET d.last_sync_time = NOW() WHERE gc.groupId IN (${clientGroup})`);
    }

    let toDeleteService = service;
    let toDeleteClient = client;

    if (!all_services && serviceGroup.length > 0) {
      const policiesIds = await models.sequelize.query(
        `SELECT serviceId FROM sdp.GroupServices WHERE groupId IN (${serviceGroup})`,
        {
          type: queryTypes.SELECT
        }
      );
      policiesIds.map(data => {
        toDeleteService.push(data.serviceId);
      });
    }

    if (!all_clients && clientGroup.length > 0) {
      const clientsIds = await models.sequelize.query(
        `SELECT clientId FROM sdp.GroupClients WHERE groupId IN (${clientGroup})`,
        {
          type: queryTypes.SELECT
        }
      );
      clientsIds.map(data => {
        toDeleteClient.push(data.clientId);
      });
    }

    await models.sequelize.query(`DELETE FROM sdp.BlockedServices bs
              WHERE bs.policy_id = ${id}
              and (service_id not in (${
                toDeleteService.length > 0 ? toDeleteService : -1
              })
                   or client_id not in (${
                     toDeleteClient.length > 0 ? toDeleteClient : -1
                   }))`);

    await transaction.commit();

    insertIntoAdminLogs(
      getPolicyEditMessage(policy.name, policyTypes.SOFTWARE_POLICY),
      req
    );
    await publishSoftwarePolicy(actionType.UPDATED, {
      id: id,
      data: {}
    });

    job(true);

    return res.json(policy);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/**
 * software device policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.deleteSoftwarePolicy = async (req, res, next) => {
  const { id } = req.params;
  //const transaction = await models.sequelize.transaction();
  try {
    const policy = await models.SoftwarePolicies.findOne({
      where: { id }
      //transaction
    });
    const blockedService = await models.sequelize.query(`
      select * from BlockedServices where policy_id=${id}`);

    if (!policy) throw new httpErrors.NotFound('Software policy not found');
    // const transaction = await models.sequelize.transaction();

    await Promise.all([
      models.ClientPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 4 }
      }),
      models.ServicePolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 4 }
      }),
      models.ClientGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 4 }
      }),
      models.ServiceGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 4 }
      }),
      models.UserPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 4 }
      }),
      models.UserGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 4 }
      }),
      policy.destroy()
    ]);

    insertIntoAdminLogs(
      getPolicyDeleteMessage(policy.name, policyTypes.SOFTWARE_POLICY),
      req
    );
    // await transaction.commit();
    await publishSoftwarePolicy(actionType.DELETED, {
      id: id,
      data: {}
    });
    return res.json(policy);
  } catch (error) {
    // await transaction.rollback();
    next(error);
  }
};

//SOFTWARE PROCESS POLICIES

/**
 * Get the list of Software Process polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getSoftwareProcessPolicies = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchValue,
      clients = [],
      allClientPolicies = false
    } = req.query;

    const options = {
      where: {},
      limit,
      offset,
      order: []
      // distinct: true
    };

    const allowedSortValues = ['id', 'name', 'compliance', 'is_blocked'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([sort, dir]);
    }

    if (searchValue) {
      options.where = {
        [Op.or]: {
          name: {
            [Op.like]: `%${searchValue.trim()}%`
          },
          compliance: {
            [Op.like]: `%${searchValue.trim()}%`
          }
        }
      };
    }

    let clientIds = [];
    if (clients && clients.length) {
      clientIds = convertToArray(clients).map(id => parseInt(id, 10)).filter(id => !isNaN(id));
      
      options.where[Op.or] = [
        { all_clients: true },
        models.sequelize.literal(`
          EXISTS (
            SELECT 1 
            FROM ClientPolicies 
            INNER JOIN Clients ON ClientPolicies.client_id = Clients.id
            WHERE ClientPolicies.policy_id = SoftwareProcessPolicies.id 
            AND Clients.id IN (?)
            AND ClientPolicies.policy_type_id = 5
          ) OR EXISTS (
            SELECT 1 
            FROM GroupClients 
            INNER JOIN ClientGroupPolicies ON ClientGroupPolicies.client_group_id = GroupClients.groupId
            INNER JOIN Clients ON GroupClients.clientId = Clients.id
            WHERE ClientGroupPolicies.policy_id = SoftwareProcessPolicies.id 
            AND Clients.id IN (?)
            AND ClientGroupPolicies.policy_type_id = 5
          )
        `)
      ];

      // When executing, provide replacements
      options.replacements = [clientIds, clientIds];
    }

    if (allClientPolicies == 'true') {
      options.where.all_clients = true;
    }

    const result = await models.SoftwareProcessPolicies.paginate({
      ...options
    });
      const enrichedData = await Promise.all(result.data.map(async (item) => {
        const [userGroups, users, clients, services, groups, serviceGroups] = await Promise.all([
          models.sequelize.query(
            'SELECT group_name as name FROM CustomUserGroups WHERE id IN (SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ? AND policy_type_id = 5)',
            { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
          ),
          models.sequelize.query(
            'SELECT name as Name FROM LocalUsers WHERE id IN (SELECT user_id FROM UserPolicies WHERE policy_id = ? AND policy_type_id = 5)',
            { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
          ),
          models.sequelize.query(
            'SELECT Name FROM Clients WHERE id IN (SELECT client_id FROM ClientPolicies WHERE policy_id = ? AND policy_type_id = 5)',
            { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
          ),
          models.sequelize.query(
            'SELECT Name FROM Services WHERE id IN (SELECT service_id FROM ServicePolicies WHERE policy_id = ? AND policy_type_id = 5)',
            { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
          ),
          models.sequelize.query(
            'SELECT name FROM `Groups` WHERE id IN (SELECT client_group_id FROM ClientGroupPolicies WHERE policy_id = ? AND policy_type_id = 5)',
            { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
          ),
          models.sequelize.query(
            'SELECT name FROM ServiceGroups WHERE id IN (SELECT service_group_id FROM ServiceGroupPolicies WHERE policy_id = ? AND policy_type_id = 5)',
            { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
          )
        ]);
      
        return {
          ...item.dataValues,
          UserGroups: userGroups || [],
          Users: users || [],
          Clients: clients,
          Services: services,
          Groups: groups,
          ServiceGroups: serviceGroups
        };
      }));
    
    result.data = enrichedData;
    res.json(result);
  } catch (error) {
    next(error);
  }
};

module.exports.createSoftwareProcessPolicy = async (req, res, next) => {
  const {
    name,
    is_blocked,
    client,
    compliance,
    selectedSoftwares,
    clientGroup,
    service,
    serviceGroup,
    all_clients,
    all_services,
    user,
    userGroup,
    all_users
  } = req.body;

  const policyObject = {
    name,
    is_blocked,
    client,
    compliance,
    clientGroup,
    service,
    all_clients,
    all_services,
    all_users,
    SPPolicyPivots: selectedSoftwares.map(so => ({ software_id: so }))
  };
  try {
    const transaction = await models.sequelize.transaction();

    // Check whether given policy name already exists.
    const result = await models.SoftwareProcessPolicies.findOne({
      where: {
        name
      }
    });

    if (result) {
      return res
        .status(400)
        .json({ message: 'Policy with this name already exists' });
    }

    const policy = await models.SoftwareProcessPolicies.create(policyObject, {
      include: [
        {
          model: models.SoftwareProcessPolicyPivots,
          as: 'SPPolicyPivots'
        }
      ],
      transaction
    });
    const CPPolicies = client.map(c => {
      return {
        client_id: c,
        policy_type_id: 5,
        policy_id: policy.id
      };
    });
    const CGroupPPolicies = clientGroup.map(c => {
      return {
        client_group_id: c,
        policy_type_id: 5,
        policy_id: policy.id
      };
    });
    const SPPolicies = service.map(c => {
      return {
        service_id: c,
        policy_type_id: 5,
        policy_id: policy.id
      };
    });
    const SGroupPPolicies = serviceGroup.map(c => {
      return {
        service_group_id: c,
        policy_type_id: 5,
        policy_id: policy.id
      };
    });
    const UPolicies = user.map(c => {
      return {
        user_id: c,
        policy_type_id: 5,
        policy_id: policy.id
      };
    });
    const UGroupPolicies = userGroup.map(c => {
      return {
        user_group_id: c,
        policy_type_id: 5,
        policy_id: policy.id,
        user_group_type: 1
      };
    });

    if (!all_clients) {
      await models.ClientPolicies.bulkCreate(CPPolicies, { transaction });
      await models.ClientGroupPolicies.bulkCreate(CGroupPPolicies, {
        transaction
      });
    }
    if (!all_services) {
      await models.ServicePolicies.bulkCreate(SPPolicies, { transaction });
      await models.ServiceGroupPolicies.bulkCreate(SGroupPPolicies, {
        transaction
      });
    }
    if (!all_users) {
      await models.UserPolicies.bulkCreate(UPolicies, { transaction });
      await models.UserGroupPolicies.bulkCreate(UGroupPolicies, {
        transaction
      });
    }
    await transaction.commit();

    insertIntoAdminLogs(
      getPolicyCreateMessage(policy.name, policyTypes.PROCESS_POLICY),
      req
    );
    await publishProcessPolicy(actionType.ADDED, {
      id: policy.id,
      mode: 2,
      data: {
        name: policy.name,
        all_clients: +policy.all_clients,
        all_services: +policy.all_services,
        is_blocked: +policy.is_blocked,
        compliance: policy.compliance,
        all_users: +policy.all_users
      }
    });

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the software policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getSoftwareProcessPolicyById = async (req, res, next) => {
  const { id } = req.params;
  try {
    let policy = await models.SoftwareProcessPolicies.findOne({
      where: {
        id
      },
      include: [
        {
          model: models.SoftwareInstances,
          attributes: ['id', 'binding_name'],
          required: false
        }
      ]
    });
    if (!policy)
      throw new httpErrors.NotFound('Software process policy not found');
    const clients = await models.sequelize.query(`
      SELECT id, Name, device_os FROM Clients
      WHERE id IN (
        SELECT client_id FROM ClientPolicies
        WHERE policy_id = ${id} AND policy_type_id = 5
      );
    `);
    const services = await models.sequelize.query(`
      SELECT id, Name FROM Services
      WHERE id IN (
        SELECT service_id FROM ServicePolicies
        WHERE policy_id = ${id} AND policy_type_id = 5
      );
    `);
    const groups = await models.sequelize.query(`
      SELECT id, name FROM \`Groups\`
      WHERE id IN (
        SELECT client_group_id FROM ClientGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 5
      );
    `);
    const serviceGroups = await models.sequelize.query(`
      SELECT id, name FROM ServiceGroups
      WHERE id IN (
        SELECT service_group_id FROM ServiceGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 5
      );
    `);
    const users = await models.sequelize.query(`
      SELECT id, Name, runtime_bind_id FROM LocalUsers
      WHERE id IN (
        SELECT user_id FROM UserPolicies
        WHERE policy_id = ${id} AND policy_type_id = 5
      );
    `);
    const userGroups = await models.sequelize.query(`
      SELECT id, group_name as name, runtime_bind_id, raw_query_id FROM CustomUserGroups
      WHERE id IN (
        SELECT user_group_id FROM UserGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 5
      );
    `);

    policy.dataValues = {
      ...policy.dataValues,
      Clients: clients[0],
      Services: services[0],
      Groups: groups[0],
      ServiceGroups: serviceGroups[0],
      Users: users[0],
      UserGroups: userGroups[0]
    };
    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

module.exports.getSoftwaresProcessBindingList = async (req, res, next) => {
  try {
    const softwareList = await models.sequelize.query(`
      SELECT id,binding_name FROM SoftwareInstances WHERE is_edr = 0;
   `);
    const finalList = softwareList[0].map(software => {
      return {
        id: software.id,
        binding_name: software.binding_name
      };
    });
    res.json(finalList);
  } catch (error) {
    next(error);
  }
};

module.exports.getEDRSoftwareProcessBindingList = async (req, res, next) => {
  try {
    const softwareList = await models.sequelize.query(`
      SELECT id,binding_name FROM SoftwareInstances WHERE is_edr = 1;
   `);
    const finalList = softwareList[0].map(software => {
      return {
        id: software.id,
        binding_name: software.binding_name
      };
    });
    res.json(finalList);
  } catch (error) {
    next(error);
  }
};

/**
 * update software polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.updateSoftwareProcessPolicy = async (req, res, next) => {
  const { id } = req.params;

  const {
    name,
    is_blocked,
    compliance,
    selectedSoftwares,
    client,
    clientGroup,
    all_clients,
    all_services,
    service,
    serviceGroup,
    all_users,
    userGroup,
    user
  } = req.body;

  const transaction = await models.sequelize.transaction();

  try {
    const policy = await models.SoftwareProcessPolicies.findOne({
      where: { id },
      include: [
        {
          model: models.SoftwareProcessPolicyPivots,
          as: 'SPPolicyPivots'
        }
      ],
      transaction
    });

    if (!policy)
      throw new httpErrors.NotFound('Software Process policy not found');

    if(!(policy.name === name)) {
      const duplicateName = await models.SoftwareProcessPolicies.findOne({
        where: {
          name
        }
      });
      
      if (duplicateName) {
        throw new httpErrors.BadRequest('Policy with this name already exists');
      }
    }

    let CPPolicies = await models.sequelize.query(
      `select * from ClientPolicies where policy_id = ${id} and policy_type_id = 5`
    );

    let SPPolicies = await models.sequelize.query(
      `select * from ServicePolicies where policy_id = ${id} and policy_type_id = 5`
    );

    let CGroupPPolicies = await models.sequelize.query(
      `select * from ClientGroupPolicies where policy_id = ${id} and policy_type_id = 5`
    );

    let SGroupPPolicies = await models.sequelize.query(
      `select * from ServiceGroupPolicies where policy_id = ${id} and policy_type_id = 5`
    );

    let UPolicies = await models.sequelize.query(
      `SELECT * FROM UserPolicies WHERE policy_id = ${id} AND policy_type_id = 5`
    );

    let UGroupPolicies = await models.sequelize.query(
      `SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ${id} AND policy_type_id = 5`
    );

    CPPolicies = (CPPolicies[0] || []).map(v => v.client_id);
    CGroupPPolicies = (CGroupPPolicies[0] || []).map(v => v.client_group_id);
    SPPolicies = (SPPolicies[0] || []).map(v => v.service_id);
    SGroupPPolicies = (SGroupPPolicies[0] || []).map(v => v.service_group_id);
    UPolicies = (UPolicies[0] || []).map(v => v.user_id);
    UGroupPolicies = (UGroupPolicies[0] || []).map(v => v.user_group_id);
    const SPPolicyPivots = policy.SPPolicyPivots.map(v => v.software_id);

    const clientToAdd = all_clients ? [] : _.difference(client, CPPolicies);
    const clientToDelete = all_clients
      ? CPPolicies
      : _.difference(CPPolicies, client);
    const clientGroupToAdd = all_clients
      ? []
      : _.difference(clientGroup, CGroupPPolicies);
    const clientGroupToDelete = all_clients
      ? CGroupPPolicies
      : _.difference(CGroupPPolicies, clientGroup);
    const serviceToAdd = all_services ? [] : _.difference(service, SPPolicies);
    const serviceToDelete = all_services
      ? SPPolicies
      : _.difference(SPPolicies, service);
    const serviceGroupToAdd = all_services
      ? []
      : _.difference(serviceGroup, SGroupPPolicies);
    const serviceGroupToDelete = all_services
      ? SGroupPPolicies
      : _.difference(SGroupPPolicies, serviceGroup);
    const softwareToAdd = _.difference(selectedSoftwares, SPPolicyPivots);
    const softwareToDelete = _.difference(SPPolicyPivots, selectedSoftwares);
    const userToAdd = all_users ? [] : _.difference(user, UPolicies);
    const userToDelete = all_users ? UPolicies : _.difference(UPolicies, user);
    const userGroupToAdd = all_users
      ? []
      : _.difference(userGroup, UGroupPolicies);
    const userGroupToDelete = all_users
      ? UGroupPolicies
      : _.difference(UGroupPolicies, userGroup);

    const [, , , , , , , , , , , , , , updatedPolicy] = await Promise.all([
      models.ClientPolicies.bulkCreate(
        clientToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 5,
          client_id: v
        })),
        { transaction }
      ),
      models.ClientPolicies.destroy({
        where: {
          client_id: {
            [Op.in]: clientToDelete
          },
          policy_type_id: 5,
          policy_id: policy.id
        },
        transaction
      }),
      models.ClientGroupPolicies.bulkCreate(
        clientGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 5,
          client_group_id: v
        })),
        { transaction }
      ),
      models.ClientGroupPolicies.destroy({
        where: {
          client_group_id: {
            [Op.in]: clientGroupToDelete
          },
          policy_type_id: 5,
          policy_id: policy.id
        },
        transaction
      }),

      models.ServicePolicies.bulkCreate(
        serviceToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 5,
          service_id: v
        })),
        { transaction }
      ),
      models.ServicePolicies.destroy({
        where: {
          service_id: {
            [Op.in]: serviceToDelete
          },
          policy_type_id: 5,
          policy_id: policy.id
        },
        transaction
      }),
      models.ServiceGroupPolicies.bulkCreate(
        serviceGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 5,
          service_group_id: v
        })),
        { transaction }
      ),
      models.ServiceGroupPolicies.destroy({
        where: {
          service_group_id: {
            [Op.in]: serviceGroupToDelete
          },
          policy_type_id: 5,
          policy_id: policy.id
        },
        transaction
      }),
      models.SoftwareProcessPolicyPivots.bulkCreate(
        softwareToAdd.map(v => ({
          policy_id: policy.id,
          software_id: v
        })),
        { transaction }
      ),
      models.SoftwareProcessPolicyPivots.destroy({
        where: {
          software_id: {
            [Op.in]: softwareToDelete
          },
          policy_id: policy.id
        },
        transaction
      }),
      models.UserPolicies.bulkCreate(
        userToAdd.map(c => {
          return {
            user_id: c,
            policy_type_id: 5,
            policy_id: policy.id
          };
        }, transaction)
      ),
      models.UserGroupPolicies.bulkCreate(
        userGroupToAdd.map(c => {
          return {
            user_group_id: c,
            policy_type_id: 5,
            policy_id: policy.id,
            user_group_type: 1
          };
        }, transaction)
      ),
      models.UserPolicies.destroy({
        where: {
          user_id: {
            [Op.in]: userToDelete
          },
          policy_type_id: 5,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserGroupPolicies.destroy({
        where: {
          user_group_id: {
            [Op.in]: userGroupToDelete
          },
          policy_type_id: 5,
          policy_id: policy.id
        },
        transaction
      }),
      await policy.update(
        {
          name,
          is_blocked,
          compliance,
          all_clients,
          all_services,
          all_users
        },
        {
          transaction,
          updateData: {
            client,
            clientGroup,
            service,
            serviceGroup,
            selectedSoftwares,
            all_clients,
            all_services,
            all_users
          }
        }
      )
    ]);

    await transaction.commit();

    await publishProcessPolicy(actionType.UPDATED, {
      id: updatedPolicy.id,
      data: {
        name: updatedPolicy.name,
        all_clients: +updatedPolicy.all_clients,
        all_services: +updatedPolicy.all_services,
        is_blocked: +updatedPolicy.is_blocked,
        compliance: updatedPolicy.compliance,
        all_users: +updatedPolicy.all_users
      }
    });

    insertIntoAdminLogs(
      getPolicyEditMessage(policy.name, policyTypes.PROCESS_POLICY),
      req
    );
    return res.json(policy);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/**
 * delete software process policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.deleteSoftwareProcessPolicy = async (req, res, next) => {
  const { id } = req.params;
  try {
    const policy = await models.SoftwareProcessPolicies.findOne({
      where: { id }
    });

    if (!policy)
      throw new httpErrors.NotFound('Software process policy not found');

    await Promise.all([
      models.ClientPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 5 }
      }),
      models.ServicePolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 5 }
      }),
      models.ClientGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 5 }
      }),
      models.ServiceGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 5 }
      }),
      models.UserPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 5 }
      }),
      models.UserGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 5 }
      }),
      policy.destroy()
    ]);

    insertIntoAdminLogs(
      getPolicyDeleteMessage(policy.name, policyTypes.PROCESS_POLICY),
      req
    );
    await publishProcessPolicy(actionType.DELETED, {
      id: policy.id,
      data: {}
    });
    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

// OS POLICIES

/**
 * Get the list of Os polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getOsPolicies = async (req, res, next) => {
  try {
    const { limit, offset, sort = 'os', dir = 'asc', os } = req.query;

    const options = {
      where: {},
      limit,
      offset,
      order: []
      // distinct: true
    };

    const allowedSortValues = ['os', 'os_version'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    if (os) {
      options.where = {
        ...options.where,
        os_display_name: {
          [Op.like]: '%' + os + '%'
        }
      };
    }

    const result = await models.OsPolicies.paginate({
      ...options
    });
    res.json(result);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the Os policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getOsPolicyById = async (req, res, next) => {
  const { id } = req.params;
  try {
    const policy = await models.OsPolicies.findOne({
      where: {
        id
      }
    });
    if (!policy) throw new httpErrors.NotFound('Policy not found');

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Create a OS policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.createOsPolicy = async (req, res, next) => {
  const { os, os_version, is_blocked } = req.body;

  const policyObject = {
    os,
    os_version,
    is_blocked,
    os_display_name: os
  };
  try {
    const policy = await models.OsPolicies.create(policyObject);

    await publishOsPolicy(actionType.ADDED, {
      id: policy.id,
      data: {
        os: policy.os,
        os_version: policy.os_version,
        is_blocked: +policy.is_blocked
      }
    });

    insertIntoAdminLogs(
      getPolicyCreateMessage(policy.os, policyTypes.OS_POLICY),
      req
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Update a policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.updateOsPolicy = async (req, res, next) => {
  const { id } = req.params;

  const { os, os_version, is_blocked } = req.body;

  try {
    const policy = await models.OsPolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');

    await policy.update({
      os,
      os_version,
      is_blocked,
      os_display_name: os
    });

    await publishOsPolicy(actionType.UPDATED, {
      id: policy.id,
      data: {
        os: policy.os,
        os_version: policy.os_version,
        is_blocked: +policy.is_blocked
      }
    });

    insertIntoAdminLogs(
      getPolicyEditMessage(policy.os, policyTypes.OS_POLICY),
      req
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Delete a policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.deleteOsPolicy = async (req, res, next) => {
  const { id } = req.params;
  try {
    const policy = await models.OsPolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Policy not found');

    await policy.destroy();

    await publishOsPolicy(actionType.DELETED, {
      id: policy.id,
      data: {
        os: policy.os,
        os_version: policy.os_version,
        is_blocked: +policy.is_blocked
      }
    });

    insertIntoAdminLogs(
      getPolicyDeleteMessage(policy.os, policyTypes.OS_POLICY),
      req
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Get route policies
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getRoutingPolicies = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchValue,
      gateways = []
    } = req.query;

    const options = {
      where: {},
      limit,
      offset,
      order: []
      // distinct: true
    };

    const allowedSortValues = ['id', 'name'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    if (searchValue) {
      options.where = {
        ...options.where,
        name: {
          [Op.like]: '%' + searchValue + '%'
        }
      };
    }

    let gatewayWhereClause = {
      is_unique: 1
    };
    let gatewayIds = [];
    if (gateways && gateways.length) {
      gatewayIds = convertToArray(gateways);
      gatewayWhereClause = {
        id: {
          [Op.in]: gatewayIds
        },
        is_unique: 1
      };
      options.where.$or = [
        {
          all_clients: true
        },
        models.sequelize.literal(`
        (
          SELECT
              \`GatewayPolicies\`.\`id\`
          FROM
              \`GatewayPolicies\` AS \`GatewayPolicies\`
          INNER JOIN \`Clients\` AS \`Client\` ON
              \`GatewayPolicies\`.\`client_id\` = \`Client\`.\`id\`
              AND \`Client\`.\`id\` IN (${gatewayIds.join(',')})
          WHERE
              (\`RoutingPolicies\`.\`id\` = \`GatewayPolicies\`.\`policy_id\`
              AND \`GatewayPolicies\`.\`policy_type_id\` = 3)
          LIMIT 1 ) IS NOT NULL
        `)
      ];
    }

    const result = await models.RoutingPolicies.paginate({
      ...options
    });
    for (let j = 0; j < result.data.length; j++) {
      const gateways = await models.sequelize.query(`
        SELECT Name FROM Gateways WHERE is_unique = 1 AND id IN (
          SELECT gateway_id FROM GatewayPolicies
          WHERE policy_id = ${result.data[j].id} AND policy_type_id = 3
        );
      `);
      const services = await models.sequelize.query(`
        SELECT Name FROM Services WHERE id IN (
          SELECT service_id FROM ServicePolicies
          WHERE policy_id = ${result.data[j].id} AND policy_type_id = 3
        );
      `);
      const gatewayGroups = await models.sequelize.query(`
        SELECT name FROM GatewayGroups WHERE id IN (
          SELECT gateway_group_id FROM GatewayGroupPolicies
          WHERE policy_id = ${result.data[j].id} AND policy_type_id = 3
        );
      `);
      const serviceGroups = await models.sequelize.query(`
        SELECT name FROM ServiceGroups WHERE id IN (
          SELECT service_group_id FROM ServiceGroupPolicies
          WHERE policy_id = ${result.data[j].id} AND policy_type_id = 3
        );
      `);
      result.data[j].dataValues.Gateways = gateways[0];
      result.data[j].dataValues.Services = services[0];
      result.data[j].dataValues.GatewayGroups = gatewayGroups[0];
      result.data[j].dataValues.ServiceGroups = serviceGroups[0];
    }
    res.json(result);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the routing policy by id
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getRoutingPolicyById = async (req, res, next) => {
  const { id } = req.params;
  try {
    let policy = await models.RoutingPolicies.findOne({
      where: {
        id
      }
    });
    if (!policy) throw new httpErrors.NotFound('Routing policy not found');
    const gateways = await models.sequelize.query(`
      SELECT id, Name, proxy_enabled FROM Gateways
      WHERE is_unique = 1 AND id IN (
        SELECT gateway_id FROM GatewayPolicies
        WHERE policy_id = ${id} AND policy_type_id = 3
      );
    `);
    const services = await models.sequelize.query(`
      SELECT id, Name, proxy_enabled FROM Services
      WHERE id IN (
        SELECT service_id FROM ServicePolicies
        WHERE policy_id = ${id} AND policy_type_id = 3
      );
    `);
    const gatewayGroups = await models.sequelize.query(`
      SELECT id, name FROM GatewayGroups
      WHERE id IN (
        SELECT gateway_group_id FROM GatewayGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 3
      );
    `);
    const serviceGroups = await models.sequelize.query(`
      SELECT id, name FROM ServiceGroups
      WHERE id IN (
        SELECT service_group_id FROM ServiceGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 3
      );
    `);
    policy.dataValues = {
      ...policy.dataValues,
      Gateways: gateways[0],
      Services: services[0],
      GatewayGroups: gatewayGroups[0],
      ServiceGroups: serviceGroups[0]
    };

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the list of device polices
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.createRoutingPolicy = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();

  try {
    let {
      name,
      gateway,
      gatewayGroup,
      service,
      serviceGroup,
      all_gateways,
      all_services
    } = req.body;

    const existingPolicy = await models.RoutingPolicies.findOne({
      where: {
        name
      }
    });

    if (existingPolicy) {
      throw new httpErrors.Conflict('Policy with this name already exists');
    }

    let clonedGWArray = [];
    for (let j = 0; j < gateway.length; j++) {
      let currGW = await models.Gateway.findOne({
        where: {
          id: gateway[j]
        },
        attributes: ['Name']
      });

      let clonedGW = await models.Gateway.findAll({
        where: {
          Name: {
            [Op.like]: `%${currGW.Name.trim()}%`
          }
        },
        attributes: ['id']
      });

      clonedGW = clonedGW.map(gateway => gateway.id);

      for (let k = 0; k < clonedGW.length; k++) {
        if (!gateway.includes(clonedGW[k])) clonedGWArray.push(clonedGW[k]);
      }
    }
    gateway = [...gateway, ...clonedGWArray];

    let anyProxyService = false;
    let allProxyGateway = false;
    if (all_services) {
      const allProxies = await models.sequelize.query(
        'SELECT proxy_enabled FROM Services group by proxy_enabled'
      );
      if (allProxies[0].some(service => service.proxy_enabled)) {
        anyProxyService = true;
      }
    } else {
      const allServicesProxy = await models.Service.findAll({
        where: {
          id: service
        },
        attributes: ['proxy_enabled']
      });

      if (allServicesProxy.some(service => service.proxy_enabled)) {
        anyProxyService = true;
      }
    }

    if (anyProxyService) {
      if (all_gateways) {
        const allProxies = await models.sequelize.query(
          'SELECT proxy_enabled FROM Gateways group by proxy_enabled'
        );
        if (
          allProxies[0].every(gateway => {
            return gateway.proxy_enabled;
          })
        ) {
          allProxyGateway = true;
        }
      } else {
        const allGatewaysProxy = await models.Gateway.findAll({
          where: {
            id: gateway
          },
          attributes: ['proxy_enabled']
        });

        console.log('found gateways');

        if (
          allGatewaysProxy.every(gateway => {
            return gateway.dataValues.proxy_enabled;
          })
        ) {
          allProxyGateway = true;
        }
      }

      if (!allProxyGateway) {
        throw new createHttpError(
          400,
          'Proxied services cannot be assigned to non-proxied gateways'
        );
      }
    }

    const policyObject = {
      name,
      all_gateways,
      all_services,
      is_enabled: true
    };

    const policy = await models.RoutingPolicies.create(policyObject, {
      transaction
    });

    const GRPolicies = gateway.map(c => {
      return {
        gateway_id: c,
        policy_type_id: 3,
        policy_id: policy.id
      };
    });
    const GGroupRPolicies = gatewayGroup.map(c => {
      return {
        gateway_group_id: c,
        policy_type_id: 3,
        policy_id: policy.id
      };
    });
    const SRPolicies = service.map(c => {
      return {
        service_id: c,
        policy_type_id: 3,
        policy_id: policy.id
      };
    });
    const SGroupRPolicies = serviceGroup.map(c => {
      return {
        service_group_id: c,
        policy_type_id: 3,
        policy_id: policy.id
      };
    });

    if (!all_gateways) {
      await models.GatewayPolicies.bulkCreate(GRPolicies, { transaction });
      await models.GatewayGroupPolicies.bulkCreate(GGroupRPolicies, {
        transaction
      });
    }
    if (!all_services) {
      await models.ServicePolicies.bulkCreate(SRPolicies, { transaction });
      await models.ServiceGroupPolicies.bulkCreate(SGroupRPolicies, {
        transaction
      });
    }
    await transaction.commit();

    await publishRoutingPolicy(actionType.ADDED, {
      id: policy.id,
      mode: 3
    });

    insertIntoAdminLogs(
      getPolicyCreateMessage(policy.name, policyTypes.ROUTING_POLICY),
      req
    );

    return res.json(policy);
  } catch (error) {
    transaction.rollback();
    next(error);
  }
};

/**
 * Update routing policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.updateRoutingPolicy = async (req, res, next) => {
  const { id } = req.params;

  const {
    name,
    gateway,
    gatewayGroup,
    service,
    serviceGroup,
    all_gateways,
    all_services
  } = req.body;

  const transaction = await models.sequelize.transaction();
  try {
    const policy = await models.RoutingPolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Routing policy not found');

    if(!(policy.name === name)) {
      const duplicateName = await models.RoutingPolicies.findOne({
        where: {
          name
        }
      });

      if (duplicateName) {
        throw new httpErrors.BadRequest('Policy with this name already exists');
      }
    }

    // let GRPolicies = await models.GatewayPolicies.findAll({
    //   where: { policy_id: id, policy_type_id: 3 }
    // });
    // let SRPolicies = await models.ServicePolicies.findAll({
    //   where: { policy_id: id, policy_type_id: 3 }
    // });
    // let GGroupRPolicies = await models.GatewayGroupPolicies.findAll({
    //   where: { policy_id: id, policy_type_id: 3 }
    // });
    // let SGroupRPolicies = await models.ServiceGroupPolicies.findAll({
    //   where: { policy_id: id, policy_type_id: 3 }
    // });

    let GRPolicies = await models.sequelize.query(
      `select * from GatewayPolicies where policy_id = ${id} and policy_type_id = 3`
    );

    let SRPolicies = await models.sequelize.query(
      `select * from ServicePolicies where policy_id = ${id} and policy_type_id = 3`
    );

    let GGroupRPolicies = await models.sequelize.query(
      `select * from GatewayGroupPolicies where policy_id = ${id} and policy_type_id = 3`
    );

    let SGroupRPolicies = await models.sequelize.query(
      `select * from ServiceGroupPolicies where policy_id = ${id} and policy_type_id = 3`
    );

    GRPolicies = (GRPolicies[0] || []).map(v => v.gateway_id);
    GGroupRPolicies = (GGroupRPolicies[0] || []).map(v => v.gateway_group_id);
    SRPolicies = (SRPolicies[0] || []).map(v => v.service_id);
    SGroupRPolicies = (SGroupRPolicies[0] || []).map(v => v.service_group_id);

    let gatewayToAdd = all_gateways ? [] : _.difference(gateway, GRPolicies);
    let gatewayToDelete = all_gateways
      ? GRPolicies
      : _.difference(GRPolicies, gateway);
    const gatewayGroupToAdd = all_gateways
      ? []
      : _.difference(gatewayGroup, GGroupRPolicies);
    const gatewayGroupToDelete = all_gateways
      ? GGroupRPolicies
      : _.difference(GGroupRPolicies, gatewayGroup);
    const serviceToAdd = all_services ? [] : _.difference(service, SRPolicies);
    const serviceToDelete = all_services
      ? SRPolicies
      : _.difference(SRPolicies, service);
    const serviceGroupToAdd = all_services
      ? []
      : _.difference(serviceGroup, SGroupRPolicies);
    const serviceGroupToDelete = all_services
      ? SGroupRPolicies
      : _.difference(SGroupRPolicies, serviceGroup);

    let clonedGWArray = [];
    for (let j = 0; j < gatewayToAdd.length; j++) {
      let currGW = await models.Gateway.findOne({
        where: {
          id: gatewayToAdd[j]
        },
        attributes: ['Name']
      });

      let clonedGW = await models.Gateway.findAll({
        where: {
          Name: {
            [Op.like]: `%${currGW.Name.trim()}%`
          }
        },
        attributes: ['id']
      });

      clonedGW = clonedGW.map(gateway => gateway.id);

      for (let k = 0; k < clonedGW.length; k++) {
        if (!gatewayToAdd.includes(clonedGW[k]))
          clonedGWArray.push(clonedGW[k]);
      }
    }
    gatewayToAdd = [...gatewayToAdd, ...clonedGWArray];

    clonedGWArray = [];
    for (let j = 0; j < gatewayToDelete.length; j++) {
      let currGW = await models.Gateway.findOne({
        where: {
          id: gatewayToDelete[j]
        },
        attributes: ['Name']
      });
      console.log('lajsdf');

      let clonedGW = await models.Gateway.findAll({
        where: {
          Name: {
            [Op.like]: `%${currGW.Name.trim()}%`
          }
        },
        attributes: ['id']
      });

      clonedGW = clonedGW.map(gateway => gateway.id);

      for (let k = 0; k < clonedGW.length; k++) {
        if (!gatewayToDelete.includes(clonedGW[k]))
          clonedGWArray.push(clonedGW[k]);
      }
    }
    gatewayToDelete = [...gatewayToDelete, ...clonedGWArray];

    let anyProxyService = false;
    let allProxyGateway = false;
    if (all_services) {
      const allProxies = await models.sequelize.query(
        'SELECT proxy_enabled FROM Services group by proxy_enabled'
      );
      if (allProxies[0].some(service => service.proxy_enabled)) {
        anyProxyService = true;
      }
    } else {
      const serviceIds = [...serviceToAdd, ...SRPolicies];
      const allServicesProxy = await models.Service.findAll({
        where: {
          id: serviceIds
        },
        attributes: ['proxy_enabled']
      });

      if (allServicesProxy.some(service => service.proxy_enabled)) {
        anyProxyService = true;
      }
    }

    if (anyProxyService) {
      if (all_gateways) {
        const allProxies = await models.sequelize.query(
          'SELECT proxy_enabled FROM Gateways group by proxy_enabled'
        );
        if (
          allProxies[0].every(gateway => {
            return gateway.proxy_enabled;
          })
        ) {
          allProxyGateway = true;
        }
      } else {
        const gatewayIds = [...serviceToAdd, ...SRPolicies];

        const allGatewaysProxy = await models.Gateway.findAll({
          where: {
            id: gatewayIds
          },
          attributes: ['proxy_enabled']
        });

        if (
          allGatewaysProxy.every(gateway => {
            return gateway.dataValues.proxy_enabled;
          })
        ) {
          allProxyGateway = true;
        }
      }

      if (!allProxyGateway) {
        throw new createHttpError(
          400,
          'Proxied services cannot be assigned to non-proxied gateways'
        );
      }
    }

    await Promise.all([
      models.GatewayPolicies.bulkCreate(
        gatewayToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 3,
          gateway_id: v
        })),
        { transaction }
      ),
      models.GatewayPolicies.destroy({
        where: {
          gateway_id: {
            [Op.in]: gatewayToDelete
          },
          policy_type_id: 3,
          policy_id: policy.id
        },
        transaction
      }),
      models.GatewayGroupPolicies.bulkCreate(
        gatewayGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 3,
          gateway_group_id: v
        })),
        { transaction }
      ),
      models.GatewayGroupPolicies.destroy({
        where: {
          gateway_group_id: {
            [Op.in]: gatewayGroupToDelete
          },
          policy_type_id: 3,
          policy_id: policy.id
        },
        transaction
      }),

      models.ServicePolicies.bulkCreate(
        serviceToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 3,
          service_id: v
        })),
        { transaction }
      ),
      models.ServicePolicies.destroy({
        where: {
          service_id: {
            [Op.in]: serviceToDelete
          },
          policy_type_id: 3,
          policy_id: policy.id
        },
        transaction
      }),
      models.ServiceGroupPolicies.bulkCreate(
        serviceGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 3,
          service_group_id: v
        })),
        { transaction }
      ),
      models.ServiceGroupPolicies.destroy({
        where: {
          service_group_id: {
            [Op.in]: serviceGroupToDelete
          },
          policy_type_id: 3,
          policy_id: policy.id
        },
        transaction
      }),
      await policy.update(
        {
          name,
          all_gateways,
          all_services
        },
        {
          transaction,
          updateData: {
            gateway,
            gatewayGroup,
            service,
            serviceGroup,
            all_gateways,
            all_services
          }
        }
      )
    ]);

    await transaction.commit();

    await publishRoutingPolicy(actionType.UPDATED, {
      id: policy.id,
      mode:
        (gatewayToAdd.length ||
          gatewayToDelete.length ||
          gatewayGroupToAdd.length ||
          gatewayGroupToDelete.length) &&
        (serviceToAdd.length ||
          serviceToDelete.length ||
          serviceGroupToAdd.length ||
          serviceGroupToDelete.length)
          ? 3
          : gatewayToAdd.length ||
            gatewayToDelete.length ||
            gatewayGroupToAdd.length ||
            gatewayGroupToDelete.length
          ? 1
          : 2
    });

    insertIntoAdminLogs(
      getPolicyEditMessage(policy.name, policyTypes.ROUTING_POLICY),
      req
    );
    return res.json(policy);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/**
 * Delete Routing policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.deleteRoutingPolicy = async (req, res, next) => {
  const { id } = req.params;
  //const transaction = await models.sequelize.transaction();
  try {
    const policy = await models.RoutingPolicies.findOne({
      where: { id }
      //transaction
    });

    if (!policy) throw new httpErrors.NotFound('Routing policy not found');
    // const transaction = await models.sequelize.transaction();
    await policy.destroy();

    const isRoutingPlocyinServiceGroup = await models.ServiceGroupPolicies.findOne(
      {
        where: { policy_id: id }
      }
    );
    if (isRoutingPlocyinServiceGroup) {
      await models.sequelize.query(
        `DELETE FROM sdp.ServiceGroupPolicies WHERE policy_id=${id};`
      );
    }

    const isRoutingPlocyinService = await models.ServicePolicies.findOne({
      where: { policy_id: id }
    });
    if (isRoutingPlocyinService) {
      await models.sequelize.query(
        `DELETE FROM sdp.ServicePolicies WHERE policy_id=${id};`
      );
    }

    const isRoutingPlocyinGatewayGroup = await models.GatewayGroupPolicies.findOne(
      {
        where: { policy_id: id }
      }
    );
    if (isRoutingPlocyinGatewayGroup) {
      await models.sequelize.query(
        `DELETE FROM sdp.GatewayGroupPolicies WHERE policy_id=${id};`
      );
    }

    const isRoutingPlocyinGateway = await models.GatewayPolicies.findOne({
      where: { policy_id: id }
    });
    if (isRoutingPlocyinGateway) {
      await models.sequelize.query(
        `DELETE FROM sdp.GatewayPolicies WHERE policy_id=${id};`
      );
    }

    await publishRoutingPolicy(actionType.DELETED, {
      id: policy.id,
      mode: 3
    });

    insertIntoAdminLogs(
      getPolicyDeleteMessage(policy.name, policyTypes.ROUTING_POLICY),
      req
    );

    // await transaction.commit();
    return res.json(policy);
  } catch (error) {
    // await transaction.rollback();
    next(error);
  }
};

/**
 * Enable/disable device policy
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */

module.exports.setRoutingEnableValue = async (req, res, next) => {
  const { id, value } = req.params;
  try {
    const policy = await models.RoutingPolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Routing Policy not found');
    await policy.update(
      { is_enabled: value },
      {
        hooks: false
      }
    );

    let logMessage;
    if (value === 'true') {
      logMessage = getPolicyEnableMessage(
        policy.name,
        policyTypes.ROUTING_POLICY
      );
    } else {
      logMessage = getPolicyDisableMessage(
        policy.name,
        policyTypes.ROUTING_POLICY
      );
    }

    insertIntoAdminLogs(logMessage, req);

    await publishRoutingPolicy(
      value == 'true' ? actionType.ADDED : actionType.DELETED,
      {
        id: policy.id,
        mode: 3
      }
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

/** get all gateways
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getAllGatewaysByName = async (req, res, next) => {
  try {
    const {
      limit = 10,
      offset = 0,
      dir = 'asc',
      sort = 'Name',
      name
    } = req.query;

    const options = {
      where: {},
      order: [],
      attributes: ['id', 'Name'],
      limit: +limit,
      offset: +offset * +limit || 0
    };

    const allowedSortValues = ['Name'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    if (name) {
      options.where = {
        ...options.where,
        Name: {
          [Op.like]: name + '%'
        }
      };
    }

    const gateways = await models.Gateway.findAll(options);

    return res.json(gateways);
  } catch (error) {
    next(error);
  }
};

/**
 * Get the services, gateways, gateway groups and service groups
 * @param {Request} req
 * @param {Response} res
 * @param {Function} next
 * @return {Object}
 */
module.exports.getAllGatwayAndServices = async (req, res, next) => {
  const policyId = verifyPolicyId(req.query.policyId)
    ? req.query.policyId
    : false;
  const gatewayName = isSafeString(req.query.gatewayName)
    ? req.query.gatewayName
    : null;
  const serviceName = isSafeString(req.query.serviceName)
    ? req.query.serviceName
    : null;
  let gatewayOffset = +req.query.gatewayOffset || 0;
  let serviceOffset = +req.query.serviceOffset || 0;

  const limit = 10;
  gatewayOffset = gatewayOffset * limit;
  serviceOffset = serviceOffset * limit;
  let where = 'where 1=1 ';
  let gatewayWhere = where + ` and is_unique = 1 `;
  let gatewayGroupWhere = 'where TRUE';
  let serviceWhere = where;
  if (gatewayName && gatewayName !== 'null') {
    gatewayWhere += ` and (c.Name LIKE '%${gatewayName}%' or c.id like '%${gatewayName}%') `;
    gatewayGroupWhere += ` and (g.name LIKE '%${gatewayName}%' or g.id like '%${gatewayName}%') `;
  }
  if (serviceName && serviceName !== 'null') {
    serviceWhere += ` and name LIKE '%${serviceName}%' `;
  }

  let gatewayLeftJoinQuery = '';
  let serviceLeftJoinQuery = '';
  let gatewayGroupLeftJoinQuery = '';
  let serviceGroupLeftJoinQuery = '';

  if (policyId) {
    gatewayLeftJoinQuery = `left join GatewayPolicies cp on cp.policy_type_id = 3 and cp.gateway_id = c.id and cp.policy_id  = ${policyId}`;
    serviceLeftJoinQuery = `left join ServicePolicies sp on sp.policy_type_id = 3 and sp.service_id = s.id and sp.policy_id = ${policyId}`;
    serviceGroupLeftJoinQuery = `left join ServiceGroupPolicies sgp on sgp.policy_type_id = 3 and sgp.service_group_id = sg.id and sgp.policy_id = ${policyId}`;
    gatewayGroupLeftJoinQuery = `left join GatewayGroupPolicies cgp on cgp.policy_type_id = 3 and cgp.gateway_group_id = g.id and cgp.policy_id = ${policyId}`;
  }

  try {
    const query = `
    (SELECT c.id ,c.Name as name, c.buildOs,"gateway" as type, c.proxy_enabled as proxy from Gateways c ${gatewayLeftJoinQuery} ${gatewayWhere} LIMIT ${gatewayOffset},${limit})
    UNION ALL
    (SELECT s.id,s.name,s.port, "service" as type,s.proxy_enabled as proxy from Services s ${serviceLeftJoinQuery} ${serviceWhere} LIMIT ${serviceOffset},${limit})
    UNION ALL
    (SELECT g.id ,g.name,g.createdAt,"gatewayGroup" as type, g.updatedAt  from GatewayGroups g ${gatewayGroupLeftJoinQuery} ${gatewayGroupWhere} LIMIT ${gatewayOffset},${limit})
    UNION ALL
    (SELECT sg.id,sg.name, sg.updatedAt, "serviceGroup" as type, 0 as is_deleted from ServiceGroups sg ${serviceGroupLeftJoinQuery} ${serviceWhere} LIMIT ${serviceOffset},${limit})
    `;

    console.log('running query', query);

    const result = await models.sequelize.query(query, {
      type: models.sequelize.QueryTypes.SELECT
    });
    return res.json(result);
  } catch (error) {
    next(error);
  }
};

module.exports.getDataVolumePolicies = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchValue,
      clients = [],
      allClientPolicies = false
    } = req.query;

    const options = {
      where: {},
      limit,
      offset,
      order: []
      // distinct: true
    };

    const allowedSortValues = [
      'id',
      'name',
      'is_blocked',
      'client_download_rate',
      'client_upload_rate'
    ];

    if (allowedSortValues.includes(sort)) {
      options.order.push([sort, dir]);
    }

    if (searchValue) {
      options.where = {
        [Op.or]: {
          name: {
            [Op.like]: `%${searchValue.trim()}%`
          }
        }
      };
    }

    let clientIds = [];
    if (clients && clients.length) {
      clientIds = convertToArray(clients).map(id => parseInt(id, 10)).filter(id => !isNaN(id));
      
      options.where[Op.or] = [
        { all_clients: true },
        models.sequelize.literal(`
          EXISTS (
            SELECT 1 
            FROM ClientPolicies 
            INNER JOIN Clients ON ClientPolicies.client_id = Clients.id
            WHERE ClientPolicies.policy_id = DataVolumePolicies.id 
            AND Clients.id IN (?)
            AND ClientPolicies.policy_type_id = 6
          ) OR EXISTS (
            SELECT 1 
            FROM GroupClients 
            INNER JOIN ClientGroupPolicies ON ClientGroupPolicies.client_group_id = GroupClients.groupId
            INNER JOIN Clients ON GroupClients.clientId = Clients.id
            WHERE ClientGroupPolicies.policy_id = DataVolumePolicies.id 
            AND Clients.id IN (?)
            AND ClientGroupPolicies.policy_type_id = 6
          )
        `)
      ];

      // When executing, provide replacements
      options.replacements = [clientIds, clientIds];
    }

    if (allClientPolicies == 'true') {
      options.where.all_clients = true;
    }

    const result = await models.DataVolumePolicies.paginate({
      ...options
    });

    const enrichedData = await Promise.all(result.data.map(async (item) => {
      const [userGroups, users, clients, services, groups, serviceGroups] = await Promise.all([
        models.sequelize.query(
          'SELECT group_name as name FROM CustomUserGroups WHERE id IN (SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ? AND policy_type_id = 6)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name as Name FROM LocalUsers WHERE id IN (SELECT user_id FROM UserPolicies WHERE policy_id = ? AND policy_type_id = 6)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT Name FROM Clients WHERE id IN (SELECT client_id FROM ClientPolicies WHERE policy_id = ? AND policy_type_id = 6)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT Name FROM Services WHERE id IN (SELECT service_id FROM ServicePolicies WHERE policy_id = ? AND policy_type_id = 6)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name FROM `Groups` WHERE id IN (SELECT client_group_id FROM ClientGroupPolicies WHERE policy_id = ? AND policy_type_id = 6)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        ),
        models.sequelize.query(
          'SELECT name FROM ServiceGroups WHERE id IN (SELECT service_group_id FROM ServiceGroupPolicies WHERE policy_id = ? AND policy_type_id = 6)',
          { replacements: [item.id], type: models.sequelize.QueryTypes.SELECT }
        )
      ]);
    
      return {
        ...item.dataValues,
        UserGroups: userGroups || [],
        Users: users || [],
        Clients: clients,
        Services: services,
        Groups: groups,
        ServiceGroups: serviceGroups
      };
    }));

    result.data = enrichedData;

    res.json(result);
  } catch (error) {
    next(error);
  }
};

module.exports.createDataVolumePolicy = async (req, res, next) => {
  const {
    name,
    is_blocked,
    client,
    client_upload_rate,
    client_download_rate,
    selectedSoftwares,
    clientGroup,
    service,
    serviceGroup,
    all_clients,
    all_services,
    user,
    userGroup,
    all_users
  } = req.body;

  const policyObject = {
    name,
    is_blocked,
    client,
    client_upload_rate,
    client_download_rate,
    clientGroup,
    service,
    all_clients,
    all_services,
    all_users
  };

  try {
    // check if policy with same name already exists

    const result = await models.DataVolumePolicies.count({
      where: { name }
    });

    if (result)
      throw new httpErrors.Conflict('Policy with this name already exists');

    const transaction = await models.sequelize.transaction();
    const policy = await models.DataVolumePolicies.create(policyObject, {
      transaction
    });
    const CPPolicies = client.map(c => {
      return {
        client_id: c,
        policy_type_id: 6,
        policy_id: policy.id
      };
    });
    const CGroupPPolicies = clientGroup.map(c => {
      return {
        client_group_id: c,
        policy_type_id: 6,
        policy_id: policy.id
      };
    });
    const SPPolicies = service.map(c => {
      return {
        service_id: c,
        policy_type_id: 6,
        policy_id: policy.id
      };
    });
    const SGroupPPolicies = serviceGroup.map(c => {
      return {
        service_group_id: c,
        policy_type_id: 6,
        policy_id: policy.id
      };
    });
    const UPolicies = user.map(c => {
      return {
        user_id: c,
        policy_type_id: 6,
        policy_id: policy.id
      };
    });
    const UGroupPolicies = userGroup.map(c => {
      return {
        user_group_id: c,
        policy_type_id: 6,
        policy_id: policy.id,
        user_group_type: 1
      };
    });

    if (!all_clients) {
      await models.ClientPolicies.bulkCreate(CPPolicies, { transaction });
      await models.ClientGroupPolicies.bulkCreate(CGroupPPolicies, {
        transaction
      });
    }
    if (!all_services) {
      await models.ServicePolicies.bulkCreate(SPPolicies, { transaction });
      await models.ServiceGroupPolicies.bulkCreate(SGroupPPolicies, {
        transaction
      });
    }
    if (!all_users) {
      await models.UserPolicies.bulkCreate(UPolicies, { transaction });
      await models.UserGroupPolicies.bulkCreate(UGroupPolicies, {
        transaction
      });
    }
    await transaction.commit();

    insertIntoAdminLogs(
      getPolicyCreateMessage(policy.name, policyTypes.DATA_VOLUME_POLICY),
      req
    );
    await publishDataVolumePolicy(actionType.ADDED, {
      id: policy.id,
      mode: 2,
      data: {
        name: policy.name,
        all_clients: +policy.all_clients,
        all_services: +policy.all_services,
        is_blocked: +policy.is_blocked,
        all_users: +policy.all_users
      }
    });

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

module.exports.getDataVolumePoliciesById = async (req, res, next) => {
  const { id } = req.params;
  try {
    let policy = await models.DataVolumePolicies.findOne({
      where: {
        id
      }
    });
    if (!policy) throw new httpErrors.NotFound('Software policy not found');
    const clients = await models.sequelize.query(`
      SELECT id, Name, device_os FROM Clients
      WHERE id IN (
        SELECT client_id FROM ClientPolicies
        WHERE policy_id = ${id} AND policy_type_id = 6
      );
    `);
    const services = await models.sequelize.query(`
      SELECT id, Name FROM Services
      WHERE id IN (
        SELECT service_id FROM ServicePolicies
        WHERE policy_id = ${id} AND policy_type_id = 6
      );
    `);
    const groups = await models.sequelize.query(`
      SELECT id, name FROM \`Groups\`
      WHERE id IN (
        SELECT client_group_id FROM ClientGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 6
      );
    `);
    const serviceGroups = await models.sequelize.query(`
      SELECT id, name FROM ServiceGroups
      WHERE id IN (
        SELECT service_group_id FROM ServiceGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 6
      );
    `);
    const users = await models.sequelize.query(`
      SELECT id, Name, runtime_bind_id FROM LocalUsers
      WHERE id IN (
        SELECT user_id FROM UserPolicies
        WHERE policy_id = ${id} AND policy_type_id = 6
      );
    `);
    const userGroups = await models.sequelize.query(`
      SELECT id, group_name as name, runtime_bind_id, raw_query_id FROM CustomUserGroups
      WHERE id IN (
        SELECT user_group_id FROM UserGroupPolicies
        WHERE policy_id = ${id} AND policy_type_id = 6
      );
    `);
    policy.dataValues = {
      ...policy.dataValues,
      Clients: clients[0],
      Services: services[0],
      Groups: groups[0],
      ServiceGroups: serviceGroups[0],
      Users: users[0],
      UserGroups: userGroups[0]
    };
    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

module.exports.updateDataVolumePolicies = async (req, res, next) => {
  const { id } = req.params;

  const {
    name,
    is_blocked,
    client,
    clientGroup,
    all_clients,
    all_services,
    service,
    serviceGroup,
    client_upload_rate,
    client_download_rate,
    all_users,
    userGroup,
    user
  } = req.body;

  const transaction = await models.sequelize.transaction();

  try {
    const policy = await models.DataVolumePolicies.findOne({
      where: { id },
      transaction
    });

    if (!policy)
      throw new httpErrors.NotFound('Software Process policy not found');

    if(!(policy.name === name)) {
      const duplicateName = await models.DataVolumePolicies.findOne({
        where: {
          name
        }
      });

      if (duplicateName) {
        throw new httpErrors.BadRequest('Policy with this name already exists');
      }
    }

    let CPPolicies = await models.sequelize.query(
      `select * from ClientPolicies where policy_id = ${id} and policy_type_id = 6`
    );

    let SPPolicies = await models.sequelize.query(
      `select * from ServicePolicies where policy_id = ${id} and policy_type_id = 6`
    );

    let CGroupPPolicies = await models.sequelize.query(
      `select * from ClientGroupPolicies where policy_id = ${id} and policy_type_id = 6`
    );

    let SGroupPPolicies = await models.sequelize.query(
      `select * from ServiceGroupPolicies where policy_id = ${id} and policy_type_id = 6`
    );

    let UPolicies = await models.sequelize.query(
      `SELECT * FROM UserPolicies WHERE policy_id = ${id} AND policy_type_id = 6`
    );

    let UGroupPolicies = await models.sequelize.query(
      `SELECT user_group_id FROM UserGroupPolicies WHERE policy_id = ${id} AND policy_type_id = 6`
    );

    CPPolicies = (CPPolicies[0] || []).map(v => v.client_id);
    CGroupPPolicies = (CGroupPPolicies[0] || []).map(v => v.client_group_id);
    SPPolicies = (SPPolicies[0] || []).map(v => v.service_id);
    SGroupPPolicies = (SGroupPPolicies[0] || []).map(v => v.service_group_id);
    UPolicies = (UPolicies[0] || []).map(v => v.user_id);
    UGroupPolicies = (UGroupPolicies[0] || []).map(v => v.user_group_id);

    const clientToAdd = all_clients ? [] : _.difference(client, CPPolicies);
    const clientToDelete = all_clients
      ? CPPolicies
      : _.difference(CPPolicies, client);
    const clientGroupToAdd = all_clients
      ? []
      : _.difference(clientGroup, CGroupPPolicies);
    const clientGroupToDelete = all_clients
      ? CGroupPPolicies
      : _.difference(CGroupPPolicies, clientGroup);
    const serviceToAdd = all_services ? [] : _.difference(service, SPPolicies);
    const serviceToDelete = all_services
      ? SPPolicies
      : _.difference(SPPolicies, service);
    const serviceGroupToAdd = all_services
      ? []
      : _.difference(serviceGroup, SGroupPPolicies);
    const serviceGroupToDelete = all_services
      ? SGroupPPolicies
      : _.difference(SGroupPPolicies, serviceGroup);
    const userToAdd = all_users ? [] : _.difference(user, UPolicies);
    const userToDelete = all_users ? UPolicies : _.difference(UPolicies, user);
    const userGroupToAdd = all_users
      ? []
      : _.difference(userGroup, UGroupPolicies);
    const userGroupToDelete = all_users
      ? UGroupPolicies
      : _.difference(UGroupPolicies, userGroup);

    const [, , , , , , , , , , , , updatedPolicy] = await Promise.all([
      models.ClientPolicies.bulkCreate(
        clientToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 6,
          client_id: v
        })),
        { transaction }
      ),
      models.ClientPolicies.destroy({
        where: {
          client_id: {
            [Op.in]: clientToDelete
          },
          policy_type_id: 6,
          policy_id: policy.id
        },
        transaction
      }),
      models.ClientGroupPolicies.bulkCreate(
        clientGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 6,
          client_group_id: v
        })),
        { transaction }
      ),
      models.ClientGroupPolicies.destroy({
        where: {
          client_group_id: {
            [Op.in]: clientGroupToDelete
          },
          policy_type_id: 6,
          policy_id: policy.id
        },
        transaction
      }),

      models.ServicePolicies.bulkCreate(
        serviceToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 6,
          service_id: v
        })),
        { transaction }
      ),
      models.ServicePolicies.destroy({
        where: {
          service_id: {
            [Op.in]: serviceToDelete
          },
          policy_type_id: 6,
          policy_id: policy.id
        },
        transaction
      }),
      models.ServiceGroupPolicies.bulkCreate(
        serviceGroupToAdd.map(v => ({
          policy_id: policy.id,
          policy_type_id: 6,
          service_group_id: v
        })),
        { transaction }
      ),
      models.ServiceGroupPolicies.destroy({
        where: {
          service_group_id: {
            [Op.in]: serviceGroupToDelete
          },
          policy_type_id: 6,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserPolicies.bulkCreate(
        userToAdd.map(c => {
          return {
            user_id: c,
            policy_type_id: 6,
            policy_id: policy.id
          };
        }, transaction)
      ),
      models.UserGroupPolicies.bulkCreate(
        userGroupToAdd.map(c => {
          return {
            user_group_id: c,
            policy_type_id: 6,
            policy_id: policy.id,
            user_group_type: 1
          };
        }, transaction)
      ),
      models.UserPolicies.destroy({
        where: {
          user_id: {
            [Op.in]: userToDelete
          },
          policy_type_id: 6,
          policy_id: policy.id
        },
        transaction
      }),
      models.UserGroupPolicies.destroy({
        where: {
          user_group_id: {
            [Op.in]: userGroupToDelete
          },
          policy_type_id: 6,
          policy_id: policy.id
        },
        transaction
      }),

      await policy.update(
        {
          name,
          is_blocked,
          client_upload_rate,
          client_download_rate,
          all_clients,
          all_services,
          all_users
        },
        {
          transaction,
          updateData: {
            client,
            clientGroup,
            service,
            serviceGroup,
            client_upload_rate,
            client_download_rate,
            all_clients,
            all_services,
            all_users
          }
        }
      )
    ]);

    await transaction.commit();

    await publishDataVolumePolicy(actionType.UPDATED, {
      id: updatedPolicy.id,
      data: {
        name: updatedPolicy.name,
        all_clients: +updatedPolicy.all_clients,
        all_services: +updatedPolicy.all_services,
        is_blocked: +updatedPolicy.is_blocked,
        client_upload_rate: updatedPolicy.client_upload_rate,
        client_download_rate: updatedPolicy.client_download_rate,
        all_users: +updatedPolicy.all_users
      }
    });

    insertIntoAdminLogs(
      getPolicyEditMessage(policy.name, policyTypes.DATA_VOLUME_POLICY),
      req
    );
    return res.json(policy);
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

module.exports.setDataVolumeEnableValue = async (req, res, next) => {
  const { id, value } = req.params;
  try {
    const policy = await models.DataVolumePolicies.findOne({
      where: { id }
    });

    if (!policy) throw new httpErrors.NotFound('Routing Policy not found');
    await policy.update(
      { is_enabled: value },
      {
        hooks: false
      }
    );

    let logMessage;
    if (value === 'true') {
      logMessage = getPolicyEnableMessage(
        policy.name,
        policyTypes.DATA_VOLUME_POLICY
      );
    } else {
      logMessage = getPolicyDisableMessage(
        policy.name,
        policyTypes.DATA_VOLUME_POLICY
      );
    }

    insertIntoAdminLogs(logMessage, req);

    await publishDataVolumePolicy(
      value == 'true' ? actionType.ADDED : actionType.DELETED,
      {
        id: policy.id,
        mode: 3
      }
    );

    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

module.exports.deleteDataVolumePolicy = async (req, res, next) => {
  const { id } = req.params;
  try {
    const policy = await models.DataVolumePolicies.findOne({
      where: { id }
    });

    if (!policy)
      throw new httpErrors.NotFound('Software process policy not found');

    await Promise.all([
      models.ClientPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 6 }
      }),
      models.ServicePolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 6 }
      }),
      models.ClientGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 6 }
      }),
      models.ServiceGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 6 }
      }),
      models.UserPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 6 }
      }),
      models.UserGroupPolicies.destroy({
        where: { policy_id: policy.id, policy_type_id: 6 }
      }),
      policy.destroy()
    ]);

    await publishDataVolumePolicy(actionType.DELETED, {
      id: id,
      data: {},
      mode: 2
    });

    // insertIntoAdminLogs(
    //     getPolicyDeleteMessage(policy.name, policyTypes.SOFTWARE_POLICY),
    //     req
    // )
    // await publishProcessPolicy(actionType.DELETED, {
    //     id: policy.id,
    //     data: {}
    // });
    return res.json(policy);
  } catch (error) {
    next(error);
  }
};

function verifyPolicyId(num) {
  if (typeof num === 'number' && num > -1) {
    return true;
  } else {
    return false;
  }
}

function isSafeString(inputString) {
  const forbiddenCharacters = [';', '--', '/*', '*/', '"', "'"];

  for (let i = 0; i < forbiddenCharacters.length; i++) {
    if (inputString.includes(forbiddenCharacters[i])) {
      return false;
    }
  }

  return true;
}
