const argon2 = require('argon2');
var moment = require('moment');
var httpErrors = require('http-errors');
var jwt = require('jsonwebtoken');
var blacklist = require('express-jwt-blacklist');
var uuid = require('uuid');
var models = require('../models');
let helper = require('../routes/common/helper');
let language = require('../routes/common/language');
var projectConfig = require('../config/config')['project'][
  process.env.NODE_ENV
];
const {
  generateKeyValuePair
} = require('../utils/common');

var emailLinkExpiryInMinutes = 60;
var emailTemplate = 'email_verification_link.html';
const _ = require('lodash');
const {
  insertIntoAdminLogs
} = require('../routes/common/adminLogs');
const {
  getUserDeletMessage,
  logsPriority,
  getUpdateUserMessage,
  getUserCreateMessage
} = require('../config/adminLogs.config');

let prepareEmailOptions = (
  recipient,
  emailSubject,
  templateFileName,
  on_premise_confirm_link,
  off_premise_confirm_link,
  recipientName,
  linkExpiryInMinutes,
  invLogoLink,
  ebrLogoLink
) => {
  return {
    recipient: recipient,
    emailSubject: emailSubject,
    templateFileName: templateFileName,
    on_premise_confirm_link: on_premise_confirm_link,
    off_premise_confirm_link: off_premise_confirm_link,
    recipientName: recipientName,
    linkExpiryInMinutes: linkExpiryInMinutes,
    invLogoLink: invLogoLink,
    ebrLogoLink: ebrLogoLink
  };
};

/* GET users listing. */
const getUsers = async (req, res, next) => {
  try {
    const {
      limit,
      offset,
      sort = 'id',
      dir = 'desc',
      searchKey = ''
    } = req.query;

    const options = {
      attributes: [
        'firstName',
        'lastName',
        'id',
        'createdAt',
        'email',
        'is_locked',
        'is_locked_auto',
        'email_verified',
        'public_key',
        'secret_key'
      ],
      include: [{
        model: models.Role
      }],
      where: {
        is_active: 1
      },
      limit: limit,
      offset: offset,
      order: []
    };

    if (searchKey.trim()) {
      options.where = {
        ...options.where,
        [Op.or]: [{
            firstName: {
              [Op.like]: `${searchKey.trim()}%`
            }
          },
          {
            lastName: {
              [Op.like]: `${searchKey.trim()}%`
            }
          },
          models.sequelize.where(
            models.sequelize.fn(
              'concat',
              models.sequelize.col('firstName'),
              ' ',
              models.sequelize.col('lastName')
            ), {
              [Op.like]: `%${searchKey.trim()}%`
            }
          )
        ]
      };
    }

    var sessionUser = req.user;

    if (sessionUser.role === 'Policy-Admin') {
      options.include[0].where = {
        title: 'Policy-Admin'
      };
    }

    if (sessionUser.role === 'Team-Manager') {
      options.include[0].where = {
        title: 'Team-Manager'
      };
    }

    // if not admin or infra admin
    if (req.user.tempId !== 1 && req.user.tempId !== 2) {
      options.where.id = req.user.id;
    }

    const allowedSortValues = ['firstName', 'id'];

    if (allowedSortValues.includes(sort)) {
      options.order.push([models.sequelize.literal(sort), dir]);
    }

    const result = await models.User.paginate(options);
    const settings = await models.Settings.findOne({
      attributes: ['admin_mfa'],
      order: [
        ['id', 'DESC']
      ]
    });

    result.admin_mfa = settings.admin_mfa;

    for (let i = 0; i < result.data.length; i++) {
      let query = `
        SELECT DISTINCT t.id as temp_id, t.name, t.description
        FROM Templates t
        INNER JOIN TemplatePermissions tp ON tp.temp_id = t.id
        INNER JOIN Permissions p ON p.id = tp.permission_id
        INNER JOIN Privilege pv ON pv.id = p.privilege_id
        INNER JOIN UserTemplates ut ON ut.temp_permission_id = tp.id
        INNER JOIN Users u ON u.id = ut.user_id
        WHERE u.id = ${result.data[i].id}
      `;

      let template = (await models.sequelize.query(query))[0][0];

      if (template) {
        let templateName = template.name;

        // storing template data in user object
        result.data[i].dataValues.template = templateName;
      }
    }

    let filteredResult = {
      ...result,
      data: (result.data &&
        result.data.map(user =>
          user.id === req.user.id || isPublicUser(user) ?
          user :
          _.omit(user.dataValues, ['public_key', 'secret_key'])
        )) || []
    };

    res.json(filteredResult);
  } catch (error) {
    next(error);
  }
};

const resetMfa = async (req, res, next) => {
  try {
    await models.User.update({
      is_mfa_assigned: 0,
      otp_image_data: null,
      otp_secret: null
    }, {
      where: {
        id: req.body.id
      }
    });

    res.json('MFA reset');
  } catch (error) {
    next(error);
  }
};

/* update single user */
const updateUser = async (req, res, next) => {
  try {
    const sessionUser = req.user;
    const {
      user: reqUser
    } = req.body;

    if (!reqUser.password) {
      delete reqUser.password;
    }
    if (!reqUser.answer_1) {
      delete reqUser.answer_1;
    }
    if (!reqUser.answer_2) {
      delete reqUser.answer_2;
    }
    if (!reqUser.answer_3) {
      delete reqUser.answer_3;
    }

    const dbUser = await models.User.findOne({
      where: {
        email: reqUser.email,
        id: {
          [Op.ne]: req.params.id
        }
      }
    });

    if (dbUser) {
      dbUser.is_dark = req.body.is_dark;
      dbUser.save();

      let message = language.get['USER_EXIST_WITH_EMAIL'];
      return res.json({
        success: false,
        message: message
      });
    }

    const user = await models.User.findOne({
      where: {
        id: req.params.id
      },
      include: [{
        model: models.Role
      }]
    });

    if (!user) {
      throw new httpErrors.NotFound('No user found to update.');
    }

    let emailChanged = false;

    let key =
      Math.random()
      .toString(36)
      .substring(2) +
      Math.random()
      .toString(36)
      .substring(2);

    if (user.email !== reqUser.email) {
      let email_verification_link_expiry = moment().add(
        emailLinkExpiryInMinutes,
        'minutes'
      );
      reqUser.email_verification_link = key;
      reqUser.email_verification_link_expiry = email_verification_link_expiry;
      reqUser.email_verified = false;
      emailChanged = true;
    }

    let answer1Hash, answer2Hash, answer3Hash;
    if (reqUser.answer_1) {
      answer1Hash = await argon2.hash(reqUser.answer_1);
      if (!answer1Hash) {
        throw new httpErrors.BadRequest('Answer 1 hashing failed.');
      }
      reqUser.answer_1 = answer1Hash;
    }
    if (reqUser.answer_2) {
      answer2Hash = await argon2.hash(reqUser.answer_2);
      if (!answer2Hash) {
        throw new httpErrors.BadRequest('Answer 2 hashing failed.');
      }
      reqUser.answer_2 = answer2Hash;
    }
    if (reqUser.answer_2) {
      answer3Hash = await argon2.hash(reqUser.answer_3);
      if (!answer3Hash) {
        throw new httpErrors.BadRequest('Answer 3 hashing failed.');
      }
      reqUser.answer_3 = answer3Hash;
    }

    if (!reqUser.password) {
      await user.update(reqUser);

      if (emailChanged) {
        helper.sendVerificationEmail(
          prepareEmailOptions(
            user.email,
            language.get['USER_EMAIL_VERIFICATION'],
            emailTemplate,
            prepareOnPremiseVerificationLink(key),
            prepareOffPremiseVerificationLink(key),
            user.firstName,
            emailLinkExpiryInMinutes,
            prepareInvLogoLink(),
            prepareEbrLogoLink()
          )
        );
      }

      blacklist.revoke(sessionUser);
      var iat = moment().unix();
      let jsonUser = {
        id: user.id,
        jti: uuid.v4(),
        iat,
        exp: sessionUser.exp,
        role: req.user.role,
        roleId: reqUser.roles,
        tempId: sessionUser.tempId
      };

      insertIntoAdminLogs(
        getUpdateUserMessage(`${user.firstName} ${user.lastName}`),
        req
      );

      return res.json({
        success: true,
        message: 'Update Successfully',
        newToken: jwt.sign(JSON.stringify(jsonUser), process.env.JWT_SECRET_KEY)
      });
    }

    const hash = await argon2.hash(reqUser.password);

    if (!hash) {
      throw new httpErrors.BadRequest('Password hash failed.');
    }

    let oldpassword = user.password;
    reqUser.password = hash;

    if (oldpassword !== reqUser.password) {
      var time = moment()
        .utc()
        .local()
        .format('YYYY-MM-DD HH:mm:ss');
      reqUser.reset_password_time = time;
    }

    reqUser.roles = sessionUser.roleId;

    await user.update(reqUser);

    if (emailChanged) {
      helper.sendVerificationEmail(
        prepareEmailOptions(
          user.email,
          language.get['USER_EMAIL_VERIFICATION'],
          emailTemplate,
          prepareOnPremiseVerificationLink(key),
          prepareOffPremiseVerificationLink(key),
          user.firstName,
          emailLinkExpiryInMinutes,
          prepareInvLogoLink(),
          prepareEbrLogoLink()
        )
      );
    }

    blacklist.revoke(sessionUser);
    var iat = moment().unix();
    let jsonUser = {
      id: user.id,
      jti: uuid.v4(),
      iat,
      exp: sessionUser.exp,
      role: req.user.role,
      roleId: reqUser.roles,
      tempId: sessionUser.tempId
    };

    insertIntoAdminLogs(
      getUpdateUserMessage(`${user.firstName} ${user.lastName}`),
      req
    );

    res.json({
      success: true,
      message: 'Update Successfully',
      newToken: jwt.sign(JSON.stringify(jsonUser), process.env.JWT_SECRET_KEY)
    });
  } catch (error) {
    next(error);
  }
};

/* Update User Theme */
const updateUserTheme = async (req, res, next) => {
  try {
    const transaction = await models.sequelize.transaction();
    console.log("Hello")
    const sessionUser = req.user;
    console.log("firstName: " + req.body.isDark)

    const user = await models.User.findOne({
      where: {
        id: sessionUser.id
      }
    });

    if (!user) {
      throw new httpErrors.NotFound('No user found to update.');
    }

    await user.update({
      is_dark: req.body.isDark
    }, {
      transaction
    });
    await transaction.commit();

    return res.json({
      success: true,
      message: 'Update Successfully'
    });
  } catch (error) {
    console.log(error)
    await transaction.rollback();
    next(error);
  }
};

/* update user email */
const updateEmail = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();
  try {
    const {
      useremail
    } = req.body;

    const dbUser = await models.User.findOne({
      where: {
        email: useremail
      }
    });

    if (dbUser) {
      const message = language.get['USER_EXIST_WITH_EMAIL'];
      return res.json({
        success: false,
        message: message
      });
    }

    const user = await models.User.findOne({
      where: {
        id: req.params.id
      }
    });

    if (!user) {
      throw new httpErrors.NotFound('No user to update.');
    }

    const [setting] = await Promise.all([
      models.Settings.findOne({
        where: {
          id: 1
        }
      }),
      user.update({
        email: useremail
      }, {
        transaction
      })
    ]);

    await setting.update({
      isFirstLogin: false
    }, {
      transaction
    });

    await transaction.commit();

    res.json({
      success: true,
      message: 'Email Changed Successfully.'
    });
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/* Lock single user */
const lockUser = async (req, res, next) => {
  try {
    const user = await models.User.findOne({
      where: {
        id: req.body.id
      }
    });

    if (!user) {
      throw new httpErrors.NotFound('No user found to lock.');
    }

    await user.update({
      is_locked: true
    });

    res.json({
      success: true,
      message: language.get['LOCKED_SUCCESS']
    });
  } catch (error) {
    next(error);
  }
};

/* Unlock single user */
const unlockUser = async (req, res, next) => {
  try {
    const user = await models.User.findOne({
      where: {
        id: req.body.id
      }
    });

    if (!user) {
      throw new httpErrors.NotFound('No user found to unlock.');
    }

    await user.update({
      is_locked: false,
      is_locked_auto: false,
      questions_verified: false,
      recaptcha_verified: false,
      wrong_login_attempts: 0
    });

    res.json({
      success: true,
      message: language.get['LOCKED_SUCCESS']
    });
  } catch (error) {
    next(error);
  }
};

/* Get a single user detail */
const getUser = async (req, res, next) => {
  try {
    const data = await models.User.findOne({
      // include: [{ model: models.Role }],
      attributes: [
        'id',
        'firstName',
        'lastName',
        'email',
        'question_1_id',
        'question_2_id',
        'question_3_id',
        'is_dark'
      ],
      where: {
        id: req.params.id
      }
    });

    console.log('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>', data.is_dark);
    res.json(data);
  } catch (error) {
    next(error);
  }
};

const updateSecurityQuestions = async (req, res, next) => {
  try {
    const {
      user
    } = req.body;
    const data = await models.User.findOne({
      attributes: [
        'question_1_id',
        'question_2_id',
        'question_3_id',
        'answer_1',
        'answer_2',
        'answer_3'
      ],
      where: {
        id: req.params.id
      }
    });

    if (
      user.question_1_id == user.question_2_id ||
      user.question_1_id == user.question_3_id ||
      user.question_2_id == user.question_3_id
    ) {
      throw new httpErrors.Conflict('Please select different questions');
    }

    if (user.answer_1 == '' || user.answer_2 == '' || user.answer_3 == '') {
      throw new httpErrors.Conflict('You can not have empty answer');
    }

    if (
      data.dataValues.question_1_id == user.question_1_id &&
      data.dataValues.question_2_id == user.question_2_id &&
      data.dataValues.question_3_id == user.question_3_id
    ) {
      const answer1Validate = await argon2.verify(
        data.dataValues.answer_1,
        user.answer_1
      );
      const answer2Validate = await argon2.verify(
        data.dataValues.answer_2,
        user.answer_2
      );
      const answer3Validate = await argon2.verify(
        data.dataValues.answer_3,
        user.answer_3
      );
      const validation = answer1Validate && answer2Validate && answer3Validate;
      if (validation) {
        res.json({
          state: '0'
        });
      } else {
        const ans1 = await argon2.hash(user.answer_1);
        const ans2 = await argon2.hash(user.answer_2);
        const ans3 = await argon2.hash(user.answer_3);
        const updateSecurity = await models.User.update({
          answer_1: ans1,
          answer_2: ans2,
          answer_3: ans3,
          questions_verified: true,
          is_question_changed: true
        }, {
          where: {
            id: req.params.id
          }
        });

        if (updateSecurity) {
          res.json({
            state: '1'
          });
        }
      }
    } else {
      const ans1 = await argon2.hash(user.answer_1);
      const ans2 = await argon2.hash(user.answer_2);
      const ans3 = await argon2.hash(user.answer_3);
      const updateSecurity = await models.User.update({
        question_1_id: user.question_1_id,
        question_2_id: user.question_2_id,
        question_3_id: user.question_3_id,
        answer_1: ans1,
        answer_2: ans2,
        answer_3: ans3,
        questions_verified: true,
        is_question_changed: true
      }, {
        where: {
          id: req.params.id
        }
      });

      if (updateSecurity) {
        res.json({
          state: '2'
        });
      }
    }
  } catch (error) {
    next(error);
  }
};

const resendVerification = async (req, res, next) => {
  try {
    const user = await models.User.findOne({
      where: {
        id: req.body.id
      }
    });

    if (!user) {
      throw new httpErrors.NotFound('No user found to send verification mail.');
    }

    const key =
      Math.random()
      .toString(36)
      .substring(2) +
      Math.random()
      .toString(36)
      .substring(2);

    const email_verification_link_expiry = moment().add(
      emailLinkExpiryInMinutes,
      'minutes'
    );

    await user.update({
      email_verification_link_expiry: email_verification_link_expiry,
      email_verification_link: key
    });

    helper.sendVerificationEmail(
      prepareEmailOptions(
        user.email,
        language.get['USER_EMAIL_VERIFICATION'],
        emailTemplate,
        prepareOnPremiseVerificationLink(key),
        prepareOffPremiseVerificationLink(key),
        user.firstName,
        emailLinkExpiryInMinutes,
        prepareInvLogoLink(),
        prepareEbrLogoLink()
      )
    );

    res.json({
      success: true,
      message: language.get['LOCKED_SUCCESS']
    });
  } catch (error) {
    next(error);
  }
};

/* add new user */
const createUser = async (req, res, next) => {
  const transaction = await models.sequelize.transaction();

  try {
    const {
      user
    } = req.body;

    const dbUser = await models.User.findOne({
      where: {
        email: user.email
      }
    });

    if (dbUser) {
      dbUser.is_dark = req.body.is_dark;
      dbUser.save();

      const message = language.get['USER_EXIST_WITH_EMAIL'];
      return res.json({
        success: false,
        message: message
      });
    }

    if (!user.tempId) {
      throw new httpErrors.BadRequest('Template id is required');
    }

    let tempId = user.tempId; // extracting template Id from user object
    delete user.tempId;

    const key =
      Math.random()
      .toString(36)
      .substring(2) +
      Math.random()
      .toString(36)
      .substring(2);
    user.email_verification_link_expiry = moment().add(
      emailLinkExpiryInMinutes,
      'minutes'
    );
    user.email_verification_link = key;
    const hash = await argon2.hash(user.password);

    if (!hash) {
      throw new httpErrors.BadRequest('Password hashing failed.');
    }

    user.password = hash;

    const time = moment()
      .utc()
      .local()
      .format('YYYY-MM-DD HH:mm:ss');
    user.reset_password_time = time;

    const data = await models.User.create(user);

    let query = `
      SELECT DISTINCT t.id as temp_id, t.name, t.description, tp.id as temp_permission_id, p.id as permission_id
      FROM Templates t 
      INNER JOIN TemplatePermissions tp ON tp.temp_id = t.id 
      INNER JOIN Permissions p ON p.id = tp.permission_id 
      WHERE t.id  = ${tempId}
    `;

    let results = (await models.sequelize.query(query))[0];

    if (results.length !== 0) {
      // Assign template to user:
      for (let i = 0; i < results.length; i++) {
        await models.UserTemplates.create({
          temp_permission_id: results[i].temp_permission_id,
          user_id: data.id
        }, {
          transaction: transaction
        });
      }

      if (tempId === 2) {
        let query2 = `
          SELECT id, name
          FROM Templates
          WHERE id BETWEEN 2 AND 6
          ORDER BY id
        `;

        let templates = (await models.sequelize.query(query2))[0];

        if (templates.length !== 0) {
          for (let i = 0; i < templates.length; i++) {
            await models.UserTemplatePivot.create({
              temp_id: templates[i].id,
              user_id: data.id
            }, {
              transaction: transaction
            });
          }
        }
      } else {
        await models.UserTemplatePivot.create({
          temp_id: results[0].temp_id,
          user_id: data.id
        }, {
          transaction: transaction
        });
      }
    }

    if (req.body.user.roles) {
      data.setRoles(parseInt(req.body.user.roles));
    }

    helper.sendVerificationEmail(
      prepareEmailOptions(
        user.email,
        language.get['USER_EMAIL_VERIFICATION'],
        emailTemplate,
        prepareOnPremiseVerificationLink(key),
        prepareOffPremiseVerificationLink(key),
        user.firstName,
        emailLinkExpiryInMinutes,
        prepareInvLogoLink(),
        prepareEbrLogoLink()
      )
    );

    insertIntoAdminLogs(
      getUserCreateMessage(`${user.firstName} ${user.lastName}`),
      req
    );

    await transaction.commit();
    res.json({
      success: true,
      message: 'Add Successfully'
    });
  } catch (error) {
    await transaction.rollback();
    next(error);
  }
};

/* check email validity */
const duplicationCheck = async (req, res, next) => {
  try {
    const {
      value
    } = req.body;

    const user = await models.User.findOne({
      where: {
        email: value
      }
    });

    if (user) {
      return res.send(400);
    }

    res.send(200);
  } catch (error) {
    next(error);
  }
};

/* delete a single user */
const deleteUser = async (req, res, next) => {
  try {
    const user = await models.User.findOne({
      where: {
        id: req.params.id
      }
    });

    if (!user) {
      throw new httpErrors.NotFound('No user found to delete.');
    }

    let query = `
      SELECT DISTINCT u.id, u.createdAt 
      FROM Users u
      INNER JOIN UserTemplates ut ON ut.user_id = u.id 
      INNER JOIN TemplatePermissions tp ON tp.id  = ut.temp_permission_id 
      INNER JOIN Templates t ON t.id = tp.temp_id 
      WHERE t.id = 1 OR t.id = 2
      ORDER BY u.createdAt ASC
      LIMIT 2
    `;

    let results = (await models.sequelize.query(query))[0];

    let extractedIds = results.map(result => result.id); // id of the first Infra Admin and first Admin

    let incomingId = parseInt(req.params.id);

    // If the incoming id is of first Infra Admin or Admin, don't allow to delete
    if (extractedIds.includes(incomingId)) {
      throw new httpErrors.NotFound(
        'Cannot delete this admin, user does not have the required permission'
      );
    }

    await models.AdminAccessLogs.destroy({
      where: {
        userId: req.params.id
      }
    });
    await models.UserRole.destroy({
      where: {
        userId: req.params.id
      }
    });
    await models.UserTimeSetting.destroy({
      where: {
        user_id: req.params.id
      }
    });
    await models.UserTemplatePivot.destroy({
      where: {
        user_id: req.params.id
      }
    });
    await models.UserTemplates.destroy({
      where: {
        user_id: req.params.id
      }
    });
    // not deleting, just updating as the same template could be used by other users
    await models.Templates.update({
      created_by: null
    }, {
      where: {
        created_by: req.params.id
      }
    });

    await models.Service.update({
      createdById: null
    }, {
      where: {
        createdById: req.params['id']
      }
    });

    await user.destroy({});

    insertIntoAdminLogs(
      getUserDeletMessage(`${user.firstName} ${user.lastName}`),
      req
    );

    res.json({
      success: true,
      message: 'Delete Successfully'
    });
  } catch (error) {
    next(error);
  }
};

const regenerateUserKeys = async (req, res, next) => {
  try {
    const reqUser = helper.verifyJwtToken(req.header('authorization'));

    const {
      id
    } = req.params;

    if (reqUser.id != id) {
      blacklist.revoke(reqUser);
      throw new httpErrors.Forbidden('Operation not allowed.');
    }

    const user = await models.User.findById(id);
    if (!user) {
      throw new httpErrors.NotFound('No user found.');
    }

    const {
      publicKey,
      privateKey
    } = await generateKeyValuePair();
    (user.public_key = publicKey), (user.secret_key = privateKey);

    await user.save();

    res.json({
      success: true,
      publicKey,
      secretKey: privateKey,
      message: 'Keys updated'
    });
  } catch (e) {
    next(e);
  }
};

/* ################################# Utility Functions ######################################## */
const prepareOnPremiseVerificationLink = key => {
  return (
    `${projectConfig.onPremUiUrl.split('web')[0]}api/verify/user/` +
    new Buffer(key).toString('base64')
  );
};

const prepareOffPremiseVerificationLink = key => {
  return (
    `${projectConfig.offPremUiUrl.split('web')[0]}api/verify/user/` +
    new Buffer(key).toString('base64')
  );
};

const prepareInvLogoLink = () => {
  let fullWebUrl = projectConfig.offPremApiUrl;
  let mainUrl = fullWebUrl.split('/api')[0];
  return `${mainUrl}/assets/images/invisily-logo.png`;
};

const prepareEbrLogoLink = () => {
  let fullWebUrl = projectConfig.offPremApiUrl;
  let mainUrl = fullWebUrl.split('/api')[0];
  return `${mainUrl}/assets/images/ebryx-logo.png`;
};

function isPublicUser(user) {
  return (
    user && user.dataValues.template && user.dataValues.template === 'API User'
  );
}

module.exports = {
  getUsers,
  resetMfa,
  deleteUser,
  duplicationCheck,
  getUser,
  updateUser,
  updateEmail,
  unlockUser,
  lockUser,
  resendVerification,
  createUser,
  regenerateUserKeys,
  updateSecurityQuestions,
  prepareOnPremiseVerificationLink,
  prepareOffPremiseVerificationLink,
  prepareInvLogoLink,
  prepareEbrLogoLink,
  prepareEmailOptions,
  updateUserTheme
};