const router = require('express').Router();
const httpErrors = require('http-errors');
const models = require('../models');

const { 
    getDomainCreateMessage,
    getDomainUpdateMessage,
    getDomainDeleteMessage
  } = require('../config/adminLogs.config');
const { insertIntoAdminLogs } = require('../routes/common/adminLogs');


const getDomains = async (req, res, next) => {
    try {
        const {
            limit,
            offset,
            sort = 'id',
            dir = 'desc',
            searchKey = ''
        } = req.query;

        const options = {
            attributes: ['id', 'domain'],
            limit: +limit,
            offset: +offset,
            order: []
        }

        const allowedSortValues = [
            'id',
            'domain'
        ];

        if (searchKey.trim()) {
            options.where = {
                domain: {
                    [Op.like]: `${searchKey.trim()}%`
                }
            }
        }

        if (allowedSortValues.includes(sort)) {
            options.order.push([models.sequelize.literal(sort), dir]);
        }

        const result = await models.ValidDomain.paginate(options);
        res.json(result);

    } catch (error) {
        next(error);
    }
}


const addDomain = async (req, res, next) => {
    const reqBody = req.body;
    try {
        const isDomainExist = await models.ValidDomain.findOne({
            where: {
                domain: reqBody.domain
            },
            attributes: [
                'id',
                'domain'
            ]
        });

        if (isDomainExist) {
            throw new httpErrors.Conflict('This domain already exist.');

        }

        const domain = await models.ValidDomain.create(reqBody);

        insertIntoAdminLogs(
            getDomainCreateMessage(reqBody.domain),
            req
        )

        res.json(domain);
    } catch (error) {
        next(error);
    }
}


const updateDomain = async (req, res, next) => {
    const reqBody = req.body;
    try {

        const previousDomain = await models.ValidDomain.findOne({
            where: {
                id: req.params.id
            }
        })

        const isDomainExist = await models.ValidDomain.findOne({
            where: {
                domain: reqBody.domain,
                id: {
                    $ne: req.params.id
                }
            }
        });

        if (isDomainExist) {
            throw new httpErrors.Conflict('This domain already exist.');
        }

        const domain = await models.ValidDomain.update(reqBody, {
            where: {
                id: req.params.id
            }
        });

        insertIntoAdminLogs(
            getDomainUpdateMessage(previousDomain.dataValues.domain, reqBody.domain),
            req
        )

        res.send(domain);
    } catch (error) {
        next(error);
    }
}


const deleteDomain = async (req, res, next) => {
    try {
        const existedDomain = await models.ValidDomain.findOne({
            where: {
                id: req.params.id
            }
        });

        if (!existedDomain) {
            throw new httpErrors.NotFound('No domain found against this Id.');
        }

        const associatedUsers = await models.LocalUsers.findOne({
            where: {
                email: {
                    [Op.like]: `%@${existedDomain.domain}`
                }
            }
        });

        if(associatedUsers){
            throw new httpErrors.Conflict('Cannot delete domain which is associated with user(s).');
        }

        const domain = await models.ValidDomain.destroy({
            where: {
                id: req.params.id
            }
        });

        insertIntoAdminLogs(
            getDomainDeleteMessage(existedDomain.domain),
            req
        )

        res.json(domain);

    } catch (error) {
        next(error);
    }
}


module.exports = {
    getDomains,
    addDomain,
    updateDomain,
    deleteDomain
};