const crypto = require('crypto');

const cache = require('../config/cache');

class Queue {
  /**
 * RabbitMQ Abstract Queue class
 * @param {string} queueName Name of the queue 
 * @param {Object} opts Options available to configure
 * @param {Object} opts.cache Redis Object
 * @param {Number} opts.prefetch How many parallel message can be processed. Default value is 1.
 * @param {Number} opts.retryDelay Delay between each retry in milliseconds. Default value is 5000.
 * @param {string} opts.delayedExchangeName Name of delayed exchange. Default value is queue-delayed-exchange.
 * @param {string} opts.offlineQueueName Name of the offline queue. Default name is ${queueName}-offline.
 * @param {Number} opts.offlineQueueTTL Maximum time (seconds) message remains in offline queue. Default value is 120.
 * @param {Number} opts.consumerInstance Total number of consumers need to run. Default value is 0 (0 Will start maximum processes depending on available CPUs)
 */
  constructor(queueName = 'queue', opts = {}) {

    if (this.constructor === Queue) {
      throw new Error("Can't instantiate abstract class!");
    }

    const _defualts = {
      queueName,
      connection: null,
      channel: null,
      cache,
      prefetch: 1,
      retryDelay: 5000,
      maxRetries: 5,
      delayedExchangeName: null,
      offlineQueueName: 'offline',
      offlineQueueTTL: 120,
      consumerInstance: 0,
      delayedExchangeName: 'queue-delayed-exchange'
    };

    this.config = {..._defualts, ...opts};

    this.status = {
      PROCESSING: 'processing',
      COMPLETED: 'completed',
      FAILED: 'failed',
      PENDING: 'pending'
    }
  }

  init(connection) {
    this.config.connection = connection;

    // Creating queue and start consumers
    this.createQueue().then(this._consumer.bind(this));
  }

  async createQueue() {

    try {
      this.config.channel = await this.createChannel();

      this.config.channel.on('close', () => {
        this.config.channel = null;
      });

      await this.config.channel.assertExchange(
        this.config.delayedExchangeName,
        'x-delayed-message',
        {
          arguments:{"x-delayed-type":"direct"},
          durable: false,
          noAck: false
        }
      )

      await this.config.channel.assertQueue(this.config.queueName, { 
        durable: true,
        arguments: {
          'x-message-deduplication': true,
        }
      }),

      await this.config.channel.bindQueue(
        this.config.queueName,
        this.config.delayedExchangeName,
        this.config.queueName
      );
      
      if(this.config.cache && this.config.offlineQueueTTL) {
        this.config.cache.smembers(
          this.setCacheName(this.config.offlineQueueName),
          (err, list) => {
            while(list && list.length) {
              list.shift();
              this.config.cache.spop(this.setCacheName(this.config.offlineQueueName), (err, msg) => {
                if(!err) 
                  this.addtoQueue(JSON.parse(msg));
              })
            }
          }
        )
      }

    } catch (error) {
      console.log('RabbitMQ channel error: ', error.message);
    }

  }

  createChannel() {
    if(!this.config.connection) {
      throw new Error('connection value can\'t be null');
    }
    
    return this.config.connection.createChannel();
  }

  /**
   * Send message to queue
   * @param {Object} message Content need to send into rabbitmq
   * @param {Object} [headers] Headers that will be passed with message
   * @param {boolean} [isBuffer] Is content is already buffer
   */
  addtoQueue(message, headers = {}, isBuffer = false) {
    if(!this.config.channel && !isBuffer) {

      if(this.config.cache && this.config.offlineQueueTTL) {
        this.config.cache.sadd(
          this.setCacheName(this.config.offlineQueueName),
          JSON.stringify(message),
          _ => {}
        );
  
        this.config.cache.expire(
          this.setCacheName(this.config.offlineQueueName),
          this.config.offlineQueueTTL,
          _ => {}
        );
      }

      return;
    }

    const msg = isBuffer ? message : JSON.stringify(message);

    const hash = headers['x-deduplication-header']
      ? headers['x-deduplication-header'] 
      : this.createHash(msg);

    console.log(`Adding to ${this.config.queueName} queue: `, msg);
    
    this.config.cache.get(this.setCacheName(hash), (err, value) => {
      
      if ( value && !headers['x-retries']) {
        if(value === this.status.COMPLETED)
          this.config.cache.del(this.setCacheName(hash), _ => {});
        return;
      }
  
      this.config.channel.publish(
        this.config.delayedExchangeName,
        this.config.queueName,
        isBuffer ? msg : Buffer.from(msg),
        {
          persistent: true,
          headers: {
            'x-deduplication-header': hash,
            ...headers
          }
        }
      );
      console.log(`Published to ${this.config.queueName} queue: `, msg);
      this.config.cache.set(this.setCacheName(hash), this.status.PENDING, _ => {}); 
    });
  }

  async _consumer() {
    
    if (this.isMaxConsumerInstanceReached()) {
      return;
    }

    let chConsumer = null;
    try {
      chConsumer = await this.createChannel();
    } catch (error) {
      console.error('Consumer channel error: ', error.message);
      return;
    }

    chConsumer.prefetch(this.config.prefetch);

    chConsumer.consume(
      this.config.queueName,
      async msg => {
        const hash = msg.properties.headers['x-deduplication-header'];
        let value = null;
        try {
          value = await this.config.cache.getAsync(this.setCacheName(hash));  
        } catch {}
  
        if(value === this.status.COMPLETED) {
          chConsumer.ack(msg);
          this.config.cache.del(this.setCacheName(hash), _ => {});
          return;
        }
        this.config.cache.set(this.setCacheName(hash), this.status.PROCESSING, _ => {});

        try {
          const data = JSON.parse(msg.content.toString());
          await this.consumer(data);
          this.config.cache.set(this.setCacheName(hash), this.status.COMPLETED, _ => {});

          chConsumer.ack(msg);

          this.config.cache.del(this.setCacheName(hash), _ => {});
          
        } catch (error) {
          console.error('Consumer error: ', error.message);
          const headers = msg.properties.headers;
          let xRetries = headers['x-retries'] || 0;
          xRetries += 1;
          let value = null;
          try {
            value = await this.config.cache.getAsync(this.setCacheName(hash));  
          } catch {}

          if(value !== this.status.COMPLETED) {
            this.config.cache.set(this.setCacheName(hash), this.status.FAILED, _ => {});
          } else {
            return;
          }
          console.log("Retry no: ", xRetries, "Delay : ", this.config.retryDelay);
          
          if(xRetries < this.config.maxRetries) {
            this.addtoQueue(msg.content, this.setHeaders(headers, xRetries), true);
          } else{
            console.log("Max retry limit reached");
            this.config.cache.del(this.setCacheName(hash), _ => {});
          }
          
          chConsumer.ack(msg);
        }
      },
      {
        noAck: false
      }
    )
  }

  /**
   * Consumer Method
   * @param {object} msg 
   */
  async consumer(msg) {
    throw new Error("Method 'consumer(msg)' must be implemented.");
  }

  isMaxConsumerInstanceReached() {

    if(this.config.consumerInstance === -1) {
      return true;
    }

    if (
      this.config.consumerInstance === 0 ||
      isNaN(process.env.NODE_APP_INSTANCE) ||
      Number(process.env.NODE_APP_INSTANCE) < this.config.consumerInstance
    ) {
      return false;
    }

    return true;

  }

  setCacheName(name) {
    return `${this.config.queueName}:${name}`
  }

  createHash(msg) {
    return crypto.createHash('md5').update(msg).digest('hex');
  }

  /**
   * Delay for next iteration
   * @param {number} retryNo 
   */
  retryStrategy(retryNo = 1) {
    return retryNo * this.config.retryDelay;
  }

  setHeaders(headers, retryNo) {
    return {
      'x-deduplication-header': headers['x-deduplication-header'],
      'x-delay': this.retryStrategy(retryNo),
      'x-retries': retryNo
    }
  }
}

module.exports = Queue;