const express = require('express');
const router = express.Router();
const httpErrors = require('http-errors');

const models = require('../../models');
const PagedControl = require('ldapjs/lib/controls/paged_results_control');

const LdapManager = require('../common/ldap-manager');

class LDAPConfigMissingError extends Error {
  constructor(status, allowedStatuses) {
    super('LDAP Configuration missing in the database.');
    this.name = 'LDAPConfigMissingError';
    this.state = status;
    this.allowedStates = allowedStatuses;
  }
}

router.post('/test', async (req, res, next) => {
  try {
    const { ldap } = req.body;
    const ldapManager = new LdapManager();
    const response = await ldapManager.createLdapConnection(ldap);
    res.send(response);
  } catch (error) {
    next(error);
  }
});

const ldapSearchManager = new LdapManager();

router.post('/:term?', (req, response, next) => {
  const term = new Buffer(req.params.term, 'base64').toString('ascii');
  console.log('---ldap search', term);
  //const { term } = req.params;
  const _nextPage = req.body.nextPage;
  const _cookie = _nextPage && _nextPage.cookie;

  const connectionId = _nextPage && _nextPage.connectionId;
  const pageNumber = (_nextPage && _nextPage.pageNumber) || 0;
  const requestCookie = (_cookie && Buffer.from(_cookie, 'base64')) || null;

  const searchTermFilter = `(|(cn=*${term}*)(mail=*${term}*)(uid=*${term}*)(userPrincipalName=*${term}*))`;
  const presenceFilter = '(cn=*)(|(&(mail=*)(uid=*))(userPrincipalName=*))';
  const searchFilter = `(&(objectClass=person)${presenceFilter}${searchTermFilter})`;

  let options = {
    scope: 'sub',
    filter: searchFilter,
    attributes: ['cn', 'mail', 'uid', 'userPrincipalName']
  };

  const controls = [];

  const searchCallback = (err, res) => {
    if (err) {
      console.log(err && (err.message || err));
      //throw new LDAPConfigMissingError();
      response.status(500).json({ status: 'OperationFailed' });
      return;
    }

    const entries = [];

    res.on(
      'searchEntry',
      ({ object: { userPrincipalName, uid, mail, cn } }) => {
        entries.push({
          uid: userPrincipalName ? userPrincipalName.split('@')[0] : uid,
          mail: userPrincipalName || mail,
          cn
        });
      }
    );

    res.on('error', err => {
      console.log(err && (err.message || err));
      response.status(500).json({ status: 'OperationFailed' });
    });

    res.on('end', result => {
      const pagedResponse = result.controls.filter(
        control => control instanceof PagedControl
      )[0];

      response.json({
        entries,
        nextPage: {
          cookie: pagedResponse
            ? pagedResponse.value.cookie.toString('base64')
            : null,
          connectionId: ldapSearchManager.connectionId,
          pageNumber: pageNumber + 1
        }
      });
    });
  };

  models.Ldap.findOne({ where: { } })
    .then(ldap => {
      if (ldap) {
        console.log('FOUND LDAP SETTINGS');
        return ldapSearchManager.reuseOrCreateLdapConnection(ldap).catch(() => {
          throw new LDAPConfigMissingError();
        });
      } else {
        throw new LDAPConfigMissingError();
      }
    })
    .then(() => {
      // If we lost connection while the user was searching, we need to skip
      // entries that had already been transferred, we create a PagedControl
      // that searches for pageNumber * pageSize entries and do nothing.
      if (connectionId && connectionId !== ldapSearchManager.connectionId) {
        return ldapSearchManager
          .do(
            'skipTo',
            options,
            pageNumber,
            ldapSearchManager.PREFFERED_PAGE_SIZE
          )
          .then(skipCookie => {
            controls.push(
              new PagedControl({
                value: {
                  size: ldapSearchManager.PREFFERED_PAGE_SIZE,
                  cookie: skipCookie
                }
              })
            );

            ldapSearchManager.do(
              'search',
              ldapSearchManager.settings.LDAP_dc,
              options,
              controls,
              searchCallback
            );
          });
      } else {
        controls.push(
          new PagedControl({
            value: {
              size: ldapSearchManager.PREFFERED_PAGE_SIZE,
              cookie: requestCookie
            }
          })
        );

        ldapSearchManager.do(
          'search',
          ldapSearchManager.settings.LDAP_dc,
          options,
          controls,
          searchCallback
        );
      }
    })
    .catch(err => {
      if (err instanceof LDAPConfigMissingError) {
        response.json({
          entries: []
        });
      } else {
        console.log(err && (err.message || err));
        response.status(500).json({ status: 'OperationFailed' });
      }
    });
});

module.exports = router;
