const express = require('express');
const router = express.Router();
const _ = require('lodash');
const { convertToArray } = require('../../utils/common')

const validate = require('../../validators/validate');
const serviceValidator = require('../../validators/services');
const authenticateMiddleware = require('../../middlewares/publicServiceAuthenticate');

const models = require('../../models');
const httpErrors = require('http-errors');

//public api to add services
router.post('/create',authenticateMiddleware,async (req,res,next) => {

    const transaction = await models.sequelize.transaction();

    try{
        
        const { gateway_name,Name,service_group_name } = req.body;

        if(await models.Service.count({ where: { Name }})){
            throw new httpErrors.BadRequest('Service with this name already exists');
        }

        const serviceObject = _.omit({
            service_id: 0,
            ...req.body
        },['signature','timestamp','api_key']);

        const createdService = await models.Service.create(serviceObject, { transaction });
        
        createdService.service_id = createdService.id;

        await createdService.save({transaction});

        if(gateway_name){

            const gateway = await models.Gateway.findOne({
                where:{
                    Name:gateway_name
                }
            });
    
            if(!gateway) throw new httpErrors.NotFound('Gateway with this name does not exist');
    
            await models.GatewayService.create({
                gatewayId:gateway.id,
                serviceId:createdService.id
            } , { transaction });

        }

        if(service_group_name){

            const toFindServiceGroup = convertToArray(service_group_name);
            const serviceGroup = await models.ServiceGroup.findAll({ 
                where:{ 
                    name:{
                        [Op.in]:toFindServiceGroup
                    }
                },
                attributes:['id']
            });

            if(!serviceGroup.length) throw new httpErrors.NotFound('Service group with this name(s) does not exist');

            await models.GroupService.bulkCreate(
                serviceGroup.map(service => ( {
                    groupId:service.id,
                    serviceId:createdService.id
                })),
                { transaction }
            );
        }
        
        await transaction.commit();

        res.json({
            success: true,
            message: 'Service added Successfully.'
        });

    }catch(error){
        await transaction.rollback();
        next(error);
    }
});

//public api to delete servcices
router.delete('/delete',authenticateMiddleware, async(req,res,next) => {

    const transaction = await models.sequelize.transaction();

    try{

        const { Name } = req.body;

        const service = await models.Service.findOne({
            where: {
              Name
            },
            attributes:['id']
        });
      
        if(!service){
            throw new httpErrors.NotFound('Service not found.');
        }

        await Promise.all([
            service.destroy({ where:{ } }, { transaction }),
            models.GroupService.destroy({
                where:{
                    serviceId:service.id
                }
            }, { transaction }),
            models.GatewayService.destroy({
                where: { 
                  serviceId: service.id 
                },
            }, { transaction })
        ]);

        await transaction.commit();
      
        res.status(200).json({
            success: true,
            message: 'Deleted Successfully'
        });

    }catch(error){
        await transaction.rollback();
        next(error);
    }
})

module.exports = router;