#!/bin/bash

SCRIPT_NAME="$0"
TIMESTAMP=$(date +"%Y-%m-%d %H:%M:%S.%3N")

set -eE -o functrace

function failure() {
  local LINE=$1
  local CMD=$2
  echo >& "[FTL] FATAL ERROR! [$SCRIPT_NAME:$LINE] | $CMD"
}

trap 'failure "$LINENO" "$BASH_COMMAND"' ERR

function log() {
  echo -e "$TIMESTAMP" "$@"
}

# --- script ---

DEB_PKG_NAME='invisily_client.deb'
INSTALLED_PKG_NAME='invisily-client'
SNIFFER_CONFIG_FILE_PATH='/opt/invisily/linuxclientd/client_data/configs/sniffer_config.cfg'

log "[INF] Running $SCRIPT_NAME"

log "[INF] Configuring and installing S2S client"

log "[INF] Validating root user"
if (( EUID != 0 )); then
  log "[ERR] Run with root user!"
  log "[INF] Usage: sudo $SCRIPT_NAME"
  exit 1
fi
log "[INF] Validated root user successfully!"

log "[INF] Validating debian package [$DEB_PKG_NAME]"
if [[ ! -f ./$DEB_PKG_NAME ]]; then
  log "[ERR] Package not found! [$DEB_PKG_NAME]"
  exit 1
fi
log "[INF] Validated debian package! [$DEB_PKG_NAME]"

log "[INF] Validating installed package [$INSTALLED_PKG_NAME]"
if dpkg -s $INSTALLED_PKG_NAME > /dev/null 2>&1; then
  log "[ERR] Package is already installed! [$INSTALLED_PKG_NAME]"
  log "[INF] Please uninstall it and then try again!"
  exit 1
fi
log "[INF] Package is not installed! [$INSTALLED_PKG_NAME]"

log "[INF] Installing package [$DEB_PKG_NAME]"
chmod a+x ./$DEB_PKG_NAME
if ! apt install ./$DEB_PKG_NAME -y; then
  log "[ERR] Failed to install package! [$DEB_PKG_NAME]"
  exit 1
fi
log "[INF] Package installed successfully! [$DEB_PKG_NAME]"

log "[INF] Generating sniffer configuration [$SNIFFER_CONFIG_FILE_PATH]"

echo "Select network interface:"
INTERFACES=$(ip -4 -br addr | tr -s '[:space:]' | grep -E '[[:print:]] UP' | cut -d ' ' -f1)
select INTERFACE in $INTERFACES; do
  if [[ -n $INTERFACE ]]; then
    echo "[INF] Interface selected! [$INTERFACE]"
    break
  else
    echo "[ERR] Invalid option!"
  fi
done

IP=$(ip -4 -br addr show "$INTERFACE" | tr -s '[:space:]' | cut -d ' ' -f3 | cut -d '/' -f1)

{
  echo "interface=\"$INTERFACE\""
  echo "ip=\"$IP\""
} > $SNIFFER_CONFIG_FILE_PATH

log "[INF] Dumping sniffer configuration [$SNIFFER_CONFIG_FILE_PATH]"
log "[INF] --- SNIFFER CONFIGURATION STARTS ---"
cat $SNIFFER_CONFIG_FILE_PATH
log "[INF] ---- SNIFFER CONFIGURATION ENDS ----"

log "[INF] Sniffer configuration generated successfully! [$SNIFFER_CONFIG_FILE_PATH]"

log "[INF] Configured and installed S2S client successfully!"

