import { Component, OnInit, OnDestroy, ViewChild } from '@angular/core';
import { AdminAccessLogService } from './admin-access-log.service';
import { GlobalMessageService } from '../global-messages/global-messages.service';
import { Pagination } from '../models/pagination';
import { Router } from '@angular/router';
import { AuthenticationService } from '../login/authentication.service';
import * as moment from 'moment';
import { Subject, Observable, of } from 'rxjs';
import { SettingsService } from '../settings/settings.service';
import { debounceTime, switchMap } from 'rxjs/operators';
import { ErrorCodeEnum } from '../shared/emuns/error.enum';
import { LicenseErrorEnum } from '../shared/emuns/license.enum';
import { VersionService } from '../version/version.service';
import { Title } from '@angular/platform-browser';

@Component({
  selector: 'app-admin-access-log',
  templateUrl: './admin-access-log.component.html',
  styleUrls: ['./admin-access-log.component.css']
})
export class AdminAccessLogComponent implements OnInit, OnDestroy {
  logs: any = {};
  allLogs: any = {};
  logsCount: number;
  pagination: Pagination = new Pagination();
  $searchLog = new Subject();
  searchText = '';
  start_date: Date = null;
  end_date: Date = null;
  sort: string;
  sortDir: string;
  isPrevious = false;
  private dateFormat: string = 'dd/MM/yyyy HH:mm';

  loading: Boolean = true;
  dateRangePicker: Date[] = [];

  today: Date;

  pageSize = [10, 25, 50, 100, 200];
  timeZone = '';
  colorsByPriority = {};
  priorites = [];
  selectedPriority = null;
  portalVersionInfo: string;

  constructor(
    private adminLogsService: AdminAccessLogService,
    private globalMessageService: GlobalMessageService,
    private router: Router,
    private authenticationService: AuthenticationService,
    private versionService: VersionService,
    private titleService: Title
  ) {
    this.titleService.setTitle('Access Logs | Admin');
    this.today = new Date();

    var localDateFormat = localStorage.getItem('dateFormat');
    if (typeof localDateFormat != 'undefined' || localDateFormat != null) {
      this.dateFormat = localDateFormat;
    }

    let timeZoneOffset = new Date().getTimezoneOffset();
    this.timeZone =
      'UTC' + (timeZoneOffset > 0 ? '-0' : '+0') + (timeZoneOffset / 60) * -1;
  }

  ngOnInit() {
    this.getLogs();
    this.subscribeToSearch();
    this.colorsByPriority = this.adminLogsService.getAllColors();
    this.priorites = Object.values(this.colorsByPriority);
    this.portalVersionInfo = this.versionService.portalVersion;
  }

  private getLogs() {
    this.loading = true;
    this.adminLogsService
      .searchLog(
        this.searchText,
        this.pagination,
        this.start_date,
        this.end_date,
        this.sort,
        this.sortDir,
        this.selectedPriority
      )
      .subscribe(
        (logs: any) => {
          this.loading = false;
          this.logs = this.allLogs = logs;
          this.pagination.totalItems = this.logsCount = logs.recordsTotal;
        },
        err => {
          this.loading = false;
          if (err.status === ErrorCodeEnum.Forbidden) {
            if (err._body == LicenseErrorEnum.Invalid) {
              this.authenticationService.showLicenseOnly(true);
              this.router.navigate(['/license']);
            } else {
              this.showAlert(
                'error',
                'You are not authorized to access this page.',
                false
              );
            }
          } else {
            // this.showAlert('danger', err._body, false);
          }
        }
      );
  }

  clearfun() {
    this.searchText = '';
    this.getLogs();
  }

  private showAlert(type: string, message: string, sticky: boolean) {
    // const alert = {
    //   closed: false,
    //   type: type,
    //   message: message,
    //   sticky: sticky
    // };
    // this.globalMessageService.pushAlert(alert);
  }

  subscribeToSearch() {
    this.$searchLog
      .pipe(
        debounceTime(500),
        switchMap(
          (search: {
            searchWord: string;
            typing: boolean;
            date_change: boolean;
            sort_change: boolean;
          }) => {
            this.loading = true;

            if (
              (search.searchWord && /\S/.test(search.searchWord)) ||
              this.start_date ||
              this.end_date ||
              search.sort_change
            ) {
              if (search.typing || search.date_change || search.sort_change) {
                this.pagination.currentPage = 1;
              }
              this.isPrevious = false;
              return this.adminLogsService.searchLog(
                search.searchWord,
                this.pagination,
                this.start_date,
                this.end_date,
                this.sort,
                this.sortDir,
                this.selectedPriority
              );
            } else {
              this.isPrevious = true;
              this.pagination.totalItems = this.logsCount;
              this.pagination.currentPage = 1;
              return of({
                data: this.allLogs.data,
                recordsTotal: this.logsCount
              });
            }
          }
        )
      )
      .subscribe((logs: any) => {
        this.loading = false;
        this.pagination.totalItems = logs.recordsTotal;
        this.logs = logs;
      });
  }

  pageChange(page) {
    this.pagination.currentPage = page;
    if (this.searchText === '' && !this.start_date && !this.end_date) {
      this.getLogs();
    } else {
      this.$searchLog.next({
        searchWord: this.searchText,
        typing: false,
        date_change: false,
        sort_change: false
      });
    }
  }

  ngOnDestroy() {
    this.$searchLog.unsubscribe();
  }

  dateChange() {
    this.$searchLog.next({
      searchWord: this.searchText,
      typing: false,
      date_change: true,
      sort_change: false
    });
  }

  dateRangePickerChange() {
    if (this.dateRangePicker[0]) {
      this.start_date = this.dateRangePicker[0];
    }

    if (this.dateRangePicker[1]) {
      this.end_date = this.dateRangePicker[1];
    }

    this.dateChange();
  }

  clearDate() {
    this.dateRangePicker = [];
    this.start_date = null;
    this.end_date = null;
    this.dateChange();
  }

  exportCSV() {
    this.adminLogsService.exportCSV(
      this.searchText,
      this.start_date,
      this.end_date,
      this.selectedPriority,
      this.portalVersionInfo
    );
  }

  sortRecords(term: string) {
    this.sortDir = this.sortDir == 'asc' ? 'desc' : 'asc';
    this.sort = term;
    this.searchText == ''
      ? this.getLogs()
      : this.$searchLog.next({ searchWord: this.searchText });
  }

  pageSizeChange(size) {
    this.pagination.itemsPerPage = +size;

    //not a proper way to do, pagination dynamic change erorr fix
    setTimeout(() => {
      this.pagination.currentPage = 1;
      this.searchText == ''
        ? this.getLogs()
        : this.$searchLog.next({ searchWord: this.searchText });
    }, 0);
  }

  priorityChange(priority) {
    this.pagination.currentPage = 1;
    this.getLogs();
  }
}
