import { Injectable } from '@angular/core';
import { HttpClient as AuthHttp } from '@angular/common/http';
import { ConfigService } from '../services/config.service';
import { AuthShortlivedCookieService } from '../auth-shortlived-cookie.service';
import { Pagination } from '../models/pagination';
import * as moment from 'moment';
import { map } from 'rxjs/operators';
import { RoleEnum } from '../shared/emuns/global.enum';
import { Router } from '@angular/router';
import { AuthenticationService } from '../login/authentication.service';
import { BehaviorSubject } from 'rxjs';

@Injectable()
export class AlertServices {
  private alertApiUrl: string;
  showLicenseOnly: boolean = false;
  public alertLastId: number = null;
  public userrole: string = '';
  public alertCount: number = 0;
  public alerts: any = [];

  private notificationPurge = new BehaviorSubject<boolean>(false);
  private notificationPurge$ = this.notificationPurge.asObservable();

  constructor(
    private authHttp: AuthHttp,
    private configService: ConfigService,
    private authShortlivedCookie: AuthShortlivedCookieService,
    private router: Router,
    private authenticationManager: AuthenticationService
  ) {
    const baseUrl = configService.getConfiguration().apiBaseUrl;
    this.alertApiUrl = baseUrl + '/alerts';

    this.userrole = localStorage.getItem('ur');
    if (typeof this.userrole != 'undefined' && this.userrole != null) {
      this.userrole = atob(this.userrole);
    }

    this.authenticationManager.showLicenseOnlyEmitter.subscribe(mode => {
      this.showLicenseOnly = mode;
    });
  }

  getNotificationPurge() {
    return this.notificationPurge$;
  }
  setNotificationPurge(value: boolean) {
    return this.notificationPurge.next(value);
  }
  getDistinctType() {
    return this.authHttp
      .get(`${this.alertApiUrl}/distinctType`)
      .pipe(map(response => response));
  }
  getDistinctSeverity() {
    return this.authHttp
      .get(`${this.alertApiUrl}/distinctSeverity`)
      .pipe(map(response => response));
  }

  getAlerts(
    pageSize: number = 10,
    page: number = 1,
    message: string = '',
    startDate: Date,
    endDate: Date,
    alertType = '',
    alertSeverity = '',
    sort = 'createdAt',
    dir = 'desc'
  ) {
    if (this.showLicenseOnly) return;
    const urlEncoded = encodeURIComponent(message);
    let _startDate = '1900-01-01';
    let _endDate = '9999-01-01';
    if (startDate) {
      _startDate = moment(startDate)
        .startOf('day')
        .format('YYYY-MM-DD HH:mm:ss');
    }
    if (endDate) {
      _endDate = moment(endDate)
        .endOf('day')
        .format('YYYY-MM-DD HH:mm:ss');
    }

    let query = `?limit=${pageSize}&offset=${page -
      1}&type=${message}&message=${urlEncoded}&startDate=${_startDate}&endDate=${_endDate}&sort=${sort}&dir=${dir}`;
    if (alertType) {
      query = query + `&alertType=${alertType}`;
    }
    if (alertSeverity) {
      query = query + `&alertSeverity=${alertSeverity}`;
    }
    const now = moment.utc().subtract(1, 'minute');
    return this.authHttp.get(`${this.alertApiUrl + query}`).pipe(
      map(response => response),
      map((response: any) => {
        response.data = response.data.map(d => {
          return {
            ...d,
            isNew: moment.utc(d.createdAt) >= now
          };
        });
        return response;
      })
    );
  }

  getAlertsNotification(pageSize: number = 10, id: number) {
    let query = `?limit=${pageSize}&id=${id ? id : ''}`;
    return this.authHttp
      .get(`${this.alertApiUrl + '/notifications' + query}`)
      .pipe(map(response => response));
  }

  purgeData() {
    return this.authHttp
      .delete(`${this.alertApiUrl}`)
      .pipe(map(response => response));
  }

  exportCSV(
    searchKeyword: string = '',
    startDate: Date,
    endDate: Date,
    alertType = '',
    portalVersionInfo
  ) {
    const urlEncoded = encodeURIComponent(searchKeyword);
    let _startDate = '1900-01-01';
    let _endDate = '9999-01-01';
    if (startDate) {
      _startDate = moment(startDate)
        .startOf('day')
        .format('YYYY-MM-DD HH:mm:ss');
    }
    if (endDate) {
      _endDate = moment(endDate)
        .endOf('day')
        .format('YYYY-MM-DD HH:mm:ss');
    }

    let query = `?searchKeyword=${urlEncoded}&startDate=${_startDate}&endDate=${_endDate}&version=${portalVersionInfo}`;
    if (alertType) {
      query = query + `&alertType=${alertType}`;
    }

    this.authShortlivedCookie.setFor(5);
    window.location.href = `${this.alertApiUrl}/export-csv${query}`;
  }

  getUserAlerts() {
    const token = localStorage.getItem('token');
    if (this.userrole != RoleEnum.Admin || this.showLicenseOnly || !token)
      return;
    this.getAlertsNotification(10, this.alertLastId).subscribe({
      next: (res: any) => {
        if (this.alertLastId) {
          const newAlerts = (res || []).filter(a => a.id > this.alertLastId);
          this.alertCount += newAlerts.length;
          this.alerts = [...(newAlerts || []), ...this.alerts];
        } else {
          this.alerts = res || [];
        }
        if (this.alerts.length != 0) this.alertLastId = this.alerts[0].id;
      },
      error: err => {
        if (err?.error?.message == 'jwt expired') {
          localStorage.removeItem('user');
          localStorage.removeItem('currentUser');
          localStorage.removeItem('token');

          this.router.navigate(['/login']);
        }
      }
    });
  }
}
