import { Component, OnInit } from '@angular/core';
import { Router, ActivatedRoute } from '@angular/router';
import { ControllerService } from '../controller.service';
import { Controller } from '../../models/controller';
import { Pagination } from '../../models/pagination';
import { PreserveSearch } from '../../models/preserve-search';
import * as _ from 'lodash';
import * as moment from 'moment';
import { timer } from 'rxjs';
import { GlobalMessageClass } from '../../global-messages/global-messages';
import { GlobalMessageService } from '../../global-messages/global-messages.service';
import { AuthenticationService } from '../../login/authentication.service';
import { Popup } from '../../Popup';
import { Subscription } from 'rxjs';
import { Subject } from 'rxjs';
import { Title } from '@angular/platform-browser';
import {
  debounceTime,
  distinctUntilChanged,
  first,
  switchMap,
  tap
} from 'rxjs/operators';
import { ErrorCodeEnum } from '../../shared/emuns/error.enum';
import { InstallerStatusEnum, RoleEnum } from '../../shared/emuns/global.enum';
import { LicenseErrorEnum } from '../../shared/emuns/license.enum';
import { ConfirmModalComponent } from '../../../app/shared/confirm-modal.component';
import { MatDialog } from '@angular/material/dialog';
import { OtpAuthenticationService } from '../../otp-verification/otp-authentication.service';
import { RolesAndPermissionService } from '../../roles-and-permission/roles-and-permission.service';

@Component({
  selector: 'app-controller-list',
  templateUrl: './controller-list.component.html',
  styleUrls: ['./controller-list.component.css']
})
export class ControllerListComponent implements OnInit {
  showHelpText: boolean = false;
  public downloadInitialtedFor = new Array();
  public alert: GlobalMessageClass;
  controllers: any = {};
  private controllersList = new Array();
  private selectedController: Controller;
  selectedControllerId: number;
  installerStatus: Popup = new Popup();
  private activeStatus = true;
  private hearbeatTimeThresholdInSeconds: number = 120;
  pagination: Pagination = new Pagination();
  private timerSubscription: Subscription;
  showPasswordConfimation = false;
  showHealth = false;
  currentControllerHealth: any;

  preserveSearch = new PreserveSearch();
  navigationToEdit: boolean = false;

  loading: Boolean = true;

  pageSize = [10, 25, 50, 100, 200];

  sort: string;
  sortDir: string;

  $searchController = new Subject();
  searchKeyword;

  InstallerStatus = InstallerStatusEnum;
  Role = RoleEnum;
  isDarkMode: boolean;
  showOtpDialog: boolean = false;
  otptoken: string = '';
  option: string;
  private otpStatusSubscription: Subscription;
  

  createPermission: boolean = true;
  readPermission: boolean = true;
  updatePermission: boolean = true;
  deletePermission: boolean = true;

  constructor(
    private router: Router,
    private globalMessageService: GlobalMessageService,
    private controllerService: ControllerService,
    private route: ActivatedRoute,
    private authenticationService: AuthenticationService,
    public dialog: MatDialog,
    private titleService: Title,
    private OtpAuthenticationService: OtpAuthenticationService,
    private rolesPermissionService : RolesAndPermissionService
  ) {
    this.titleService.setTitle('Infrastructure | Controllers');
  }

  ngOnInit() {
    this.isDarkMode = JSON.parse(localStorage.getItem('darkMode'));
    this.getCRUDdata(24);
    let controllerSearch = this.preserveSearch.getSearch('controllerSearch');
    if (controllerSearch) {
      this.pagination = controllerSearch.pagination;
      this.searchKeyword = controllerSearch.searchKeyword;
    }
    this.getPage(this.pagination.currentPage);
    this.refreshData();
    this.subscibeToSearch();
  }

  getCRUDdata(privId){
    this.rolesPermissionService.getCrudData(privId).subscribe((res: any) => {
      this.createPermission = res?.create_permission === 1 ? true : false
      this.readPermission = res?.read_permission === 1 ? true : false
      this.updatePermission = res?.update_permission === 1 ? true : false
      this.deletePermission = res?.delete_permission === 1 ? true : false
    });
  }
  subscibeToSearch() {
    this.$searchController
      .pipe(
        debounceTime(500),
        distinctUntilChanged(),
        switchMap(data => {
          this.loading = true;
          this.searchKeyword = data;
          this.pagination.currentPage = 1;
          return this.controllerService.getControllers(
            this.pagination.currentPage,
            this.pagination.itemsPerPage,
            this.sort,
            this.sortDir,
            this.searchKeyword
          );
        })
      )
      .subscribe({
        next: (gateways: any) => {
          this.loading = false;
          const { recordsTotal } = gateways;
          this.controllers = gateways;
          this.pagination.totalItems = recordsTotal;
        },
        error: err => {
          this.loading = false;
        }
      });
  }

  public refreshData(number = null): void {
    if (this.timerSubscription) {
      this.timerSubscription.unsubscribe();
    }
    if (number != null) {
      this.pagination.currentPage = number;
    }
    this.getPage(this.pagination.currentPage);
    this.subscribeToData();
  }

  private subscribeToData(): void {
    this.timerSubscription = timer(10000)
      .pipe(first())
      .subscribe(() => this.refreshData());
  }

  public ngOnDestroy(): void {
    if (this.timerSubscription) {
      this.timerSubscription.unsubscribe();
    }

    if (!this.navigationToEdit) {
      this.preserveSearch.saveSearch('controllerSearch', '');
    }

    if (this.otpStatusSubscription) {
      this.otpStatusSubscription.unsubscribe();
    }
  }

  statusClicked(controller) {
    this.currentControllerHealth = controller.ControllerHealth
      ? controller.ControllerHealth
      : {};
    this.showHealth = true;
    this.selectedControllerId = controller.id;
  }

  cancelHealth(event) {
    this.showHealth = false;
  }

  cancelStatusPopup() {
    this.installerStatus.show = false;
  }

  private onStatus(controller: Controller): void {
    this.installerStatus.heading = 'Installer Status';
    this.installerStatus.body = controller.installer_status;
    this.installerStatus.action = 'Confirm';
    this.installerStatus.show = true;
    setTimeout(() => {
      this.cancelStatusPopup();
    }, 2000);
  }

  getPage(page: number) {
    this.loading = true;
    this.pagination.currentPage = page;
    this.controllerService
      .getControllers(
        this.pagination.currentPage,
        this.pagination.itemsPerPage,
        this.sort,
        this.sortDir,
        this.searchKeyword
      )
      .pipe(
        tap((res: any) => {
          this.loading = false;
          this.pagination.totalItems = res.recordsTotal;
          this.computeOnlineStatus(res.data || []);
          this.initiateDownload(res.data || []);
          this.controllers = res;
          if (
            res.data.length !== 0 &&
            this.currentControllerHealth &&
            this.currentControllerHealth.id
          ) {
            try {
              const health =
                this.controllers &&
                this.controllers.data &&
                this.controllers.data.filter(
                  (_controller: any) =>
                    _controller.ControllerHealth &&
                    _controller.ControllerHealth.id ==
                      this.currentControllerHealth.id
                )[0].ControllerHealth;
              this.currentControllerHealth = health ? health : {};
            } catch (e) {
              console.log('error was', e);
            }
          }
          this.pagination.currentPage = page;
        })
      )
      .subscribe({
        next: next => {},
        error: err => {
          this.loading = false;
          if (err.status === ErrorCodeEnum.Forbidden) {
            if (err._body == LicenseErrorEnum.Invalid) {
              this.authenticationService.showLicenseOnly(true);
              this.router.navigate(['/license']);
            } else {
              this.alert = {
                closed: false,
                type: 'danger',
                message: 'You are not authorized to access this page.'
              };
              this.globalMessageService.pushAlert(this.alert);
            }
          }
        },
        complete: () => {}
      });
  }

  clearfun() {
    this.searchKeyword = '';
    this.getPage(this.pagination.currentPage);
  }

  private initiateDownload(controllers) {
    var index = -1;
    controllers.forEach(controller => {
      if (this.downloadInitialtedFor.includes(controller.id)) {
        if (
          controller.installer_status === InstallerStatusEnum.New &&
          controller.isDownloadPackage === true &&
          controller.packageDownloaded === false
        ) {
          index = this.downloadInitialtedFor.indexOf(controller.id);
          if (index > -1) {
            this.downloadInitialtedFor.splice(index, 1);
          }
          window.location.assign(controller.onPrenDownloadLink);
        }
      }
    });
  }

  private computeOnlineStatus(controllers) {
    let heartBeatSentAt;
    let allowedTime: moment.Moment = moment().subtract(
      this.hearbeatTimeThresholdInSeconds,
      'seconds'
    );
    let diffInSeconds: number;
    let heartBeat;

    controllers.forEach(controller => {
      heartBeat = controller.heart_beat;
      heartBeatSentAt = moment(controller.heart_beat_sent_at);
      diffInSeconds = heartBeatSentAt.diff(allowedTime, 'second');

      if (heartBeat) {
        if (diffInSeconds >= 0) {
          controller.is_online = true;
        } else {
          controller.is_online = false;
        }
      } else {
        controller.is_online = false;
      }
    });
  }

  private onSelect(controller: Controller): void {
    let mfa = localStorage.getItem('adminMfaValue');
    if (mfa === 'true') {
      this.showOtpDialog = true;
      this.option = 'edit';
      this.otpStatusSubscription = this.OtpAuthenticationService.otpStatusCheckEdit$.subscribe(
        (res: any) => {
          if (res == true) {
            let controllerSearch = {
              pagination: this.pagination,
              searchKeyword: this.searchKeyword
            };
            this.preserveSearch.saveSearch(
              'controllerSearch',
              controllerSearch
            );
            this.navigationToEdit = true;
            this.selectedController = controller;
            this.router.navigate(['/controllers', this.selectedController.id]);
            this.OtpAuthenticationService.setOtpStatusValueEdit(false);
          }
        }
      );
    } else {
      let controllerSearch = {
        pagination: this.pagination,
        searchKeyword: this.searchKeyword
      };
      this.preserveSearch.saveSearch('controllerSearch', controllerSearch);
      this.navigationToEdit = true;
      this.selectedController = controller;
      this.router.navigate(['/controllers', this.selectedController.id]);
    }
  }

  cancelVerificationPassword(event) {
    this.showPasswordConfimation = false;
  }

  submitVerificationPassword(event) {
    this.controllerService
      .delete_controller(this.selectedController.id, event)
      .subscribe(response => {
        this.showPasswordConfimation = false;
        if (response.success) {
          this.alert = {
            closed: false,
            type: 'success',
            message: 'Controller has been deleted successfully.'
          };
          this.globalMessageService.pushAlert(this.alert);
          this.getPage(this.pagination.currentPage);
        } else {
          this.alert = {
            closed: false,
            type: 'danger',
            message: response.message
          };
          this.globalMessageService.pushAlert(this.alert);
        }
      });
  }

  private onDelete(controller: Controller): void {
    let mfa = localStorage.getItem('adminMfaValue');
    if (mfa === 'true') {
      this.showPasswordConfimation = false;
      this.showOtpDialog = true;
      this.option = 'delete';
      this.otpStatusSubscription = this.OtpAuthenticationService.otpStatusCheckDel$.subscribe(
        (res: any) => {
          console.log('RES', res);

          if (res == true) {
            this.controllerService
              .delete_controller_by_otp(controller.id)
              .subscribe(response => {
                this.showPasswordConfimation = false;
                if (response.success) {
                  this.alert = {
                    closed: false,
                    type: 'success',
                    message: 'Controller has been deleted successfully.'
                  };
                  this.globalMessageService.pushAlert(this.alert);
                  this.getPage(this.pagination.currentPage);
                } else {
                  this.alert = {
                    closed: false,
                    type: 'danger',
                    message: response.message
                  };
                  this.globalMessageService.pushAlert(this.alert);
                }
              });
            this.showOtpDialog = false;
            this.OtpAuthenticationService.setOtpStatusValueDel(false);
          }
        }
      );
    } else {
      const secondaryCount = this.controllers.data.find(
        controller => controller.Role.role === RoleEnum.Secondary
      );
      if (
        controller.Role.role === RoleEnum.Secondary ||
        (!secondaryCount && controller.Role.role === RoleEnum.Primary)
      ) {
        this.showPasswordConfimation = true;
        this.selectedController = controller;
      } else {
        this.alert = {
          closed: false,
          type: 'danger',
          message:
            'Cannot delete primary controller if secondary controller exist.'
        };
        this.globalMessageService.pushAlert(this.alert);
      }
    }
  }

  onDownload(controller: Controller): void {
    const dialogRef = this.dialog.open(ConfirmModalComponent, {
      panelClass: this.isDarkMode ? 'dark-theme' : '',
      data: {
        title: 'Confirmation',
        message:
          'Are you sure you want to request a new service installer and download it?'
      }
    });

    dialogRef.afterClosed().subscribe(result => {
      if (result) {
        this.selectedController = controller;
        this.controllerService.download(controller);
      }
    });
  }

  private onGenerate(controller: Controller): void {
    const dialogRef = this.dialog.open(ConfirmModalComponent, {
      panelClass: this.isDarkMode ? 'dark-theme' : '',
      data: {
        title: 'Confirmation',
        message: `Are you sure you want to request a new service installer?`
      }
    });

    dialogRef.afterClosed().subscribe(result => {
      if (result) {
        this.selectedController = controller;

        this.controllerService
          .generate_client(controller.id, 0)
          .subscribe(response => {
            if (response.status) {
              this.alert = {
                closed: false,
                type: 'success',
                message: 'Installer has been sent successfully.'
              };
              this.globalMessageService.pushAlert(this.alert);
              this.getPage(this.pagination.currentPage);
            } else {
              this.alert = {
                closed: false,
                type: 'danger',
                message: 'Failed to send the installer.'
              };
              this.globalMessageService.pushAlert(this.alert);
            }
          });
      }
    });
  }

  add_controller(): void {
    this.router.navigate(['/create-controller']);
  }

  toggle(controller: Controller, index) {
    this.controllerService.change_status(controller).subscribe(response => {
      if (response.success) {
        this.controllers.data[index].is_active = !this.controllers.data[index]
          .is_active;
        if (this.controllers.data[index].is_active) {
          this.alert = {
            closed: false,
            type: 'success',
            message: 'Controller Activated successfully.'
          };
          this.globalMessageService.pushAlert(this.alert);
        } else {
          this.alert = {
            closed: false,
            type: 'danger',
            message: 'Controller De-Activated successfully.'
          };
          this.globalMessageService.pushAlert(this.alert);
        }
      } else {
        this.alert = {
          closed: false,
          type: 'danger',
          message: 'Failed to change Controller status.'
        };
        this.globalMessageService.pushAlert(this.alert);
      }
    });
  }

  pageSizeChange(size) {
    this.pagination.itemsPerPage = +size;
    this.pagination.totalItems = 0;
    this.pagination.currentPage = 1;
    this.getPage(this.pagination.currentPage);
  }

  sortRecord(term: string) {
    this.sortDir = this.sortDir == 'asc' ? 'desc' : 'asc';
    this.sort = term;

    this.getPage(this.pagination.currentPage);
  }

  cancelOtpVerificationPassword(event: any) {
    this.showOtpDialog = false;
    this.OtpAuthenticationService.setOtpStatusValueEdit(false);
    this.OtpAuthenticationService.setOtpStatusValueDel(false);
  }

  submitOtpVerificationPassword(event: any) {
    this.otptoken = event;
    // this.otptoken = event
  }
  getClosePopup(event: any) {
    this.showOtpDialog = false;
  }
}
