import { Injectable } from '@angular/core';

import { ConfigService } from '../services/config.service';
import { HttpClient as AuthHttp, HttpHeaders } from '@angular/common/http';
import { AuthenticationService } from '../login/authentication.service';
import { PaginationResponse } from '../models/pagination-response';
import { Controller } from '../models/controller';
import { Gateway } from '../models/gateway';
import { Observable } from 'rxjs';

import { AuthShortlivedCookieService } from '../auth-shortlived-cookie.service';
import { map } from 'rxjs/operators';

@Injectable()
export class ControllerService {
  private controllersUrl = 'http://localhost:3000/controllers';
  private controllersApiUrl = '/controllers';
  private whiteListedHostApiUrl = '/whitelisted-hosts';
  private controllersChangeStatusApiUrl = '/controllers/status';
  private controllers: Controller[];
  private apiBaseUrl: string;
  httpOptions = {
    headers: new HttpHeaders({
      'Content-Type': 'application/json'
    })
  };

  constructor(
    private authenticationService: AuthenticationService,
    private authShortlivedCookie: AuthShortlivedCookieService,
    private authHttp: AuthHttp,
    private configService: ConfigService
  ) {
    this.apiBaseUrl = this.configService.getConfiguration().apiBaseUrl;
    this.controllersApiUrl = this.apiBaseUrl + this.controllersApiUrl;
    this.controllersChangeStatusApiUrl =
      this.apiBaseUrl + this.controllersChangeStatusApiUrl;
    this.whiteListedHostApiUrl = this.apiBaseUrl + this.whiteListedHostApiUrl;
  }

  /** GET from the server */

  getControllers(
    page: number = 1,
    pageSize: number = 10,
    sort: string = 'id',
    dir: string = 'desc',
    searchKey = ''
  ): any {
    var queryString = `?limit=${pageSize}&offset=${page -
      1}&sort=${sort}&dir=${dir}&searchKey=${searchKey}`;
    return this.authHttp
      .get(this.controllersApiUrl + queryString)
      .pipe(map(response => response));
  }

  getAllControllers(): any {
    return this.authHttp
      .get(this.controllersApiUrl)
      .pipe(map(response => response));
  }

  getcontrollerGraphs(controllerId: number, interval: number = 1) {
    return this.authHttp
      .get(
        `${this.controllersApiUrl}/health/graph/${controllerId}?interval=${interval}`
      )
      .pipe(map(response => response));
  }

  add_controller(controller: Controller): any {
    let headers = new HttpHeaders({ 'Content-Type': 'application/json' });
    let options = { headers: headers };
    let queryString = '';

    if (controller.id > 0 && typeof controller.id != 'undefined') {
      return this.authHttp
        .put(
          this.controllersApiUrl + '/' + controller.id,
          { controller: controller },
          options
        )
        .pipe(map(this.extractData));
      //.catch(this.handleError);
    } else {
      return this.authHttp
        .post(this.controllersApiUrl, { controller: controller }, options)
        .pipe(map(response => response));
    }
  }

  change_status(controller: Controller): Observable<any> {
    let headers = new HttpHeaders({ 'Content-Type': 'application/json' });
    let options = { headers: headers };
    return this.authHttp
      .put(
        this.controllersChangeStatusApiUrl,
        { controller: controller },
        options
      )
      .pipe(map(response => response));
  }

  delete_controller(id: number, verificationPassword): any {
    var queryString = '';

    if (id > 0) {
      queryString = '/' + id;
    }
    this.httpOptions['body'] = { verificationPassword };
    return this.authHttp
      .delete(this.controllersApiUrl + queryString, this.httpOptions)
      .pipe(map(this.extractData));
    //.catch(this.handleError);
  }

  delete_controller_by_otp(id: number): any {
    var queryString = '';

    if (id > 0) {
      queryString = '/otp/' + id;
    }

    return this.authHttp
      .delete(this.controllersApiUrl + queryString)
      .pipe(map(this.extractData));
    //.catch(this.handleError);
  }

  getcontrollerdetail(id): Observable<Controller | any> {
    var queryString = '';

    if (id != '') {
      queryString = '/' + id;
    }
    return this.authHttp
      .get(this.controllersApiUrl + queryString)
      .pipe(map(response => response));
  }


  generate_client(id: number, isDownload = 0): any {
    var queryString = '';

    if (id > 0) {
      queryString = '/' + id;
    }
    if (isDownload == 1) {
      queryString += '/' + 1;
    } else {
      queryString += '/' + 0;
    }
    return this.authHttp
      .get(`${this.controllersApiUrl}/generate/service${queryString}`)
      .pipe(map(this.extractData));
    //.catch(this.handleError);
  }

  download(controller) {
    const { id } = controller;

    this.authShortlivedCookie.setFor(5);

    window.location.href =
      this.apiBaseUrl + `/controllers/download/service/${id}`;
  }

  getWhiteListedHostes(controllerId) {
    return this.authHttp
      .get(`${this.whiteListedHostApiUrl}/hosts/${controllerId}`)
      .pipe(map(this.extractData));
    //.catch(this.handleError);
  }

  createWhiteListedHostes(port, protocol, host, controllerId) {
    return this.authHttp
      .post(`${this.whiteListedHostApiUrl}/hosts/${controllerId}`, {
        port,
        protocol,
        host
      })
      .pipe(map(this.extractData));
    //.catch(this.handleError);
  }

  updateWhiteListedHostes(port, protocol, host, id, controllerId) {
    return this.authHttp
      .put(`${this.whiteListedHostApiUrl}/${id}`, {
        port,
        protocol,
        host,
        controllerId
      })
      .pipe(map(this.extractData));
    //.catch(this.handleError);
  }

  deleteWhiteListedHostes(id) {
    return this.authHttp
      .delete(`${this.whiteListedHostApiUrl}/${id}`)
      .pipe(map(this.extractData));
    //.catch(this.handleError);
  }

  private extractData(res: Response) {
    let body = res;
    return body || {};
  }

  // private handleError(error: any) {
  //   // In a real world app, we might use a remote logging infrastructure
  //   let errMsg: string;
  //   if (error) {
  //     const body = error || '';
  //     const err = body.error || JSON.stringify(body);
  //     errMsg = `${error.status} - ${error.statusText || ''} ${err}`;
  //   } else {
  //     errMsg = error.message ? error.message : error.toString();
  //   }
  //   console.error(errMsg);
  //   return Observable.throw(errMsg);
  // }
}
