import { Injectable } from '@angular/core';
import { HttpClient as AuthHttp } from '@angular/common/http';
import { ConfigService } from '../services/config.service';
import { Pagination } from '../models/pagination';
import { map } from 'rxjs/operators';
import { Title } from '@angular/platform-browser';

@Injectable()
export class AccessPolicyService {
  private policyApiUrl: string;
  private localUserApiUrl: string;
  constructor(
    private authHttp: AuthHttp,
    private configService: ConfigService,
    private titleService: Title
  ) {
    this.titleService.setTitle('Policy Management | Entitlements');
    const baseUrl = configService.getConfiguration().apiBaseUrl;
    this.policyApiUrl = baseUrl + '/policies/access';
    this.localUserApiUrl = baseUrl + '/localUsers';
  }

  getAllPolicies(
    pagination: Pagination,
    name: string = '',
    sort: string = 'id',
    dir: string = 'desc',
    selectedClientIds = [],
    allClientPolicies: boolean = false,
    policyType = ''
  ) {
    const currentPage = pagination.currentPage - 1 || 0;
    const itemsPerPage = pagination.itemsPerPage || 10;

    let _name: string = '';

    if (name.trim()) {
      _name = name.trim();
    }

    let _clients = selectedClientIds.join(',');

    return this.authHttp
      .get(
        this.policyApiUrl +
          `?offset=${currentPage}&limit=${itemsPerPage}&name=${_name}&sort=${sort}&dir=${dir}&clients=${_clients}&allClientPolicies=${allClientPolicies}&policy=${policyType}`
      )
      .pipe(
        map(response => response),
        map((res: any) => {
          return {
            ...res,
            data: (res.data || []).map(v => ({
              ...v,
              clientName: v.all_clients
                ? 'All Clients'
                : v.Clients && v.Clients[0] && v.Clients[0].Name
                ? v.Clients[0].Name +
                  `${
                    v.Clients.length + (v.Groups || []).length > 1 ? '...' : ''
                  }`
                : v.Groups && v.Groups[0] && v.Groups[0].name
                ? v.Groups[0].name + `${v.Groups.length > 1 ? '...' : ''}`
                : 'Unknown',
              serviceName: v.all_services
                ? 'All Services'
                : v.Services && v.Services[0] && v.Services[0].Name
                ? v.Services[0].Name +
                  `${
                    v.Services.length + (v.ServiceGroups || []).length > 1
                      ? '...'
                      : ''
                  }`
                : v.ServiceGroups &&
                  v.ServiceGroups[0] &&
                  v.ServiceGroups[0].name
                ? v.ServiceGroups[0].name +
                  `${v.ServiceGroups.length > 1 ? '...' : ''}`
                : 'Unknown',
              userName: v.all_users
                ? 'All Users'
                : v.Users && v.Users[0] && v.Users[0].Name
                ? v.Users[0].Name +
                  `${
                    v.Users.length + (v.UserGroups || []).length > 1
                      ? '...'
                      : ''
                  }`
                : v.UserGroups && v.UserGroups[0] && v.UserGroups[0].name
                ? v.UserGroups[0].name +
                  `${v.UserGroups.length > 1 ? '...' : ''}`
                : 'Unknown'
            }))
          };
        })
      );
  }

  getAllPoliciesWithoutPagination(
    name: string = '',
    sort: string = 'id',
    dir: string = 'desc',
    selectedClientIds = [],
    allClientPolicies: boolean = false,
    policyType = ''
  ) {
    let _name: string = '';

    if (name.trim()) {
      _name = name.trim();
    }

    let _clients = selectedClientIds.join(',');

    // To fix this Find All without Pagination
    return this.authHttp
      .get(
        this.policyApiUrl +
          `?offset=0&limit=1000&name=${_name}&sort=${sort}&dir=${dir}&clients=${_clients}&allClientPolicies=${allClientPolicies}&policy=${policyType}`
      )
      .pipe(
        map(response => response),
        map((res: any) => {
          return {
            ...res,
            data: (res.data || []).map(v => ({
              ...v,
              clientName: v.all_clients
                ? 'All Clients'
                : v.Clients && v.Clients[0] && v.Clients[0].Name
                ? v.Clients[0].Name +
                  `${
                    v.Clients.length + (v.Groups || []).length > 1 ? '...' : ''
                  }`
                : v.Groups && v.Groups[0] && v.Groups[0].name
                ? v.Groups[0].name + `${v.Groups.length > 1 ? '...' : ''}`
                : 'Unknown',
              serviceName: v.all_services
                ? 'All Services'
                : v.Services && v.Services[0] && v.Services[0].Name
                ? v.Services[0].Name +
                  `${
                    v.Services.length + (v.ServiceGroups || []).length > 1
                      ? '...'
                      : ''
                  }`
                : v.ServiceGroups &&
                  v.ServiceGroups[0] &&
                  v.ServiceGroups[0].name
                ? v.ServiceGroups[0].name +
                  `${v.ServiceGroups.length > 1 ? '...' : ''}`
                : 'Unknown',
              userName: v.all_users
                ? 'All Users'
                : v.Users && v.Users[0] && v.Users[0].Name
                ? v.Users[0].Name +
                  `${
                    v.Users.length + (v.UserGroups || []).length > 1
                      ? '...'
                      : ''
                  }`
                : v.UserGroups && v.UserGroups[0] && v.UserGroups[0].name
                ? v.UserGroups[0].name +
                  `${v.UserGroups.length > 1 ? '...' : ''}`
                : 'Unknown'
            }))
          };
        })
      );
  }

  deletePolicy(id) {
    return this.authHttp
      .delete(this.policyApiUrl + `/${id}`)
      .pipe(map(response => response));
  }

  enablePolicy(id, value) {
    return this.authHttp
      .patch(this.policyApiUrl + `/${id}/enable/${value}`, {})
      .pipe(map(response => response));
  }

  verifyPolicies(
    userId?,
    serviceGroupId?,
    selectedIds = [],
    gatewayGroupId?,
    gatewayId?,
    clientGroupId?,
    allServices?
  ) {
    return this.authHttp
      .get(
        this.policyApiUrl +
          `/verifyPolicies` +
          `?userId=${userId}` +
          `&serviceGroupId=${serviceGroupId}` +
          `&selectedIds=${selectedIds}` +
          `&gatewayGroupId=${gatewayGroupId}` +
          `&gatewayId=${gatewayId}` +
          `&clientGroupId=${clientGroupId}` +
          `&allServices=${allServices}`
      )
      .pipe(map(response => response));
  }

  getUserPolicyById(id) {
    return this.authHttp
      .get(`${this.policyApiUrl}/user/${id}`)
      .pipe(map(response => response));
  }

  getSinglePolicy(id) {
    return this.authHttp
      .get(`${this.policyApiUrl}/${id}`)
      .pipe(map(response => response));
  }

  getAllDropDownData(
    clientOffset,
    serviceOffset,
    clientName,
    serviceName,
    policyId,
    clientType,
    only_non_persistent: boolean = true
  ) {
    let policyQuery = `&policyId=`;
    if (policyId) {
      policyQuery += policyId;
    }
    return this.authHttp
      .get(
        `${this.policyApiUrl}/filters` +
          `?clientOffset=${clientOffset}` +
          `&serviceOffset=${serviceOffset}` +
          `&clientName=${clientName}` +
          `&serviceName=${serviceName}` +
          `&clientType=${clientType}` +
          `&only_non_persistent=${only_non_persistent}` +
          `${policyQuery}`
      )
      .pipe(map(response => response));
  }

  getAllDropDownDataWithoutPagination(
    clientOffset,
    serviceOffset,
    clientName,
    serviceName,
    policyId,
    clientType
  ) {
    let policyQuery = `&policyId=`;
    if (policyId) {
      policyQuery += policyId;
    }
    return this.authHttp
      .get(
        `${this.policyApiUrl}/filterswithoutpagination` +
          `?clientOffset=${clientOffset}` +
          `&serviceOffset=${serviceOffset}` +
          `&clientName=${clientName}` +
          `&serviceName=${serviceName}` +
          `&clientType=${clientType}` +
          `${policyQuery}`
      )
      .pipe(map(response => response));
  }

  getClients(clientOffset, searchTerm) {
    return this.authHttp
      .get(
        `${this.policyApiUrl}/clients?offset=${clientOffset}&name=${searchTerm}`
      )
      .pipe(map(response => response));
  }

  getUserDropDown() {
    return this.authHttp.get(`${this.localUserApiUrl}/userdropdown`);
  }

  createService(policyOption) {
    return this.authHttp
      .post(`${this.policyApiUrl}/create`, policyOption)
      .pipe(map(response => response));
  }

  updateService(policyOption, id) {
    return this.authHttp.put(`${this.policyApiUrl}/${id}`, policyOption);
  }
}
