import { Injectable } from '@angular/core';
import { HttpClient as AuthHttp } from '@angular/common/http';
import { ConfigService } from '../services/config.service';
import { AuthShortlivedCookieService } from '../auth-shortlived-cookie.service';
import { map } from 'rxjs/operators';

@Injectable()
export class DeviceService {
  private apiBaseUrl: string;
  private devicesUrl = '/devices';

  constructor(
    private authHttp: AuthHttp,
    private configService: ConfigService,
    private authShortlivedCookie: AuthShortlivedCookieService
  ) {
    this.apiBaseUrl = this.configService.getConfiguration().apiBaseUrl;
    this.devicesUrl = this.apiBaseUrl + this.devicesUrl;
  }

  getDevices(
    pagination,
    sort: string = 'id',
    dir: string = 'desc',
    searchKey = '',
    deviceType = '',
    organization = '',
    department = '',
    location = '',
    osVersion = '',
    os = '',
    referenceMachineCheck
  ) {
    let query =
      `?limit=${pagination.itemsPerPage}` +
      `&offset=${pagination.currentPage - 1}` +
      `&sort=${sort}` +
      `&dir=${dir}` +
      `&searchTerm=${searchKey}` +
      `&deviceType=${deviceType}` +
      `&organization=${organization}` +
      `&department=${department}` +
      `&location=${location}` +
      `&os=${os}` +
      `&osVersion=${osVersion}` +
      `&referenceMachineCheck=${referenceMachineCheck}`;
    return this.authHttp
      .get(`${this.devicesUrl}${query}`)
      .pipe(map(response => response));
  }

  getWorkloadDevices(
    pagination,
    sort: string = 'id',
    dir: string = 'desc',
    searchKey = '',
    deviceType = '',
    organization = '',
    department = '',
    location = '',
    osVersion = '',
    os = '',
    referenceMachineCheck
  ) {
    let query =
      `?limit=${pagination.itemsPerPage}` +
      `&offset=${pagination.currentPage - 1}` +
      `&sort=${sort}` +
      `&dir=${dir}` +
      `&searchTerm=${searchKey}` +
      `&deviceType=${deviceType}` +
      `&organization=${organization}` +
      `&department=${department}` +
      `&location=${location}` +
      `&os=${os}` +
      `&osVersion=${osVersion}` +
      `&referenceMachineCheck=${referenceMachineCheck}`;
    return this.authHttp
      .get(`${this.devicesUrl}/workload/${query}`)
      .pipe(map(response => response));
  }

  getAvailableDevices(serverSideDeviceFound, mobileDeviceFound, departmentId) {
    return this.authHttp
      .get(
        `${this.devicesUrl}/available/${serverSideDeviceFound}/${mobileDeviceFound}/${departmentId}`
      )
      .pipe(map(response => response));
  }

  clientsforAvailableDevices(devices) {
    return this.authHttp
      .put(
        `${this.devicesUrl}/available-devices`,
        { devices }
      )
      .pipe(map(response => response));
  }

  getAllLocations() {
    return this.authHttp
      .get(`${this.devicesUrl}/locations`)
      .pipe(map(response => response));
  }

  devicesValidation() {
    return this.authHttp
      .get(`${this.devicesUrl}/devicesValidation`)
      .pipe(map(response => response));
  }

  getDevicesForUser(
    pagination,
    sort: string = 'id',
    dir: string = 'desc',
    searchKey = '',
    userId = ''
  ) {
    let query =
      `?limit=${pagination.itemsPerPage}` +
      `&offset=${pagination.currentPage - 1}` +
      `&sort=${sort}` +
      `&dir=${dir}` +
      `&searchTerm=${searchKey}`;

    return this.authHttp
      .get(`${this.devicesUrl}/user/${userId}${query}`)
      .pipe(map(response => response));
  }

  searchDevices(
    searchQuery: string,
    pageSize: number = 10,
    page: number = 1,
    sort: string = 'id',
    dir: string = 'desc',
    deviceType = '',
    organization = '',
    department = '',
    location = '',
    osVersion = '',
    os = ''
  ) {
    const encodedUrlComponent = encodeURIComponent(searchQuery);
    let query =
      `?limit=${pageSize}` +
      `&offset=${page - 1}` +
      `&sort=${sort}` +
      `&dir=${dir}` +
      `&searchTerm=${encodedUrlComponent}` +
      `&deviceType=${deviceType}` +
      `&organization=${organization}` +
      `&department=${department}` +
      `&location=${location}` +
      `&osVersion=${osVersion}` +
      `&os=${os}`;

    return this.authHttp
      .get(`${this.devicesUrl}${query}`)
      .pipe(map(response => response));
  }

  searchDevicesForUser(
    searchQuery: string,
    pageSize: number = 10,
    page: number = 1,
    sort: string = 'id',
    dir: string = 'desc',
    userId = null
  ) {
    const encodedUrlComponent = encodeURIComponent(searchQuery);
    let query =
      `?limit=${pageSize}` +
      `&offset=${page - 1}` +
      `&sort=${sort}` +
      `&dir=${dir}` +
      `&searchTerm=${encodedUrlComponent}` +
      `&userId=${userId}`;

    return this.authHttp
      .get(`${this.devicesUrl}${query}`)
      .pipe(map(response => response));
  }

  deleteDevice(id) {
    return this.authHttp
      .delete(`${this.devicesUrl}/${id}`)
      .pipe(map(response => response));
  }

  getDevice(id) {
    return this.authHttp
      .get(`${this.devicesUrl}/${id}`)
      .pipe(map(response => response));
  }

  getGlobalDarkeningState() {
    return this.authHttp
      .get(`${this.devicesUrl}/globalDarkening`)
      .pipe(map(response => response));
  }

  getClients(deviceId) {
    return this.authHttp
      .get(`${this.devicesUrl}/clients/${deviceId}`)
      .pipe(map(response => response));
  }

  toggleDarkeningState(deviceId, currentState) {
    return this.authHttp
      .put(`${this.devicesUrl}/toggleDarkening/${deviceId}`, { currentState })
      .pipe(map(response => response));
  }

  addDevice(device, editing) {
    if (!editing) {
      return this.authHttp
        .post(`${this.devicesUrl}`, { device })
        .pipe(map(response => response));
    } else {
      return this.authHttp
        .put(`${this.devicesUrl}/${device.id}`, { device })
        .pipe(map(response => response));
    }
  }

  uploadCsvFile(fileData) {
    return this.authHttp
      .post(`${this.devicesUrl}/csv/upload`, { fileData: fileData })
      .pipe(map(response => response));
  }

  downloadCsv() {
    this.authShortlivedCookie.setFor(5);
    window.location.href = `${this.devicesUrl}/csv/download`;
  }

  bulkActions(
    type = '',
    status = '',
    selectAll = false,
    selectedIds,
    organization = '',
    department,
    location = '',
    searchKey = '',
    device_type = '',
    os
  ) {
    return this.authHttp.put(`${this.devicesUrl}/bulk-actions`, {
      type,
      status,
      selectAll,
      selectedIds: Array.from(selectedIds),
      department,
      location,
      searchKey,
      device_type,
      organization,
      os
    });
  }
}
