import { Component, OnInit } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { GatewayService } from '../gateway.service';
import { Pagination } from '../../models/pagination';
import { Gateway } from '../../models/gateway';
import { ContextHelpEnum } from '../../shared/emuns/global.enum';
import { DeviceOSEnum } from '../../shared/emuns/client.enum';
import { GlobalMessageClass } from '../../global-messages/global-messages';
import { GlobalMessageService } from '../../global-messages/global-messages.service';

@Component({
  selector: 'app-gateway-detail',
  templateUrl: './gateway-detail.component.html',
  styleUrls: ['./gateway-detail.component.css']
})
export class GatewayDetailComponent implements OnInit {
  public alert: GlobalMessageClass;

  policiesPagination = new Pagination();
  servicePagination = new Pagination();
  serviceGroupPagination = new Pagination();

  loading: Boolean = false;

  policies: Array<Object> = [];
  gatewayServices: Array<Object> = [];
  gatewayServiceGroup: Array<Object> = [];

  gateway = new Gateway();
  disableEdit = false;

  public contextualHelp: Array<any>;
  public showContext: boolean = false;

  gatewayHardware: string;
  public selectedController: string;
  public selectedProtocol: string;

  public deployment_locations = [
    { value: 0, display: 'Off-Premises' },
    { value: 1, display: 'On-Premises' }
  ];

  oss = [
    { value: DeviceOSEnum.UbuntuServer18, display: 'Ubuntu Server' },
    {
      value: DeviceOSEnum.UbuntuServer20armv8,
      display: 'Ubuntu Server (ARMv8)'
    }
    // { value: DeviceOSEnum.CentosServer, display: 'CentOS Server' }
    // { value: DeviceOSEnum.WindowsServer, display: 'Windows Server' }
  ];

  public communication_protocol = [
    { value: 0, display: 'TLS' },
    { value: 1, display: 'DTLS' }
  ];

  constructor(
    private router: Router,
    private route: ActivatedRoute,
    private gatewayService: GatewayService,
    private globalMessageService: GlobalMessageService
  ) {
    this.contextualHelp = ['', ''];
  }

  showContextHelp(name: any): void {
    if (name == ContextHelpEnum.Name) {
      this.contextualHelp = ['Name', 'Descriptive name of the gateway.'];
      this.showContext = true;
    } else if (name == ContextHelpEnum.email) {
      this.contextualHelp = [
        'email',
        'Email address of admin user responsible for gateway deployment. Gateway package will be sent to this user.'
      ];
      this.showContext = true;
    } else if (name == ContextHelpEnum.ipv4_address) {
      this.contextualHelp = [
        'ipv4_address',
        'IP of the machine where gateway will be deployed.'
      ];
      this.showContext = true;
    } else if (name == ContextHelpEnum.port) {
      this.contextualHelp = [
        'port',
        'Port where the gateway will be listening for requests.'
      ];
      this.showContext = true;
    } else if (name == ContextHelpEnum.public_ipv4_address) {
      this.contextualHelp = [
        'public_ipv4_address',
        'Forwarded public IP of the gateway.'
      ];
      this.showContext = true;
    } else if (name == ContextHelpEnum.public_ipv4_port) {
      this.contextualHelp = ['public_ipv4_port', 'Public port of the gateway.'];
      this.showContext = true;
    } else if (name == ContextHelpEnum.spa_port) {
      this.contextualHelp = [
        'spa_port',
        'UDP port where gateway will be listening for SPA requests.'
      ];
      this.showContext = true;
    } else if (name == ContextHelpEnum.hardware_id) {
      this.contextualHelp = [
        'hardware_id',
        'Unique ID of the machine the gateway is bound to. This is visible when gateway software starts and should be updated from there.'
      ];
      this.showContext = true;
    } else if (name == ContextHelpEnum.selectedController) {
      this.contextualHelp = [
        'selectedController',
        'Assigned controller of the gateway.'
      ];
      this.showContext = true;
    } else if (name == ContextHelpEnum.DeviceLocation) {
      this.contextualHelp = ['DL', 'Device Location for gateway.'];
      this.showContext = true;
    } else if (name == ContextHelpEnum.buildOs) {
      this.contextualHelp = ['buildOs', 'Supported OS of the gateway.'];
      this.showContext = true;
    }
  }

  closeContextualHelp(): void {
    this.showContext = false;
  }

  ngOnInit() {
    this.disableEdit = true;
    this.loading = true;

    this.getGatewayDetail();
    this.getGatewayServices();

    this.gateway.buildOs = this.oss[0].value;
    this.gateway.deployment_location = this.deployment_locations[0].value;
  }

  getGatewayDetail() {
    this.gatewayService
      .getgatewaydetail(this.route.snapshot.params.id)
      .subscribe(
        gateway => {
          this.loading = false;
          if (!gateway[0]) {
            this.alert = {
              closed: false,
              type: 'danger',
              message: 'No gateway found with the specified ID.'
            };
            this.globalMessageService.pushAlert(this.alert);
            this.router.navigate(['/gateways']);
          } else {
            let g = gateway[0];
            this.gateway = g;
            this.gateway.communication_protocol = g.security_protocol == 'TLS' ? 
            this.communication_protocol[0].value : this.communication_protocol[1].value;
            this.gatewayHardware = g.hardware_id;
            this.selectedController = g.Controllers[0].id;
            this.selectedProtocol = g.protocol;
          }
        },
        err => {
          this.loading = true;
        }
      );
  }

  getGatewayServices() {
    this.gatewayService
      .getGatewayServices(this.route.snapshot.params.id)
      .subscribe(
        (services: any) => {
          this.loading = false;
          this.policies = services.policies;
          this.gatewayServices = services.gatewayServices;
          this.gatewayServiceGroup = services.gatewayServiceGroup;
          this.policiesPagination.totalItems = services.policies?.length;
          this.servicePagination.totalItems = services.gatewayServices?.length;
          this.serviceGroupPagination.totalItems =
            services.gatewayServiceGroup?.length;
        },
        error => {
          this.loading = true;
          this.policies = [];
          this.gatewayServices = [];
          this.gatewayServiceGroup = [];
        }
      );
  }

  cancelHandler() {
    this.router.navigate(['/gateways']);
  }
}
