import { Component, OnInit } from '@angular/core';
import { RestoreService } from './restore.service';
import { GlobalMessageService } from '../global-messages/global-messages.service';
import { UserService } from '../users/user.service';
import { Router } from '@angular/router';
import { Popup } from '../Popup';
import { GlobalMessageClass } from '../global-messages/global-messages';
import { ErrorCodeEnum } from '../shared/emuns/error.enum';
import { KeyboardKeyCodeEnum } from '../shared/emuns/global.enum';
import { ConfirmModalComponent } from '../shared/confirm-modal.component';
import { MatDialog } from '@angular/material/dialog';
import { Title } from '@angular/platform-browser';

@Component({
  selector: 'app-restore',
  templateUrl: './restore.component.html',
  styleUrls: ['./restore.component.css']
})
export class RestoreComponent implements OnInit {
  recoverButtonDisabled = true;
  isRecovering = false;
  fileData = {};
  currentUserName: string = '';
  authPassword: string = '';
  failedPassword: number = 0;
  reSubmit: boolean = false;
  showAuthError: boolean = false;
  getAuth: Popup = new Popup();
  isDarkMode : boolean
  constructor(
    private restoreService: RestoreService,
    private globalMessageService: GlobalMessageService,
    private userService: UserService,
    private router: Router,
    public dialog: MatDialog,
    private titleService: Title
  ) {}

  ngOnInit() {
    this.isDarkMode = JSON.parse(localStorage.getItem('darkMode'));
    this.titleService.setTitle('Restore');
    this.populateCurrentUserName();
    let warningAlert: GlobalMessageClass = {
      closed: false,
      type: 'danger',
      message: 'Exercise this feature carefully to avoid data loss',
      sticky: true,
      name: 'restoreWarningAlert'
    };
    this.globalMessageService.pushAlert(warningAlert);
  }

  ngOnDestroy() {
    this.globalMessageService.removeAlertByName('restoreWarningAlert');
  }

  private populateCurrentUserName() {
    let current_user = JSON.parse(localStorage.getItem('currentUser'));
    if (current_user) {
      this.currentUserName = current_user.userName;
    } else {
      this.currentUserName = '';
    }
  }

  handleFileUpload(event) {
    if (event.target.files && event.target.files.length > 0) {
      let file = event.target.files[0];
      let name = file.name;
      let type = file.type;
      let localName = name.split('.');
      let ext = localName[localName.length - 1];
      if (ext == 'sql' || type == 'application/sql') {
        this.recoverButtonDisabled = false;
        let reader = new FileReader();
        reader.readAsDataURL(file);
        reader.onload = () => {
          let data = reader.result;
          this.fileData = {
            data: data,
            name: name,
            type: type
          };
        };
      } else {
        this.recoverButtonDisabled = true;
        alert('Please select a valid file type, these are allowed types `sql`');
      }
    }
  }

  handleRecover() {
    const dialogRef = this.dialog.open(ConfirmModalComponent, {
      panelClass: this.isDarkMode ? 'dark-theme' : '',
      data: {
        title: 'Confirmation',
        message: 'Are you sure you want to recover?',
      },
    });

    dialogRef.afterClosed().subscribe((result) => {
      if (result){
        this.showPasswordPopup();
      }
    });
  }

  private recoverDb(password, username = null) {
    this.isRecovering = true;
    this.restoreService.recoverFile(this.fileData, password).subscribe(
      response => {
        this.isRecovering = false;
        this.showAlert('info', 'Database Recovered', false);
        this.cancelPopup();
      },
      error => {
        this.isRecovering = false;
        if (error.status === ErrorCodeEnum.Forbidden) {
          // this.showAlert('danger', 'Invalid password provided', false);
        } else {
          // this.showAlert('danger', error._body, false);
        }
      }
    );
  }

  showPasswordPopup() {
    this.getAuth.heading = 'Verification required';
    this.getAuth.body = 'Please enter your current password for verification.';
    this.getAuth.action = 'Confirm';
    this.getAuth.show = true;
  }

  cancelPopup() {
    this.getAuth.show = false;
  }

  authAction() {
    this.recoverDb(this.authPassword, this.currentUserName);
  }

  submitMyAlert(event) {
    if (event.keyCode == KeyboardKeyCodeEnum.Enter) {
      if (!this.reSubmit && this.authPassword.length != 0) {
        this.authAction();
      }
    }
  }

  hideAuthError() {
    this.showAuthError = false;
  }

  private showAlert(type: string, message: string, sticky: boolean) {
    const alert = {
      closed: false,
      type: type,
      message: message,
      sticky: sticky
    };
    this.globalMessageService.pushAlert(alert);
  }
}
