import { Component, OnInit } from '@angular/core';
import { Router } from '@angular/router';
import { debounceTime, distinctUntilChanged, switchMap } from 'rxjs/operators';
import { Observable, of, Subject } from 'rxjs';
import { Pagination } from '../models/pagination';
import { PreserveSearch } from '../models/preserve-search';
import { SaasBridgeService } from './saas-bridge.service';
import { GlobalMessageService } from '../global-messages/global-messages.service';
import { Title } from '@angular/platform-browser';
import { RolesAndPermissionService } from '../roles-and-permission/roles-and-permission.service';

@Component({
  selector: 'app-saas-bridge',
  templateUrl: './saas-bridge.component.html',
  styleUrls: ['./saas-bridge.component.css']
})
export class SaasBridgeComponent implements OnInit {
  showHelpText: boolean = false;
  pagination = new Pagination();

  searchKeyword: any = '';
  isSearching: boolean = false;
  bridges: any = {};
  $searchBridge = new Subject();
  loading: boolean = true;

  bridgeId: Number = 0;

  showPasswordConfimation: Boolean = false;

  pageSize = [10, 25, 50, 100, 200];

  preserveSearch = new PreserveSearch();
  navigationToEdit: boolean = false;

  sort: string = 'id';
  dir: string = 'desc';

  createPermission: boolean = true;
  readPermission: boolean = true;
  updatePermission: boolean = true;
  deletePermission: boolean = true;

  constructor(
    private globalMessageService: GlobalMessageService,
    private saasBridgeService: SaasBridgeService,
    private router: Router,
    private titleService: Title,
    private rolesPermissionService : RolesAndPermissionService
  ) {}

  ngOnInit() {
    this.titleService.setTitle('Infrastructure | SaaS Bridges');
    this.getCRUDdata(31);
    const saveSearch = this.preserveSearch.getSearch('saasBridgeSaveSearch');
    if (saveSearch) {
      this.pagination = saveSearch.pagination;
      this.searchKeyword = saveSearch.searchKeyword;
    }
    this.getPage();
    this.subscibeToSearch();
  }

  getCRUDdata(privId){
    this.rolesPermissionService.getCrudData(privId).subscribe((res: any) => {
      this.createPermission = res?.create_permission === 1 ? true : false
      this.readPermission = res?.read_permission === 1 ? true : false
      this.updatePermission = res?.update_permission === 1 ? true : false
      this.deletePermission = res?.delete_permission === 1 ? true : false
    });
  }

  getPage() {
    this.loading = true;
    this.saasBridgeService
      .getBridges(this.pagination, this.sort, this.dir, this.searchKeyword)
      .subscribe({
        next: res => {
          this.loading = false;
          this.bridges = res;
        },
        error: err => {
          this.loading = false;
        }
      });
  }

  clearfun() {
    this.searchKeyword = '';
    this.getPage();
  }

  confirmPassword(id) {
    this.bridgeId = id;
    this.showPasswordConfimation = true;
  }

  deleteBridge(event) {
    this.loading = true;
    this.saasBridgeService.deleteBridge(this.bridgeId, event).subscribe({
      next: res => {
        this.getPage();
        this.showAlert('success', 'SaaS Bridge Deleted.', false);
        this.showPasswordConfimation = false;
      },
      error: err => {
        this.loading = false;
      }
    });
  }

  subscibeToSearch() {
    this.$searchBridge
      .pipe(
        debounceTime(500),
        distinctUntilChanged(),
        switchMap(data => {
          this.loading = true;
          this.isSearching = true;
          this.searchKeyword = data;
          this.pagination.currentPage = 1;
          this.getPage();
          return of('Fetching data...');
        })
      )
      .subscribe(() => {
        this.isSearching = false;
      });
  }

  cancelVerificationPassword() {
    this.showPasswordConfimation = false;
  }

  addSaaSBridgeHandler() {
    this.router.navigate(['/saas-bridge/create']);
  }

  pageSizeChange(size) {
    this.pagination.itemsPerPage = +size;
    this.pagination.totalItems = 0;
    this.pagination.currentPage = 1;
    this.getPage();
  }

  pageChange(page: any) {
    this.pagination.currentPage = page;
    this.getPage();
  }

  editHandler(id) {
    const preserve = {
      pagination: this.pagination,
      searchKeyword: this.searchKeyword
    };
    this.preserveSearch.saveSearch('saasBridgeSaveSearch', preserve);
    this.navigationToEdit = true;
    this.router.navigate([`saas-bridge/edit/${id}`]);
  }

  private showAlert(type: string, message: string, sticky: boolean) {
    const alert = {
      closed: false,
      type: type,
      message: message,
      sticky: sticky
    };
    this.globalMessageService.pushAlert(alert);
  }
}
